%%
% AccGuessObj
%
% Front-end to use Richard and Gabriele's accelerated cavity field calculation, R_getguessFin
%
% May 6, 2019
%    inject2IFO does not calculate transmitted and reflected field of laser by the front mirror.
%    This is done one when the IFO is built. This is to utilize fields calculated in the previous lock or calc.
%
% May 18, 2018
%    Clean up with minor update
%
% February 26th, 2018
%    function EgtList = isEgtSet( obj ) added to check if any of the external sources defined
%    function guessIndex = isGuessField( obj, fldID ) added to find the guess field index whose guess field ID is fldID
%    In calcgLoopPhase(), protection is set for the case when lockProp < 0.
%
% February 12th, 2018
%   defineGuess( guessID, convID, lockProp(srcField), ... ) these three inputs can be field names, instead of field IDs
%
% February 9th, 2018
%  In lockAction, gLoopPhase calculation changed to make the locking more stable.
%  gLoopPhase(2:3,:) returns round trip phase themselves, to be locked phase not subtracted
%
% February 8th, 2018
%  1) In lockAction, phase calculation is changed based on the suggestion by Gabriele.
%    For the field without external source, use seld round trip.
%    For the field with external source, use seld round trip and contribution of other guess fields
%  2) delL gain is set to be constant until full convergence, instead of switching to 1.
%  3) defineGuess( ... delLG ), default value of delLG is set to 0.3, not 1
%
% February 4, 2018
%  lockAction is modified to multiply gain for each length change,
%     delL = delL based on round trip phase * gain
%  For this, delLGain is added and delLGain property added
%
% March 31, 2017
%   When the object is created and lockHandle function is provided,
%   this is called as
%     phase = customgLoopPhase( obj, AccGuessObj, whichGID, startField, nextField )
%   return nan if not calculated and default value is to be used.
%
% March 30, 2017
%   mainAction( ..., Nprint ) has more control by Nprint, supporting no printout for Nprint <= 0
%
% March 24, 2017
%   Cleanup.
%   obj.ifo.ffts.zeroMat is used to fill zero field amplitudes.
%
% March 8, 2017  by Hiro Yamamoto
%   First implementation based on AdVirgoIMCIFOAcc code
%   Uses GuessObj code for tracing evolution paths for each guess gield
%
%% Usage of this object class
%
% 1) create object
%   AccG = AccGuessObj( 'IMC', ifo, AccN, AveN, @lockHandle );
%
% 2) register guess fields, convergence test field, propagator for locking, locked phase, and
%    path list for all fields starting from this guess field. EA -> EAA, EA->EBA, EA->ECA, etc
%   AccG.defineGuess( guessFieldID, convergenceTestFieldID, propIDForLock, lockPhase, pathList )
%       pathList = {{path1}, {path2}, ...}
%                  {{'MC1-HRx-o', 'MC3-HRx-o', 'MC2-HRx-o', 'MC1-HRx-o'}, ...
%                   {'MC1-HRx-o', 'MC3-HRx-o', 'MC2-HRy-o'}} );
%       The orders in the path list can be arbitrary.
%
% 3) set the external input field
%   after all defineGuess are called, define source field
%   AccG.setEt( guessIndex(1), EFFT for ID );
%
% 4) main action to calculate or lock the system
%   [err, count] = AccG.mainAction( doLock, 1000, errorLimit, Nprint );
%
%% Code structure    (methods marked by => can be overridden)
%
%    for eLoop = 1 : maxEvent
%       for m=1:obj.AccN
%          for l=1:obj.AveN(m)
%             for gid1 = 1 : obj.Nguess
% =>             obj.forEachGuess(gid1);
%                obj.GObj(gid1).goMasterPath
%             end
%          end
%       end
% =>    theErr(eLoop,1:obj.Nguess) = obj.evalErr;
%       new guess values = R_getguessFin;
% =>    obj.lockAction( eLoop );
%    end
% => finishAction;
%

%%
classdef AccGuessObj < handle
    
    properties
        name = 'unnamed'
        
        Nguess    % number of guess fields, automatically incremented by defineGuess call
        
        % inputs
        guessIDs  % array of field IDs of independent guess fields
        convIDs   % array of field IDs to be used for convergence test for each guessID
        
        AccN      % number of steps for accelaration calculation
        AveN      % number of averages to be used for each step
        % AveN(1:AccN) is [N,N+1,N,N+1,...] to avoid possible systematics
        % cauased by even or odd propagation
        
        lockPhase % round trip phase after lock
        lockProp  % if > 0, use this propagator to lock
        delLG     % delL for lock = delL based on round trip phase * this gain
        
        doublePath % true if lockProp and reverse of lockProp are in the EA->EA path
        % If true, delL for lock is halved
        
        GObj % guess field obj for each guess field
        
        Eg1  % EG1(AccN, Nguess) guess field
        Eg2  % EG2(AccN, Nguess, Nguess) guess field A to guess field B evolution
        Egt  % source field of Eg1
        
        % information of the system
        ifo
        
        % called to calculate round trip phase
        % phase = customgLoopPhase( obj, AccGuessObj, whichGID, startField, nextField )
        customgLoopPhase
        
        % lock code called at the end of lockAction
        customLockCode
        
        % custom code to fill theErr output
        customErrorBuffering
    end
    
    methods
        
        % constructor
        function obj = AccGuessObj( name, ifo, AccN, AveN )
            obj.name = name;
            obj.ifo = ifo;
            obj.guessIDs = [];
            obj.Nguess = 0;
            obj.AccN = AccN;
            obj.GObj = GuessObj.empty(0);
            
            obj.Eg1 = cell(AccN+1,0);
            obj.Eg2 = cell(AccN+1,0,0);
            if AveN==0 % simple average
                obj.AveN=ones(1,AccN);
            else       % n, n+1, n, n+1 ... to avoid even only or odd only case anomaly
                obj.AveN=AveN*ones(1,AccN);
                obj.AveN(2:2:end)=obj.AveN(2:2:end)+1;
            end
            
            obj.customgLoopPhase = '';
            obj.customLockCode = '';
            obj.customErrorBuffering = '';
        end
        
        %% mainAction
        
        % interface to acclerated conversion code R_getguessFin by Richard and Gabriele
        % Nprint < 0 : no printout even for the end notice
        % Nprint = 0 : no printout except for the last end notice
        % Nprint > 0 : printut every Nprint'th event and the last end notice
        % doLock : 1 : all lock
        %          0 : no length adjustment
        function [theErr, theCounter] = mainAction( obj, doLock, maxEvents, errLimit, Nprint )
            theErr = zeros(0,1);
            theCounter = zeros(0,1);
            
            % roundtrip phase calculated using the first loop
            gLoopPhase = zeros(3,obj.Nguess);
            
            % initialize fields used for accelaration
            
            % Eg1(A, AccStep) = guessField(A, AccStep)
            obj.Eg1 = cell( obj.Nguess, obj.AccN+1 );
            % Eg2(A, B) is the change of fields
            % Eg2(A, A) = A - P(A->A)[A], F in the paper
            % Eg2(A, B) = P(B->A)[B], G in the paper
            obj.Eg2 = cell( obj.Nguess, obj.Nguess, obj.AccN ); % A -> B
            
            % convergence test fields before and after propagation
            Econv = cell(obj.Nguess, 1);  % field used for convergence test
            
            % reset propagator counter
            obj.ifo.props.propCounter = 0;
            
            % initial values of fields used for convergence check
            for gid1 = 1 : obj.Nguess
                Econv{gid1} = obj.ifo.fields.fieldList(obj.convIDs(gid1)).EFFT;
            end
            
            % loop until convergces
            for eLoop = 1 : maxEvents
                obj.Eg1(:) = {obj.ifo.ffts.zeroMat};
                obj.Eg2(:) = {obj.ifo.ffts.zeroMat};
                
                for iAcc=1:obj.AccN
                    for iAve=1:obj.AveN(iAcc)
                        
                        % EA2EB is propagator for all A->B, i.e., E(B) = EA2EB(A, B)[E(A)]
                        % for each Acc and Ave step.
                        EA2EB = cell( obj.Nguess, obj.Nguess );
                        
                        for gid1 = 1 : obj.Nguess
                            % do whavever necessary action for each guess field
                            obj.forEachGuess(gid1);
                            % Eg1{gid1, iAcc} = sum( guessField(gid1, iAcc) ) / Nave
                            obj.Eg1{gid1,iAcc} = obj.Eg1{gid1,iAcc} + obj.GObj(gid1).guessField;
                            
                            % goMasterPath calculates all fields starting from this (gid1'th) guess field
                            % the paths to these fields are defined in pathList
                            % EA -> AEA, EA -> CEA, etc
                            for gid2 = 1 : obj.Nguess
                                obj.ifo.fields.fieldList(obj.guessIDs(gid2)).EFFT = obj.ifo.ffts.zeroMat;
                            end
                            obj.GObj(gid1).goMasterPath;
                            
                            % EA2EB{gid1,gid1} = E[ A <- A ]
                            EA2EB{gid1,gid1} = obj.ifo.fields.fieldList(obj.guessIDs(gid1)).EFFT;
                            
                            % Eg2{gid1,gid1,iAcc} = sum( guessField(A) - E[ A <- A ] ) / Nave for iAcc (accelaration step)
                            obj.Eg2{gid1,gid1,iAcc} = obj.Eg2{gid1,gid1,iAcc} + ...
                                (obj.GObj(gid1).guessField - EA2EB{gid1,gid1});
                            
                            for gid2 = 1 : obj.Nguess
                                if gid2 ~= gid1
                                    % EA2EB{gid1,gid2} = EB{gid2} <- EA{gid1}
                                    EA2EB{gid1,gid2} = ...
                                        obj.ifo.fields.fieldList(obj.guessIDs(gid2)).EFFT;
                                    % Eg2{gid1,gid2,iAcc} = sum( EA2EB ) / Nave
                                    obj.Eg2{gid1,gid2,iAcc} = obj.Eg2{gid1,gid2,iAcc} + EA2EB{gid1,gid2};
                                end
                            end
                            
                        end % for gid1 = 1 : obj.Nguess
                        
                        % once in a stationary field search
                        if iAcc == 1 && iAve == 1
                            gLoopPhase = obj.calcgLoopPhase( EA2EB, gLoopPhase(2,:) );
                        end
                        
                        % update guessField = sum( EA2EB( all src to gid1 ) ) + Et(gid1, if external source exists)
                        for gid1 = 1 : obj.Nguess
                            obj.GObj(gid1).guessField = EA2EB{gid1,gid1};
                            if ~isempty( obj.Egt{gid1} )
                                obj.GObj(gid1).guessField = ...
                                    obj.GObj(gid1).guessField + obj.Egt{gid1};
                            end
                            for gid2 = 1 : obj.Nguess
                                if gid2 ~= gid1
                                    % field start from gid2 and arriving at gid1
                                    % is added to the guessField at the beginning of gid1 loop
                                    obj.GObj(gid1).guessField = ...
                                        obj.GObj(gid1).guessField + EA2EB{gid2,gid1};
                                end
                            end
                        end % for gid1 = 1 : obj.Nguess
                        
                    end  % for iAve=1:obj.AveN(iAcc)
                    
                    for gid1 = 1 : obj.Nguess
                        for gid2 = 1 : obj.Nguess
                            obj.Eg2{gid2,gid1,iAcc} = obj.Eg2{gid2,gid1,iAcc} / obj.AveN(iAcc);
                        end
                        obj.Eg1{gid1,iAcc} = obj.Eg1{gid1,iAcc} / obj.AveN(iAcc);
                    end % for gid1 = 1 : obj.Nguess
                    
                end  % for iAcc=1:obj.AccN
                
                % do the accelaeration magic
                EGvec = R_getguessFin(reshape(obj.Eg2,obj.Nguess^2,obj.AccN), obj.Eg1, obj.Egt);
                
                % update guessFields by the estimation
                for gid1 = 1 : obj.Nguess
                    obj.GObj(gid1).guessField = EGvec{gid1};
                end
                
                % new Econv is calculated and compared with old one to estimate the convergence
                [errorEstimation, Econv] = obj.evalErr( Econv );
                errMax = max(errorEstimation);
                
                locked = false;
                if doLock
                    locked = obj.lockAction( eLoop, gLoopPhase );
                end
                % update guessFields by the estimation
                for gid1 = 1 : obj.Nguess
                    obj.ifo.fields.fieldList(obj.guessIDs(gid1)).EFFT = obj.GObj(gid1).guessField;
                end
                
                if ~isempty( obj.customErrorBuffering )
                    theErrOne = obj.customErrorBuffering( errorEstimation, gLoopPhase, obj );
                else
                    theErrOne = obj.fillErrBuf( errorEstimation, gLoopPhase );
                end
                theErr(eLoop, 1:size(theErrOne,1), 1:size(theErrOne,2)) = theErrOne;
                theCounter(eLoop)=obj.ifo.props.propCounter;
                
                if errMax < errLimit && min(errorEstimation) > 0
                    if Nprint >= 0
                        fprintf(1, 'Ending ...\n%s\n', obj.printAction( theErr, theCounter, locked ));
                    end
                    break;
                end
                
                if Nprint > 0
                    if mod( eLoop-1, Nprint ) == 0
                        fprintf(1, '%s\n', obj.printAction( theErr, theCounter, locked ));
                    end
                end
                
            end   % for eLoop = 1 : maxEvents
            
            obj.finishAction;
        end
        
        function theErr = fillErrBuf( obj, errorEstimation, gLoopPhase )
            theErr(1,1:obj.Nguess) = errorEstimation;
            theErr(2,1:obj.Nguess) = obj.ifo.power(obj.convIDs);
            theErr(3,1:obj.Nguess) = gLoopPhase(1,:);
            theErr(4,1:obj.Nguess) = gLoopPhase(2,:);
            theErr(5,1:obj.Nguess) = gLoopPhase(3,:);
        end
        
        %% calcgLoopPhase
        %  calculate loop phase, and store results in gLoopPhase
        %  gLoopPhase(1,:) = delL ( = 0 for cavities which are not locked )
        %  gLoopPhase(2,:) = round trip phase
        %  gLoopPhase(3,:) = kL - gouy phase for this loop
        %  EA2EB(gdid1, gdid2) = E(gid1) -> E(gid2)
        %  For each guessField,
        %         start amplitude is obj.GObj(gid1).guessField, and
        %         final amplitude is stored in EA2EB{gid1, gid1}
        function gLoopPhase = calcgLoopPhase( obj, EA2EB, lastPhase )
            
            gLoopPhase = zeros(3, obj.Nguess);
            % calculate angle using initial, GObj{gid1).guessField, and final, EA2EB{gid1,grid1}, fields
            for gid1 = 1 : obj.Nguess
                % Calculate fields at the guessField ports after propagations
                % EA2EB{gid1,gid1} = A -> A
                % add B -> A if A has external sources
                if ~isempty( obj.Egt{gid1} )
                    for gid2 = 1 : obj.Nguess
                        if gid2 ~= gid1
                            EA2EB{gid1,gid1} = EA2EB{gid1,gid1} + EA2EB{gid2,gid1};
                        end
                    end
                end
                
                phi = nan;
                if ~isempty( obj.customgLoopPhase )
                    % if customgLoopPhase exists, try to let it handle phase calculation
                    % it returns nan, if it cannot handle it
                    phi = obj.customgLoopPhase( obj, gid1, EA2EB{gid1,gid1} );
                end
                if isnan( phi )
                    EEProd = sum(sum( conj(obj.GObj(gid1).guessField) .* EA2EB{gid1,gid1} ));
                    phi = atan2( imag(EEProd), real(EEProd) );
                end
                % This make code stable ...
                unwrapped = unwrap( [ lastPhase(gid1), phi ] );
                phi = unwrapped(2);
                
                if obj.lockProp(gid1) > 0
                    delL = (phi - obj.lockPhase(gid1)) / obj.ifo.fields.kCR;
                    % For a triangular cavity, both paths should be changed, but delL should not be halved.
                    % doublePath is true if both are in the field evolution path
                    if obj.doublePath
                        delL = delL / 2;
                    end
                    delL = obj.delLwithGain( delL, gid1 );
                else
                    delL = 0;
                end
                
                gLoopPhase(1,gid1) = delL;
                gLoopPhase(2,gid1) = phi;
                % lockProp < 0 for unlocking propagator, so make is abs()
                if obj.lockProp(gid1) > 0
                    gLoopPhase(3,gid1) = obj.ifo.cavities.kLGouyOfClosedCavity( abs(obj.lockProp(gid1)) );
                end
            end
        end
        
        %% lockAction
        %     calls customgLoopPhase and/or customLockCode if provided
        %     acted = true if some locktion is done. If eLoop condition is done, may be false
        function acted = lockAction( obj, eLoop, gLoopPhase )
            
            done = false;
            if ~isempty( obj.customLockCode )
                % if all length changes are done, return true.
                % otherwise, return false and let this lockAction takes care
                done = obj.customLockCode( obj, eLoop, gLoopPhase );
            end
            if ~done
                for gid1 = 1 : obj.Nguess
                    % setDelL changes forward and backward propagator length
                    if obj.lockProp(gid1) > 0
                        obj.ifo.props.setDelL2( obj.lockProp(gid1), obj.ifo.props.propList(obj.lockProp(gid1)).delL + gLoopPhase(1,gid1) );
                    end
                end
            end
            acted = true;
        end
        
        % delL gain is independent on the current error, not setting to 1.
        function gainL = delLwithGain( obj, delL, gid1 )
            gainL = delL * obj.delLG(gid1); % delLG can be different for each guess field
        end
        
        %% other functions
        % function defineGuess( obj, guessID, convID, lockProp, lockPhase, pathList, delLG )
        %
        % set the field evolution flows for the guess field guessID
        % convID is used to save a field to test convergence
        % If lockProp > 0, chage this cavity length (and the back propergator if exists)
        % to adjust the self-roundtrip phase (EA->EA) to lockPhase
        % pathList is a cell arrays of paths to define evolutions from EA to EX(i)
        %  { {'ETM-HR-o', 'ITM-HR-o', 'ETM-HR-o'}, {'ETM-HR-o', 'ITM-AR-o'} } or {{1:3}, {4:5}}
        % If lockPhase is nan, sum of lockPhase of all propagators in the loop is used
        function index = defineGuess( obj, guessID, convID, lockProp, lockPhase, pathList, delLG )
            guessID = obj.ifo.fields.findNamedField( guessID );
            convID = obj.ifo.fields.findNamedField( convID );
            index = length( obj.guessIDs ) + 1;
            obj.GObj(index) = GuessObj( sprintf('AccGuess%d',index), 100*index, obj.ifo, guessID );
            obj.guessIDs(index) = guessID;
            obj.convIDs(index) = convID;
            obj.doublePath = false;
            for id = 1 : length( pathList )
                pathIndex = obj.GObj( index ).addPath( pathList{id} );
                if lockProp > 0
                    if obj.GObj(index).areBothPaths( lockProp, pathIndex )
                        obj.doublePath = true;
                    end
                end
            end
            obj.GObj( index ).buildMasterPath( GuessObj.actFillGuessField );
            obj.Nguess = index;
            obj.Egt = cell(1,index);
            obj.lockProp(index) = lockProp;
            if lockProp > 0
                if isnan( lockPhase )
                    lockPhase = obj.ifo.cavities.totalLockPhaseOfClosedCavity(lockProp);
                end
                obj.lockPhase(index) = lockPhase;
                if ~exist('delLG', 'var')
                    delLG = 0.3;
                end
                obj.delLG(index) = delLG;
            end
            
            if obj.ifo.fields.fieldList(guessID).status <= 0
                obj.ifo.fields.setDefField( guessID, 1 );
            end
            obj.GObj(index).guessField = obj.ifo.fields.fieldList(guessID).EFFT;
        end
        
        function resetGuessField( obj )
            for ig = 1 : obj.Nguess
                guessID = obj.guessIDs(ig);
                obj.ifo.fields.setDefField( guessID, 1 );
                obj.GObj(ig).guessField = obj.ifo.fields.fieldList(guessID).EFFT;
            end
        end
        
        % freeze guessID cavity by setting the one trip phase to frezePhase, if given
        function freezeOne( obj, guessIndex, freezePhase )
            propID = abs(obj.lockProp(guessIndex));
            obj.lockProp(guessIndex) = -propID;
            if exist('freezePhase', 'var')
                obj.lockPhase(guessIndex) = freezePhase;
                obj.ifo.props.propList(propID).resetDelL( freezePhase );
                backProp = obj.ifo.props.findReverseProp( propID );
                if backProp > 0
                    obj.ifo.props.propList(backProp).resetDelL( freezePhase );
                end
            end
        end
        
        function setEt( obj, guessIndex, Et )
            obj.Egt{ guessIndex } = Et;
        end
        
        % this is called before a field evolution starts
        % so prepare anything for this guessID field
        function forEachGuess( ~, ~ )
            % place holder
        end
        
        % estimate change of the fields to judge if guess fields are converging
        function [errors, EconvNew] = evalErr( obj, EconvOld )
            
            errors = zeros(1,obj.Nguess);
            EconvNew = cell(obj.Nguess, 1);  % field used for convergence test
            powerNew = zeros(obj.Nguess, 1);
            % update Econv
            for gid = 1 : obj.Nguess
                if obj.convIDs(gid) == obj.guessIDs(gid)
                    EconvNew{gid} = obj.GObj(gid).guessField;
                else
                    obj.GObj(gid).goMasterPath;
                    EconvNew{gid} = obj.ifo.fields.fieldList( obj.convIDs(gid) ).EFFT;
                end
                powerNew(gid) = sum(sum(abs(EconvNew{gid}).^2));
            end
            
            maxPwr = max( powerNew );
            for gid = 1 : obj.Nguess
                powerOld = sum(sum( abs(EconvOld{gid}).^2 ));
                errors(gid) = max( ...
                    [ sqrt( sum(sum(abs(EconvNew{gid}-EconvOld{gid}).^2)) / maxPwr ), ...
                    abs(powerNew(gid)-powerOld) / (powerNew(gid)+powerOld) ] );
            end
        end
        
        %  theErr(3,:) = gLoopPhase(1,:) = delL ( = 0 for cavities which are not locked )
        %  theErr(4,:) = gLoopPhase(2,:) = round trip phase
        %  theErr(5,:) = gLoopPhase(3,:) = kL - gouy phase for this loop
        function str = printAction( obj, theErr, theCounter, locked )
            str = sprintf('%3d (FFT %6d)', length(theCounter), theCounter(end) );
            for ig = 1:obj.Nguess
                str = sprintf('%s : ''%s'' Pwr=%10.3e, Err=%10.3e, phiLoop=%10.3e, kL-gouy=%10.3e', str, obj.ifo.fields.getName( obj.convIDs(ig) ), ...
                    obj.ifo.power(obj.convIDs(ig)), theErr(end,1,ig), theErr(end,4,ig), theErr(end,5,ig));
                if locked && obj.lockProp(ig) > 0
                    str = sprintf('%s, delL=%10.3e', str, theErr(end,3,ig));
                end
                if ig ~= obj.Nguess
                    str = sprintf('%s\n                ', str);
                end
            end
        end
        
        % called just before existing mainAction
        function finishAction( obj )
            obj.ifo.cavities.resetFieldStatus;
            obj.ifo.inject2IFO(false);
            
            for ig = 1 : obj.Nguess
                obj.ifo.fields.fieldList(obj.guessIDs(ig)).EFFT = obj.GObj(ig).guessField;
                obj.ifo.fields.fieldList(obj.guessIDs(ig)).status = 1024*3;
            end
        end
        
        % change all cavity length, delL, to the one specified by lockPhase in prop obj
        % if guessField has lockProp < 0, freeze it to lockPhase
        function resetDelL( obj )
            obj.ifo.props.resetDelL;
            for ig = 1 : obj.Nguess
                if obj.lockProp(ig) < 0
                    obj.freezeOne( ig, obj.lockPhase(ig) );
                end
            end
        end
        
        % if any guess field has external source, Egt, set
        function EgtList = isEgtSet( obj )
            EgtList = [];
            for ig = 1 : obj.Nguess
                if ~isempty( obj.Egt{ig} )
                    EgtList(end+1) = ig;
                end
            end
        end
        
        % returns the guess index whose guess field ID is given
        % If fldID is not a guess field, return 0
        function guessIndex = isGuessField( obj, fldID )
            guessIndex = find( obj.guessIDs == fldID, 1 );
            if isempty( guessIndex )
                guessIndex = 0;
            end
        end
    end
end
