%
% classdef FFTIFO
%
% * Superclass to build IFO by adding optics and progapators
%
%   1) make a file named myIFO.m and define the myIFO as follows
% classdef IFOname < FFTIFO
%     methods
%         function obj = IFOname( varargin )
%             obj@FFTIFO( varargin{:} );
%         end
%
%         function defineIFO( obj, varargin ) : mandatory to define system
%             % set parameters
%             eval( UtilTK.setVars( who, true, varargin{:} ) );
%             % add optics
%             % connection optics
%             % setup details of optics, like maps
%         end
%
%         function setupBaseMode( obj )  : optional to define mode
%             call fillQvals to define modes in IFO
%         end
%     end
% end
%
%   2) make an object and lock or calc
%     ifo = myIFO;
%     ifo.lock;  % cavity lengths are adjusted to make ths stationary fields resonant
%     ifo.calc;  % stationary field is calculated with given cavity length
%
%   3) analye
%     fieldName = 'ITM-HR-o';  % 'Optic name' - 'side' - 'i' for in or 'o' for out
%     [E, xy] = ifo.getField(fieldName);  % E(iy,ix) is the complex amplitude at x=xy(ix), y = xy(iy)
%     plot( x, E(end/2,:) )
%     Cmn = ifo.HGCoef(fieldName, 0:5,0:5);    % HG amplitude for m=0~5, n=0~5
%     P = ifo.power(fieldName);                % power of the field
%     Fmn = ifo.HGFrac(fieldName, 0:5,0:5);    % HG power fraction, = abs(Cmn).^2 / power
%     ifo.gaussFit(fieldName)                  % fit of the data by gaussian form
%
%   4) save and load state
%     ifo = myIFO;
%     ifo.lock;
%     ifo.saveIFOState('myIFOLocked.mat'); % save current fields and props
%     ifo1 = myIFO;
%     ifo1.loadIFOState('myIFOLocked.mat'); % same as ifo, resume calculation
%
% * derived class should implement the following methods
%
%       1) defineIFO : mandatory
%           This defines the IFO to be simulated
%           call addMirror and addProp to setup the configuration
%           call setHRfiles to define the HR surface details, setTRfiles the transmission and setARfiles the AR surface
%
%       2) setupBaseMode : optional
%           This defines modes in the entire IFO by specifyinf the q value of a field.
%           obj.cavities.fillQvals( portID, qInv ) will start from the given portID with
%           qInv and propagate mode base to all optics.
%           If the mode of the laser is known, just call
%               obj.cavities.fillQvals( obj.optics.findLaser, 1/q of the input to the front mirror );
%
% * To see inforamtions
%       1) ifo.printAll, which calls
%           ifo.fields.printAll
%           ifo.optics.printAll
%           ifo.props.printAll
%           ifo.cavities.printAll
%
%       2) info,printAll('filename') saves the output in 'filename'
%
% * function groups in FFTIFO
%       * "main functions" called in FFTIFO, mainly for internal use
%       * "access functions" to fetch quantities
%       * "add optics" to construct the system and setup details of optics
%       * "Field quantities" to change them
%       * "internal support functions" to implement locking etc
%
% * Things to know
%        1) In principle, don't access data directly from the class properties,
%           always use access functions defined in this class, like getField.
%           Some properties may be calculated on demand.
%        2) FFTIFO uses adaptive grid sizes, that means, window sizes on mirrors are all different.
%           So, always get proper coordinate by [E,x] = getField
%

%%
% * Recent history
%
% May 24, 2020
%   HGPower passes varargin propperly as varargin{:} when calling HGCoef
%   ToggleLaser for the future ...
%
% December 21, 2019
%   mainXXXmodes : threshold < 0 corresponds to nominal number of modes
%
% November 5, 2019
%   1) many mode coefficient related functions have varargin (thetaX, thetaY, EFFT, see FieldObj), added 
%   so that fields can be tilted or the amplitude is explicitly given. varagin is used so that the test
%   if those values are given or not is tested only at the very last function which actually uses those values.
%   2) Non00Amp has thetaX and thetaY as new variables so that the field can be tilted. Not tested enough.
%   3) tiltOn added.
%      To use this, add tiltOn calls as follows.
%        [thetaI,phiI,thetaE,phiE] = cavityManager.beam2tilt( 0, 0, dxE, dyE, invRoCIM, invRoCEM, Lcav );
%        obj.addMirror( 'IM', ..., 'optTheta', thetaI, 'optPhi', phiI);
%        obj.addMirror( 'EM', ..., 'optTheta', thetaE, 'optPhi', phiE);
%        obj.tiltOn( 'IM', -dxE/Lcav, dyE/Lcav );%
%
% October 19, 2019
%   loadIFOState returns if the file does not exists
%
% October 10, 2019
%   power = powerInR( obj, fldID, radius, x0, y0 ) supports offset (x0,y0)
%
% Sep. 24, 2019
%   adjustLaser to change the input beam direction. This is to improve the mode matching when mirrors are tilted to change the beam axis
%   modeCoefTilt : modeCoef withrespect to a tilted axis
%
% July 9, 2019
%   saveIFOState saves and loadIFOState checks Nfft consistency.
%
% June 5th, 2019
%   function out = fieldNames( fldIDs )
%       out is a string of ids and names of field fldIDs
%       fldIDs can be 'BS-ARx-o', or 'BS*o', or 20 or 20:30
%
% June 5th, 2019
%   mainModesStr (mainLGsStr, mainHGsStr) has power, in addition to fraction
%
% May 3rd, 2019
%   reset is explicitly handled in lock()
%
% April 30th, 2019
%   inject2IFO( false ) does not change the field.
%   inject2IFO( true ) is called once at the beginning, to set the initial field for the first pumping field
%
% April 30th, 2019
%   1) fieldAction( ..., varargin ) uses the same format as FFTIFOAcc, i.e,
%      'errorLimit', 1e-6, 'reset', 0 style setting, instead of the fixed order inputs.
%   2) in fieldAction, when reset is false, inject2IFO is not called assuming that all fields are filled
%
% April 29th, 2019
%   saveIFOState( fileName ) and loadIFOState( fileName ) are modified to save and load fields and props.
%   default name of the saved data file is 'objectName' + '.mat'.
%
% April 3rd, 2019
%   laser parameter c0, the beam position, is now real, dxyz(1:3). dxyz(3) is the shift of the waist position.
%   The q value is changed to be q - z, i.e., the waist position is moved toward the input mirror.
%
% March 22, 2019
%   Non00Amp(useFit) = Eall - C*Efit
%   When gaussian Fit is used, C is calculated to make the power of (Eall - C*Efit) minimum.
%
% March 5th, 2019
%   Power of HG/LG modes, expanded
%   Pmn = HGPower( obj, fldID, m, n ), LGPower
%      Power of HG(m,n)/LG(p,m) terms
%      m > 0 : nominal coefficient
%      if m = -nmmax < 0, it returns array of powers
%          n = 1 : power of elements of n=0 case : HG(0,2), HG(1,1), HG(2,0) or LG(1,0), LG(0,2), LG(0,-2)
%          n = 0 : sum of power( m+n == nmmax ) HG(0,2) + HG(1,1) + HG(2,0), or LG(1,0) + LG(0,2) + LG(0,-2)
%          n = -1 : array of powers with n+m <= nmmax, sum(HG(n+m=0)), sum(HG(n+m=1)), sum(HG(n+m=2)) and power with n+m > nmax
%
% March 1st, 2019
%   function [E, x] = Non00Amp( obj, fldID, useFit )
%   E = E(fft) - E(fit) * exp( i * phase(E(fft)/F(fit)) )
%   or
%   E = E(fft) - E(mode) * modeCoef
%   default of useFit is false, i.e., modal model prediction
%
% February 8, 2019
%   function printLoss( fid ) added
%
% November 28, 2018
%   fillQvalFromArm2BS added
%
% August 24, 2018
%   mainLGsStr and mainHGsStr added to simplify.
%
% August 8, 2018
%   refl and trans can accept standard format field amplitude as the second argument.
%   This allows to calculate reflection / transmission of a input field with a specific mode
%
% July 7, 2018
%   mainModes and mainModesStr are revised to include loss information calculated by fieldManager.lossOutMode
%
% May 24, 2018
%   members (static) mainModesStr is separated from mainModes etc.
%   use like modes = mainLGs(); mainModesStr(modes).
%
% May 21, 2018
%   demod renamed to DemodRF and the content is placed in a separate file.
%   Demod can be circular, quadrant or using external function
%
% May 18, 2018
%   0) function [dE, E0, xl, dP, L] = getPointScatter( obj, srcID, defects, WfftOut, L )
%      Front-end of getPointScatter added
%   1) demod function is updated
%   2) power can accept name lists, like ifo.pwoer({'ITM-HR-o', 'ITM-AR-i'})\
%   3) mainModes, mainLGs and mainHGs are added which shows mode conpornents whosa fractions are large
%   4) saveIFOState and loadIFOState added to save and load all fielsds
%         a = IFO; a.lock;
%         b = IFO; a.saveIFO('savedDat.mat'); b.loadIFO('savedDat.mat');
%         b is the same as a after lock, i.e., one can resume work after saveIFO condition
%
% April 3rd, 2018
%   function saveFields and reastoreFields updated for convinience
%
% March 9, 2018
%   function Cmn = modeCoef( obj, fldIDs, fieldMode, ind1, ind2 ) added
%      LGCoef and HGCoef combined. If fieldMode is not specified, laser mode is used
%
% February 28, 2018
%   function saveFields( {'field1', 'field2', ...} ) and restoreFields( {'field1', 'field2', ...} ) are added
%      saveFields save the fields in internal buffer and are restored.
%      These are to be used when CR and SB are calculated pereatedly.
%
% February 27, 2018
%   function [ ASCs, Es] = demod( obj, fldIDs, RF, modIndex, error ) added.
%
% February 23, 2018
%   function transPort = inject2IFO( obj ) returns transPort (like PRM-HR-o )
%
% February 14, 2018
%   power argument can be [ID1, ID2], {ID1, ID2}, {'name1', 'name2'}
%
% Setpember 29, 2017
%  In fieldAction, if there is no resonator and if BS exists, lock adjusts Mich to make the dark port dark.
%  [dark, ...] = darkPort returnd dark > 1 for good dark, = 0 dark is not dark, -1 when BS does not exist
%
% September 26, 2017
%  in function darkPort, calcIfNecessary(toBSX & toBSY) are called in case these fields are not ready
%
% Septemner 8, 2017
%  1) addFieldSrc is added. This accepts another IFO object and fieldID, and that is used as the laser source.
%     LaserObj is modified to support this kind of sourece
%  2) setupBaseMode( obj, varargin ) is modified to support source field
%     from external IFO by calling obj.optics.getSourceFieldMode;
%  3) updateNfft changed to support the external field source which demands a specific Nfft
%  4) addLaser now support new variables "invRoC" and "w" of the input field
%  5) fieldAction has a protection for the case when there is no real closed loop
%
% April 21, 2017
%    setLaserMap is added to setup laserMap and variables for that.
%
% April 8, 2017
%    getField code is more strict now, almost the same, but ...
%
% March 20, 2017 - no change of code, but a new way of handling the modMap function
%    Function handle passed by setModMap can be a normal method, and extra variables can be shared
%    as normal properties of the derived FFTIFO class.
%    When a function handle is called in MirrorObj -> udateMirMap -> afterCare
%    control flow comes back to where setModMap is called, and the function handle is called accordingly,
%    either as static function without the obj argument or as a normal method with the obj as the first argument.
%
% March 12, 2017
%  In startFFT, all fieldList().EFFTs are set to zeroMat.
%  Some of the fields with status = -1, no source, cause problems when added to other valid field
%
% May 12, 2016
%  new function CNXY = HG1DFit( obj, fldID )
%  beam center offset and tilt are calculated by the ratio of TEM10/TEM01 and TEM00
%
% May 8, 2016
%  addLaser has new variables, 'nvec' (pointing vector) and 'c0' (beam center)
%
% April 19, 2016
%  new : HGPower, LGPower
%  power is changed to Power, constent with other naming
%
% July 1st, 2015
%  Port specification is always in one name, like ITM-HR-i, not a
%  combination of multiple variables, like 'ITM', 'HR', 'i' as separate variables
%
% April 17th, 2015
%  new : setInputType( obj, varargin ), which can be used to change the input field mode
%
% November 4th, 2014
%   updateNfft( obj, varargin ) is chanhged to accept varargin, and Nfft can be used to force Nfft value
%
% October 26, 2014
%   handle of FFTIFO itself is saved in all manager objects,
%   instead of handle of multiple manager handles
%
% October 23, 2014
%   DebugTK is added as obj.debug, and passed to cavities.siblings, which are passed to GuessObj
%
% October 15, 2014
%   roundtripLoss without arguments returns all round trip loss
%
% October 13, 2014
%   use addMirrorGeneric, instead of addMirror2x2 and addMirror4x4
%
% September 19, 2014
%   gaussFit can accpet xList and Eamp, in addition to the field ID
%
% August 29, 2014
%   function [Fld, xl, yl] = getField( fldID, Wout, Lprop )
%   getField has a new argument, Lprop.
%
% August 12, 2014
%   trans and refl add new fields, 'outField-trans' and 'outField-refl'
%   This was done sometime ago. Some analysis codes may need fix for this...
%
% August 5th, 2014
%   auto curvature correction option added to addMirror, removeCurvature.
%   after correction, 1/RoC -> 1/RoC + optDCurv
%
% July 31, 2014
%   initial guess field is default mode field
%
% July 29, 2014
%   lock and calc accepts third argument, which sets the convergence limit of FP arms
%
% July 23, 2014
%   roundtripLoss( fieldID ) did not pickup correct guessField, fixed
%
% July 10, 2014
%   HGFrac and LGFrac added, which are abs(XXCoef).^2/power
%   setRF( RF ) : unit of RF is Hz, instead of MHz
%
% April 3, 2014
%   Initial public release
%

% October 3, 2013
%   modifed as the base class
%
% September 11, 2013
%   initial implementation by Hiro Yamamoto  hiro@ligo.caltech.edu
%

classdef FFTIFO < handle
    
    properties
        ffts
        twids      % full matrix to calculate fields in IFO, E = M E + src
        optics
        props
        fields
        cavities
        debug
        
        savedFields  % (saveID, fieldID, EFFT )
    end
    
    % implementation of methods
    methods
        % constructors
        
        %% main IFO object
        function obj = FFTIFO( varargin )
            % initialize variables
            obj.ffts = FFTTK;
            obj.twids = twiddle( obj );
            obj.fields = fieldManager( obj );
            obj.optics = opticManager( obj );
            obj.props = propagaterManager( obj );
            obj.cavities = cavityManager( obj );
            obj.debug = DebugTK( obj );
            
            % setup IFO coniguration
            % This is a mandatory method to be prepared
            obj.defineIFO( varargin{:} );
            
            % find trees and cavities
            obj.analyzeCavity;
            
            % setup base mode. All ports and fields have q value assigned
            % default : use the FP arm, if exists, to define the entire mode
            obj.setupBaseMode( varargin{:} );
            
            % after modes are defined, setup physical dimension
            obj.updateNfft( varargin{:} );
            
            % now all fields have base mode parameters
            obj.setupMisc;
            
            % inject input field
            obj.startFFT;
            
            % calculate powers using scalar field analytic formula
            obj.twiddler;
            
            % calculate something
            obj.calcIt( varargin{:} );
            
            % (saveTag, fieldID, EFFT)
            obj.clearSavedFields;
        end
        
        %% main functions called in FFTIFO
        
        function analyzeCavity( obj, root )
            % if root is not defined, use the laser source as the root
            if ~exist('root', 'var')
                root = obj.optics.findLaser;
            end
            obj.cavities.analyzeCavity( root );
        end
        
        % use the FP arm, if exists, to define the entire mode
        function setupBaseMode( obj, varargin )
            % if source file is specified, use it
            [qInv, root] = obj.optics.getSourceFieldMode;
            
            if root == 0
                % find mode based using closed loop cavity, HR cavities first
                [qInv, root] = obj.cavities.findModeBaseByClosedCavs;
            end
            
            if root == 0
                % there is no closed loop
                disp('mode is not defined, and cannot proceed');
                exit;
            else
                %                fprintf(1, 'Cavity mode defined using portID %d with q = %g + i * %g\n', root, real(1/qInv), imag(1/qInv) );
                [Rinv, w] = MMTK.qinv2RinvW( qInv, obj.fields.kCR );
                fprintf(1, 'Cavity mode defined using field ''%s'' with w = %g, RoC = %g, z = %g, z0 = %g\n', ...
                    obj.fields.findName(root), w, 1/Rinv, real(1/qInv), imag(1/qInv));
                obj.cavities.fillQvals( root, qInv );
            end
        end
        
        % qInv from BS toward PRC and SRC usin X(N) arm or Y(W) arm
        % arm name can be LIGO like ('X' or 'Y' ) or Virgo like ('N', 'W')
        % If ITM and ETM names are different from these conventions, change ifs using strcmpi
        function [qInvP, qInvS] = fillQvalFromArm2BS( obj, armName )
            % default is 'X' arm
            ETM = 'ETMX';
            ITM = 'ITMX';
            B2I = 'BS-ARx';
            if exist('armName', 'var')
                if strcmpi(armName, 'Y')
                    ETM = 'ETMY';
                    ITM = 'ITMY';
                    B2I = 'BS-HRy';
                elseif strcmpi(armName, 'N')
                    ETM = 'NE';
                    ITM = 'NI';
                    B2I = 'BS-ARx';
                elseif strcmpi(armName, 'W')
                    ETM = 'WE';
                    ITM = 'WI';
                    B2I = 'BS-HRy';
                end
            end
            ETMHRoutID = obj.optics.findPortID( [ETM, '-HR-o'] );
            ITMHRoutID = obj.optics.findPortID( [ITM, '-HR-o'] );
            ITMHRinID = obj.optics.findPortID( [ITM, '-HR-i'] );
            ITMARoutID = obj.optics.findPortID( [ITM, '-AR-o']);
            
            qInv = obj.cavities.findCavityMode( [ETMHRoutID, ITMHRoutID] );
            obj.fields.setQInv([ETM, '-HR-o'], qInv);
            obj.fields.setQInv([ETM, '-HR-i'], qInv, true); % qInv is out of HR surface, so it is reversed
            qInv = obj.cavities.findCavityMode( [ITMHRoutID, ETMHRoutID] );
            obj.fields.setQInv([ITM, '-HR-o'], qInv);
            obj.fields.setQInv([ITM, '-HR-i'], qInv, true); % qInv is out of HR surface, so it is reversed
            
            % find the q value going out of the arm cavity
            % reverse the q value going into the ITM HR, then let it goes through ITM toward AR side
            qInv = MMTK.ABCD_qinv( obj.optics.findABCD( ITMHRinID, ITMARoutID ), MMTK.reverseDir( qInv ) );
            obj.fields.setQInv([ITM, '-AR-o'], qInv);
            obj.fields.setQInv([ITM, '-AR-i'], qInv, true);
            
            % ITM AR to BS
            qInv = MMTK.ABCD_qinv( obj.props.getABCD( obj.props.getPropID( ITMARoutID ) ), qInv );
            obj.fields.setQInv([B2I, '-i'], qInv);
            obj.fields.setQInv([B2I, '-o'], qInv, true);
            
            % qInv towards PRC
            qInvP = MMTK.ABCD_qinv( obj.optics.findABCD( obj.optics.findPortID( [B2I, '-i'] ), obj.optics.findPortID( 'BS-HRx-o' ) ), qInv );
            % qInv towards SRC
            qInvS = MMTK.ABCD_qinv( obj.optics.findABCD( obj.optics.findPortID( [B2I, '-i'] ), obj.optics.findPortID( 'BS-ARy-o' ) ), qInv );
        end
        
        function updateNfft( obj, varargin )
            % change Nfft based on optics and field parameters
            Nfft = obj.optics.bookKeeping;
            % [~,~, scrNfft] = getSourceFieldMode
            % scrNfft is > 0 when field in NxN is set as the source
            [~, ~, srcNfft] = obj.optics.getSourceFieldMode;
            if srcNfft > 0
                Nfft = srcNfft;
            end
            eval( UtilTK.setVars( {'Nfft'}, true, varargin{:} ) );
            Nlarge = 1024;
            if Nfft > Nlarge
                fprintf(1,'Nfft (%d) is larger than %d, and the simulation may be slow\n', ...
                    Nfft, Nlarge );
                if Nfft > Nlarge * 10
                    reply = input('Do you want to proceed? Y/N [N]:', 's');
                    if isempty(reply) || strcmpi( 'N', reply )
                        error('Nfft too large and quit');
                    end
                end
            end
            obj.ffts.setNfft( Nfft );
        end
        
        function setupMisc( obj )
            obj.props.setupMisc;
            obj.optics.setupMisc;
            obj.cavities.setupMisc;
        end
        
        function toggleLaser( obj, pwr )
            obj.fields.setLaserField
            if exist( 'pwr', 'var' ) && pwr < 1
                obj.fields.fieldList(obj.optics.findLaser).EFFT = obj.fields.fieldList(obj.optics.findLaser).EFFT * pwr;
            end
                
            % let the field touch with input optics
            obj.inject2IFO( true );
        end
        
        function startFFT( obj )
            % function startFFT
            % fill all EFFT by empty field
            % This is to make EFFT valid matix for those ports which do not have source, or status = -1
            for id = 1 : length(obj.fields.fieldList)
                obj.fields.fieldList(id).EFFT = obj.ffts.zeroMat;
            end
            
            % find the laser source, and set the input field
            obj.fields.setLaserField
                
            % let the field touch with input optics
            obj.inject2IFO( true );
        end
        
        % find alalytic solution
        function twiddler( obj, srcID )
            if ~exist( 'srcID', 'var' )
                srcID = obj.optics.findLaser;
            end
            obj.twids.buildMat;
            obj.twids.calcFields( srcID );
        end
        
        % do whatever you like, to be overriddedn
        function calcIt( obj, varargin )
            for it = 1 : length( obj.cavities.tree )
                obj.cavities.propFieldsOnPath( obj.cavities.tree{it} );
            end
        end
        
        %% access functions
        
        function [Fld, xl, yl] = getField( obj, fldID, Wout, Lprop )
            % [Fld, x, y] = getField( obj, fldID, Wout, Lprop )
            % return the 2D field map (Fld) and coordinates (x, y)
            % of the field fldID.
            % Fld(jy,ix) is the complex amplitude at x(ix), y(jy) in units
            % of sqrt(W)/m.
            % If Wout > 0 is specified, the field is calculated in (Wout x Wout) area
            % If Wout = 0, Wfft is set to be optimal based on the beam size
            % If Wout = -1, Wfft is set to be the same as the sink.
            % If Lprop is specified, the field is calculated Lprop from the HR side of the source.
            % If Lprop is specified and if fieldID is a source of a propagation,
            % it returns the same result when the fieldID is the sink of the propagation.
            % I.e., getField('X-HR-o', w, prop) = getField('Y-HR-i', w, prop ) when X-HR-o is connected to Y-HR-i.
            
            % convert name to ID
            fldID = obj.fields.findNamedField( fldID );
            
            if ~exist( 'Wout', 'var' )
                % No Wout meaans no Lprop, just return the stored field in the standard format
                Fld = obj.fields.conv2std( fldID );
                xl = obj.fields.getXlist(fldID);
                yl = xl;
            else
                % Lprop = nan is used to indicate Lprop is not specified
                if ~exist( 'Lprop', 'var' )
                    Lprop = nan;
                end
                [Fld, xl, yl] = obj.fields.getField( fldID, Wout, Lprop );
            end
        end
        
        function [dE, E0, xl, dP, L] = getPointScatter( obj, srcID, defects, WfftOut, L )
            %       function [dE, E0, xl, dP] = getPointScatter( obj, srcID, defects, WfftOut, L )
            %       get fields scattered by defects specified by (u0,v0,du,dv,rms),
            %       and the total loss by these defects
            %       help getPointScatter for more details
            if ~exist( 'L', 'var' )
                L = 0;
                if ~exist( 'WfftOut', 'var' )
                    WfftOut = 0;
                end
            end
            [dE, E0, xl, dP, L] = getPointScatter( obj.fields, srcID, defects, WfftOut, L );
        end
        
        % outputs a string of field IDs and names of the input field specification by fldIDs
        % fldIDs can be names wild card, like 'BS*-o' or cab be ids, like 10:20.
        function out = fieldNames( obj, fldIDs )
            fldIDs = obj.fields.conv2FieldIDList( fldIDs );
            out = '';
            for id = 1 : length( fldIDs )
                out = sprintf( '%s %d : %s\n', out, fldIDs(id), obj.fields.fieldList(fldIDs(id)).name );
            end
        end
        
        function pwr = power( obj, fldIDin )
            if ~exist( 'fldIDin', 'var' )
                fldIDin = 1:length(obj.fields.fieldList);
            end
            % convert name to ID
            fldIDin = obj.fields.conv2FieldIDList( fldIDin );
            pwr = zeros(1,length(fldIDin));
            for id = 1 : length(fldIDin)
                pwr(id) = obj.fields.fieldList(fldIDin(id)).calcPower;
            end
        end
        
        % power argument can be [ID1, ID2], {ID1, ID2}, {'name1', 'name2'}
        % function power = powerInR( obj, fldID, radius, x0, y0, Lprop )
        function power = powerInR( obj, fldID, radius, x0, y0, Lprop )
            fldID = obj.fields.findNamedField( fldID );
            if ~exist('radius', 'var')
                power = obj.power( fldID );
            else
                if ~exist( 'x0', 'var' )
                    x0 = 0;
                end
                if ~exist( 'y0', 'var' )
                    y0 = 0;
                end
                if ~exist('Lprop', 'var')
                    power = obj.fields.fieldList(fldID).calcSumIn( radius, x0, y0 );
                else
                    Wout = 2*(radius + max(abs(x0),abs(y0)));
                    [Fld, xl] = getField( obj, fldID, Wout, Lprop );
                    power = UtilTK.sumIn( xl, Fld, radius, radius, x0, y0 );
                end
            end
        end
        
        function [info, cwR] = gaussFit( obj, fldIDin, Eamp )
            % [info, cwR] = gaussFit( fldIDin/xList, E )
            % if there is one argument, it is assumed to be a list of field IDs or a field name
            % if there are two arcuments, first argument is the xList and the second argument is a complex field
            %
            % fit the field by gaussian(in x plane) * gaissian(in y plane)
            % info is for text display, so ifp.gaussFit('ITMX-HR-o') will display the fit
            % exp( -(x-x0)^2*(1/wx^2) + i/(2*Rx) )
            %   cwR(1, 2) for x and y
            %   cwR(:).c0 = center of gauss field
            %   cwR(:).w  = the beam width
            %   cwR(:).invRk = k / RoC
            
            if exist('Eamp', 'var')
                % exmplicit complex amplitude given
                % fldIDin is actually an array of coordinates for the field
                [cwR, info] = modeFitData( fldIDin, fldIDin, Eamp );
            else
                if ~exist('fldIDin', 'var')
                    fldIDin = 1:length(obj.fields.fieldList);
                end
                % if field is specied by name, convert it to ID number
                fldIDin = obj.fields.findNamedField( fldIDin );
                
                info = 'Gauss fit of fields\n';
                for id = 1 : length(fldIDin)
                    [infoOut, cwROut] = obj.fields.fieldList( fldIDin(id) ).gaussFit;
                    if length(fldIDin) == 1
                        info = infoOut;
                        cwR = cwROut;
                    else
                        info = sprintf( '%s\nField %2d) %s', info, fldIDin(id), infoOut );
                        cwR{id} = cwROut;
                    end
                end
            end
        end
        
        function CNXY = HG1DFit( obj, fldID )
            % beam center offset and tilt are calculated by the ratio of TEM10 and TEM00
            % offset = real(A10/A00) * w
            % tilt = -imag(A10/A00) 2/kw + d/R
            
            fldID = obj.fields.findNamedField( fldID );
            A00 = obj.HGCoef( fldID, 0, 0 );
            w = obj.fields.fieldList(fldID).w;
            invRoC = obj.fields.fieldList(fldID).invRoC;
            CNXY = zeros(2,2);
            for ixy = 1:2
                if ixy == 1
                    AXY = obj.HGCoef( fldID, 1, 0 );
                else
                    AXY = obj.HGCoef( fldID, 0, 1 );
                end
                AXY = AXY / A00;
                CNXY(1,ixy) = real(AXY) * w;
                CNXY(2,ixy) = -imag(AXY) * 2 / obj.fields.kCR / w + CNXY(1,ixy)*invRoC;
            end
        end
        
        % main mode contents
        function [modeStr, modes] = mainLGsStr( obj, fldID, range1, range2, threshold, varargin )
            if ~exist('range1', 'var') || isempty('range1')
                range1 = 0:10;
            end
            if ~exist('range2', 'var') || isempty('range2')
                range2 = -2*max(range1):2*max(range1);
            end
            if ~exist('threshold', 'var')
                threshold = -1;
            end
            modes = obj.mainModes( fldID, 'LG', range1, range2, threshold, varargin{:} );
            modeStr = obj.mainModesStr( modes );
        end
        
        function modes = mainLGs( obj, fldID, range1, range2, threshold, varargin )
            if ~exist('range1', 'var') || isempty('range1')
                range1 = 0:10;
            end
            if ~exist('range2', 'var') || isempty('range2')
                range2 = -2*max(range1):2*max(range1);
            end
            if ~exist('threshold', 'var')
                threshold = -1;
            end
            modes = obj.mainModes( fldID, 'LG', range1, range2, threshold, varargin{:} );
        end
        
        function [modeStr, modes] = mainHGsStr( obj, fldID, range1, range2, threshold, varargin )
            if ~exist('range1', 'var') || isempty('range1')
                range1 = 0:20;
            end
            if ~exist('range2', 'var') || isempty('range2')
                range2 = 0:max(range1);
            end
            if ~exist('threshold', 'var')
                threshold = -1;
            end
            modes = obj.mainModes( fldID, 'HG', range1, range2, threshold, varargin{:} );
            modeStr = obj.mainModesStr( modes );
        end
        
        function modes = mainHGs( obj, fldID, range1, range2, threshold, varargin )
            if ~exist('range1', 'var') || isempty('range1')
                range1 = 0:20;
            end
            if ~exist('range2', 'var') || isempty('range2')
                range2 = 0:max(range1);
            end
            if ~exist('threshold', 'var')
                threshold = -1; % use default
            end
            modes = obj.mainModes( fldID, 'HG', range1, range2, threshold, varargin{:} );
        end
        
        % modes = mainModes( obj, fldID, fieldMode, range1, range2, threshold )
        %      modes(index,1:2) = mode index (m,n, or p,m)
        %      modes(index,3) = amplitude
        %      modes(index,4) = power fraction
        %      modes(index,5) = fractional loss of this mode
        %      modes(index,6) = amplitude of this mode at the source
        %      modes(index,7) = fractional loss of this field (ind,5)*(ind,6)^2/fldPower
        % threshold < 1  : returns modes whose fraction is larger than threshold
        % threshold = 0  : all
        % threshold >= 1 : returns modes whose fractions is among the top N = int(threshold)
        % threshold < 0  : max(ceil(length(range1) * length(range2) / 20), 10);
        function modes = mainModes( obj, fldID, fieldMode, range1, range2, threshold, varargin )
            fldID = obj.fields.findNamedField( fldID );
            if ~exist('threshold', 'var')
                threshold = 5;
            else
                if threshold < 0
                    threshold = max(ceil(length(range1) * length(range2) / 20), 10);
                end
            end
            if isempty( varargin )
                fldPower = obj.power( fldID );
            else
                [~, ~, Fld] = obj.fields.fieldList(fldID).varargin4ModeCoef( varargin{:} );
                fldPower = obj.fields.fieldList(fldID).calcPower( Fld );
            end
            Cmns = zeros(length(range1)*length(range2), 3);
            ind = 0;
            for ind1 = 1:length(range1)
                for ind2 = 1:length(range2)
                    ind = ind + 1;
                    Cmns(ind, 1:3) = [range1(ind1), range2(ind2), obj.modeCoef( fldID, fieldMode, range1(ind1), range2(ind2), varargin{:})];
                end
            end
            [powerMode, ind] = sort( abs(Cmns(:,3)).^2 );
            powerMode = flip(powerMode);
            ind = flip(ind);
            if threshold < 1
                ind = ind( powerMode > fldPower*threshold  );
            else
                ind = ind( 1: min(floor(threshold+0.1), length(ind)) );
            end
            modes = zeros(length(ind),7);
            for id = 1:length(ind)
                modes(id,1) = Cmns(ind(id),1);
                modes(id,2) = Cmns(ind(id),2);
                modes(id,3) = Cmns(ind(id),3);
                modes(id,4) = abs(Cmns(ind(id),3)).^2 / fldPower;
                [modes(id,5), modes(id,6)] = obj.fields.lossOutMode(fldID,fieldMode,Cmns(ind(id),1),Cmns(ind(id),2));
                modes(id,7) = modes(id,5) * abs(modes(id,6))^2 / fldPower;
            end
        end
        
        % mode coefficient using tilted mode axis.
        % this works for limited case, like tilted beam goes through the center of ITM
        % replace this by modeCoef( ,,, dthetaX, dthetaY )
        
        %{     
        function Cmn = modeCoefTilt( obj, fldID, fieldMode, ind1, ind2, dthetaX, dthetaY )
            fldID = obj.fields.findNamedField( fldID );
            xl = obj.fields.getXlist(fldID);
            [xg, yg] = meshgrid( xl, xl );
            EFFT = obj.fields.fieldList(fldID).EFFT;
            obj.fields.fieldList(fldID).EFFT = EFFT .* exp( 1i*( dthetaX * xg + dthetaY * yg ) * obj.fields.kCR);
            Cmn = obj.fields.modeCoef( fldID, fieldMode, ind1, ind2 );
            obj.fields.fieldList(fldID).EFFT = EFFT;
        end
        %}
        
        % LGCoef and HGCoef combined. If fieldMode is not specified, laser mode is used
        % varargin = (thetaX, thetaY, EFFT), which is actually tested and used in fields.modeCoef
        function Cmn = modeCoef( obj, fldIDs, fieldMode, ind1, ind2, varargin )
            if ~exist( 'fieldMode', 'var' )
                [fieldMode, ind1, ind2] = obj.optics.getInputType;
            end
            Cmn = obj.fields.modeCoef( fldIDs, fieldMode, ind1, ind2, varargin{:} );
        end
        
        % varargin = (thetaX, thetaY, EFFT), which is actually tested and used in fields.modeCoef
        function Cmn = HGCoef( obj, fldIDs, m, n, varargin )
            % Cmn = HGCoef( obj, fldID, m, n, invRoC, w )
            % calculare the coefficients of HG(m,n) terms
            
            % convert name to ID
            Cmn = obj.fields.modeCoef( fldIDs, 'HG', m, n, varargin{:} );
        end
        
        function Pmn = HGPower( obj, fldID, m, n, varargin )
            % Pmn = HGPower( obj, fldID, m, n )
            % Power of HG(m,n) terms
            % if m < 0, it returns array of powers
            %    m = -nmmax, n = 0
            %       sum of power( m+n == nmmax )
            %    m = -nmmax, n = 1
            %       power of elements of n=0 case
            %    m = -nmmax, n = -1
            %       array of powers with n+m <= nmmax
            if m >= 0
                Pmn = abs(obj.HGCoef(fldID, m, n, varargin{:})).^2;
            else
                if n >= 0
                    Cmn = obj.fields.HGCoef( fldID, m, 0, varargin{:} ); % m-input < 0 => converted to array of m+n = abs(m-input)
                    Pmn = abs(Cmn).^2;
                    if n == 0
                        Pmn = sum( Pmn );
                    end
                else
                    for imn = 0 : abs(m)
                        Pmn(imn+1) = sum( abs(obj.fields.HGCoef( fldID, -imn, 0, varargin{:} )).^2 );
                    end
                    if n == -1
                        Pmn( abs(m) + 2 ) = obj.power( fldID ) - sum( Pmn );
                    end
                end
            end
        end
        
        function Pmn = HGFrac( obj, fldID, m, n, varargin )
            % Pmn = HGFrac( obj, fldID, m, n )
            % Power fraction of HG(m,n) terms
            
            Pmn = obj.HGPower( fldID, m, n, varargin{:} ) / obj.power(fldID);
        end
        
        function Cpm = LGCoef( obj, fldIDs, p, m, varargin )
            % Cpm = LGCoef( fldID, p, m )
            % calculare the coefficients of LG(p,m) terms
            
            % convert name to ID
            Cpm = obj.fields.modeCoef( fldIDs, 'LG', p, m, varargin{:} );
        end
        
        function Ppm = LGPower( obj, fldID, p, m, varargin )
            % Ppm = LGPower( obj, fldID, p, m )
            % Power of LG(p,m) terms
            % if p < 0, it returns array of powers
            %     m =  1 : power of elements for |2p+m| : 2p+m=-2 then, LG(1,0), LG(0,2), LG(0,-2)
            %     m =  0 : sum of power of 2p+m : LG(1,0) + LG(0,2) + LG(0,-2)
            %     m = -1 : array of powers with 2p+m <= nmmax, sum(LG(2p+m=0)), sum(LG(2p+m=1))..., sum(LG(2p+m=nmax)) and power with 2p+m > nmax
            if p >= 0
                Ppm = abs(obj.LGCoef(fldID, p, m, varargin{:} )).^2;
            else
                if m >= 0
                    Cmn = obj.fields.LGCoef( fldID, p, 0, varargin{:} ); % p-input < 0 => converted to array of 2p+m = abs(p-input)
                    Ppm = abs(Cmn).^2;
                    if m == 0
                        Ppm = sum( Ppm );
                    end
                else
                    for imn = 0 : abs(p)
                        Ppm(imn+1) = sum( abs(obj.fields.LGCoef( fldID, -imn, 0, varargin{:} )).^2 );
                    end
                    if m == -1
                        Ppm( abs(p) + 2 ) = obj.power( fldID ) - sum( Ppm );
                    end
                end
            end
        end
        
        function Ppm = LGFrac( obj, fldID, p, m, varargin )
            % Cpm = LGFrac( fldID, p, m )
            % calculare the fraction of LG power
            
            Ppm = obj.LGPower( fldID, p, m, varargin{:} ) / obj.power(fldID);
        end
        
        function [E, x] = Non00Amp( obj, fldID, useFit, Wind, thetaX, thetaY )
            % returns the minor components, i.e., field - gaussFit or field - LGpm/HGmn (type specified at addLaser)
            fldID = obj.fields.findNamedField( fldID );
            if exist('Wind', 'var')
                [E, x] = obj.getField( fldID, Wind );
                useFit = true;
            else
                [E, x] = obj.getField( fldID );
            end
            if ~exist( 'useFit', 'var' )
                useFit = false;
            end
            if ~exist('thetaX','var')
                thetaX = 0;
            end
            if ~exist('thetaY', 'var')
                thetaY = 0;
            end
            E = obj.fields.fieldList(fldID).tiltField( thetaX, thetaY, E);
            if useFit
                [~, ~, Efit] = modeFitData( x, x, E );
                EEfit = sum(sum(E .* conj(Efit)));
                C = complex( real(EEfit), imag(EEfit) ) / sum(sum(abs(Efit).^2));
            else
                % amplitude of the field type specified at addLaser, i.e., the main mode
                [Efit, ~, C] = obj.fields.fieldAtFID(fldID);
            end
            E = E - C * Efit;
        end
        
        function loss = roundtripLoss( obj, fldID )
            % if nothing is specified, return all losses
            if nargin == 1
                loss = zeros( 1, length( obj.cavities.guessList )-1 );
                for ig = 1 : length( obj.cavities.guessList )-1
                    loss(ig) = obj.cavities.guessList(ig).roundtripLoss;
                end
                return
            end
            
            % loss = roundtripLoss( fldID )
            % calculate rount trip loss of a cavity which has fieldID as a part of it
            fldID = obj.fields.findNamedField( fldID );
            loss = NaN;
            for ig = 1 : length( obj.cavities.guessList )-1
                if ~isempty( find( obj.cavities.guessList(ig).pathList{1} == fldID, 1) )
                    loss = obj.cavities.guessList(ig).roundtripLoss;
                    return;
                end
            end
        end
        
        function reflFldID = refl( obj, inFldID, inField )
            % function reflFldID = refl( obj, inFldID, inField )
            % create a new field, which is induced by the reflection of the inField
            % if inField is specified, this amplitude (standard format) is reflected
            
            % incoming field
            inFldID = obj.fields.findNamedField( inFldID );
            % find the port where the reflected field is going out
            % this can have a sum of reflection and transmission
            [outFldID, optID] = obj.optics.findReflPort( inFldID );
            % create a new field which contains reflection componenent only
            % new name is old name + '-refl'
            name = strcat( obj.fields.fieldList(outFldID).name, '-refl');
            reflFldID = obj.fields.newField( name, outFldID );
            % calculate the reflection only field and store in reflFldID
            if exist('inField','var')
                % if inField, standard format, is specified, reflect that field
                obj.fields.fieldList(reflFldID).EFFT = ...
                    obj.optics.refl( inFldID, obj.fields.fieldList(inFldID).conv2int( inField ) );
            else
                % if not, use the field amplitude in inFldID
                obj.fields.fieldList(reflFldID).EFFT = obj.optics.refl( inFldID );
            end
            if obj.optics.isOnARSidePortID( inFldID )
                r = obj.optics.optList(optID).rAR;
            else
                r = obj.optics.optList(optID).rHR;
            end
            obj.fields.fieldList(reflFldID).Eamp = r * obj.fields.fieldList(inFldID).Eamp;
            % set the flag
            obj.fields.fieldList(reflFldID).status = 2048;
            obj.fields.fieldList(reflFldID).portIDfrom = outFldID;
        end
        
        function transFldID = trans( obj, inFldID, inField )
            % function transFldID = trans( obj, inFldID )
            % create a new field, which is induced by the transmission of the inField
            
            % incoming field
            inFldID = obj.fields.findNamedField( inFldID );
            % find the port where the transmitted field is going out
            % this can have a sum of reflection and transmission
            [outFldID, optID] = obj.optics.findTransPort( inFldID );
            % create a new field which contains reflection componenent only
            % new name is old name + '-trans'
            name = strcat( obj.fields.fieldList(outFldID).name, '-trans');
            transFldID = obj.fields.newField( name, outFldID );
            % calculate the reflection only field and store in reflFldID
            if exist('inField','var')
                obj.fields.fieldList(transFldID).EFFT = ...
                    obj.optics.trans( inFldID, obj.fields.fieldList(inFldID).conv2int( inField ) );
            else
                obj.fields.fieldList(transFldID).EFFT = obj.optics.trans( inFldID );
            end
            obj.fields.fieldList(transFldID).Eamp = ...
                obj.optics.optList(optID).t * obj.fields.fieldList(inFldID).Eamp;
            % set the flag
            obj.fields.fieldList(transFldID).status = 2048;
            obj.fields.fieldList(transFldID).portIDfrom = outFldID;
        end
        
        function resetStatus( obj, fldID, doInjection )
            % no more needed now that calRefl and calTrans add new fields, instead of override
            if ~exist( 'doInjection', 'var' )
                doInjection = false;
            end
            
            if exist('fldID', 'var')
                % field is specified, so reste only that
                fldID = obj.fields.findNamedField( fldID );
                obj.fields.fieldList(fldID).status = 0;
            else
                % reset status of all external fields
                obj.cavities.resetFieldStatus;
                obj.cavities.updateFieldStatus;
                obj.inject2IFO(doInjection);
            end
        end
        
        function [L0, delL] = getPropLeng( obj, propID )
            % [L0, delL] = getPropLeng( obj, propID )
            % cavity length
            [L0, delL] = obj.props.getPropLeng( propID );
        end
        
        %% DemodRF
        %{
             front end of DemodRF function
        %}
        function [ Demod1, Demod2, PRG, Es ] = DemodRF( obj, fldIDs, RF, modIndex, errorCR, errorSB )
            
            if ~exist('gamma', 'var')
                modIndex = 1;
            end
            if ~exist('errorCR', 'var')
                errorCR = 1e-6;
            end
            if ~exist('errorSB', 'var')
                errorSB = 1e-6;
            end
            [ Demod1, Demod2, PRG, Es ] = DemodRFFunc( obj, fldIDs, RF, modIndex, errorCR, errorSB );
        end
        
        %% add optics
        
        % add laser
        function id = addLaser( obj, name, varargin )
            % id = addLaser( 'name', varargin ) for LGpm or HGmn injection
            % add laser source of HG(ind1,ind2) or LG(ind1,ind2)
            %
            % variables names   : help
            % ------------------ ----------------------------------------------------------------
            % power             : laser power [1]
            % lamCR             : CR laser wavelength [1.064e-6]
            % invRoC            : [0]
            % w                 : [inf]
            % fieldMode         : 'HG' or HermiteGauss and 'LG' for Laguerre Gauss ['LG']
            % ind1              : first index for HG(ind1,ind2) and LG(ind1,ind2) [0]
            % ind2              : second index for HG(ind1,ind2) and LG(ind1,ind2) [0]
            % nvec              : [nx, ny] pointing vector of the input beam [ [0,0] ]
            % dxyz              : [cx, cy, cz] center of the gaussian beam [ [0,0, 0] ]
            %                     waist position is shifted by cz (q => q-z)
            % ------------------ ----------------------------------------------------------------
            id = obj.optics.addLaser( name, varargin{:} );
        end
        
        % add field source from other IFO object
        function id = addFieldSrc( obj, srcName, srcIFO, srcFieldID )
            srcFieldID = srcIFO.fields.findNamedField( srcFieldID );
            if isempty( srcName )
                srcName = strcat( 'src_', srcIFO.fields.findName( srcFieldID ) );
            end
            id = obj.optics.addLaser( srcName );
            obj.optics.optList(id).setLaserSrcInfo( srcIFO, srcFieldID );
        end
        
        % add detector
        function id = addDetector( obj, name )
            id = obj.optics.addDetector( name );
        end
        
        function id = addMirrorGeneric( obj, MirrorObjName, name, varargin )
            % function id = addMirrorGeneric( obj, MirrorObjH, name, varargin )
            % add a mirror object of class MirrorObjName
            if exist(MirrorObjName, 'class') == 8
                % minimal check if the MirrorObjName is a valid class name
                id = obj.optics.addMirrorGeneric( MirrorObjName, name, varargin{:} );
            else
                error('''%s'' is not a valid class name\n', MirrorObjName );
            end
        end
        
        function id = addMirror( obj, name, varargin )
            % id = addMirror( 'name', 'variableName1', variableValue1, 'variableName2', variableValue2, ... )
            %
            % Add a mirror and return the ID of the mirror.
            % First argument must be the name of the mirror.
            % If variable incAngle is specified, 4x4 mirror (two input fields and two out going fields
            % on each side of the mirror) is added. Otherwise, 2x2 mirror (one input and one output on
            % each side) is added.
            % BS, folding mirrors and MC mirrors are 4x4 mirrors even if angle can be approximated to be 0.
            %
            % variables names   : help
            % ------------------ ----------------------------------------------------------------
            % invRoC            : inverse of the HR side radius of curvature. + for concave [0/m]
            % invRoCAR          : inverse of the AR side radius of curvature. + for concave [0/m]
            % Aperture          : physical size of the mirror [-8 : 8 x beam size]
            % Thickness         : thickness of the substrate [0m]
            % T                 : power transmisstance [0]
            % Lhr               : power loss on the HR side (Reflectio on HR = 1-T-Lhr) [0]
            % Lar               : power loss on the AR side (Reflectio on HR = 1-T-Lar) [0]
            % resolution        : minimum spatial resolution of the mirror
            % optTheta          : Together with optPhi, the mirror tilt is defined.
            % optPhi            : d = cos(phi)*sin(theta)*x + sin(phi)*cos(theta)*y
            %                   :    pitch : phi=pi/2, theta = pitch,  yaw : phi=0, theta = yaw
            % optX0             : mirror is moved horizontally by this amount
            % optY0             : mirror is moved vertically by this amount
            % removeTilt        : remove tilt using gaussian field (for Mirror2x2)
            % removeCurvature   : remove curvature mismatch using gaussian field (for Mirror2x2), like pseudo RH
            %                       1/RoC -> 1/RoC + mirror.optDCurv
            % incAngle          : incident angle of the field (for BS)
            %
            % nRef              : refractive index (def:nRef(SiO2))
            % isShperical       : 1/-1 for spherical with/without Hello-Vinet off axis correction, 0 parabolic [BAD], [1]
            % invFocal          : 1/focal length in the mirror. If maps are defined, it overrides.
            % WfftUser          : If > 0, Wfft for this mirror. If < 0, abs(WfftUser)*beam size is Wfft [0]
            % virtualAperture   : extra outer space in which field is to be calculated [0]
            % ------------------ ----------------------------------------------------------------
            
            if ~ischar( name )
                error('First argument must be the name of the mirror');
            end
            
            % check if the argument format is OK
            if ~UtilTK.validArgList( varargin{:} )
                error('argument needs to be a list of ''name'' and value pairs');
            end
            
            % try to add a mirror2x2
            %            id = obj.optics.addMirror2x2( name, varargin{:} );
            id = obj.optics.addMirrorGeneric( 'Mirror2x2Obj', name, varargin{:} );
            if id == 0
                % if failed, add a mirror4x4
                %                id = obj.optics.addMirror4x4( name, varargin{:} );
                id = obj.optics.addMirrorGeneric( 'Mirror4x4Obj', name, varargin{:} );
                
                % id is 0 if this failed
            end
            if id == 0
                error('variable names incorrect for a new mirror specification');
            end
        end
        
        function tiltOn( obj, mirrorID, thetaX, thetaY )
            mID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mID).tiltX = thetaX;
            obj.optics.optList(mID).tiltY = thetaY;
        end
        
        function setHRfiles( obj, mirrorID, varargin )
            % set details of HR surface
            %   use setTRfiles to define the transmission map and setARfiles to define the AR side reflection
            %
            %   setHRfiles( 'mirrorName', 'dataFile', 'formula',
            %               aperture of power fitting (0 for not specifying)
            %               RoC of the power fitting (set to 0 for no fitting),
            %               orientation of the optic (0:no orientation, 1:90 degree, etc),
            %               'variableName1 in formula', initial value of variable1, 'vN2', v2, ... );
            %   example formula : map + tilt*x will add tilt to the data file surface
            %                     map + (x.^2+y.^2)/2*invRoC will add surface curvature change
            %   dataFile can be empty and all can be defined by formula, like loading COMSOL data
            %
            %   use updateMirMaps( 'mirrorName', 'varName1', value1, ... ) to change the formula later
            %
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).HRfiles.addMap( varargin{:} );
        end
        function setTRfiles( obj, mirrorID, varargin )
            % set details of transmission maps
            % see help for setHRfiles
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).TRfiles.addMap( varargin{:} );
        end
        function setARfiles( obj, mirrorID, varargin )
            % set details of AR side reflection maps
            % see help for setHRfiles
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).ARfiles.addMap( varargin{:} );
        end
        
        function setRPYfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).RPYfiles.addMap( varargin{:} );
        end
        function setRSXfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).RSXfiles.addMap( varargin{:} );
        end
        function setTPXfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).TPXfiles.addMap( varargin{:} );
        end
        function setTSYfiles( obj, BSID, varargin )
            BSID = obj.optics.findNamedOptic(BSID);
            obj.optics.optList(BSID).TSYfiles.addMap( varargin{:} );
        end
        
        function updateMirMaps( obj, mirrorID, varargin )
            % updateMirMaps( obj, mirrorID, varargin )
            %
            % change the variables used in the formulas speficying HRfiles, TRfiles and ARfiles
            % E.g., for a formula, map + (x.^ + y.^2)/(2*RoC) for 'ITM',
            %       ifo.updateMirMaps( 'ITM', 'RoC', 100e3 )
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).updateMirMaps( varargin{:} );
        end
        
        %
        % setModMap( obj, mirrorID, modMapH, modMapVar )
        % This set a function handle, which is called after each mirror map is created as
        % [obj.reflFFT, obj.transFFT] = obj.modMap( obj.reflFFT, obj.transFFT, x, y, obj.modMapVar );
        % modMapVal is passed as is, and one way to to pass a string like
        %    var = sprintf('BFdx = %g; BFdy = %g;', BFdx, BFdy);
        % and eval in each function called.
        %
        % Function passed by handle can be a normal method, and these extra variables can be shared
        % as normal properties of the derived FFTIFO class.
        %
        function setModMap( obj, mirrorID, modMapH, varargin )
            mirrorID = obj.optics.findNamedOptic(mirrorID);
            obj.optics.optList(mirrorID).modMap = modMapH;
            if exist('varargin', 'var' )
                obj.optics.optList(mirrorID).modMapVar = varargin;
            end
        end
        
        % setup laserMap, which is called to setup laser field
        function setLaserMap( obj, laserMapH, varargin )
            [~,laserOpt] = obj.optics.findLaser;
            obj.optics.optList(laserOpt).laserMap = laserMapH;
            if exist('varargin', 'var' ) && ~isempty( varargin )
                obj.optics.optList(laserOpt).laserMapVar = varargin;
            end
        end
        
        % add propagator
        function id = addOneProp( obj, varargin )
            % id = addOneProp( obj, varargin )
            %   Same syntax as addProp, but adds one way propagator
            id = obj.props.addOneProp( varargin{:} );
        end
        
        function [idf, idb] = addProp( obj, varargin )
            % [idf, idb] = addProp( 'name', source, sink, macroLength, microLength )
            %   Add two proagators, from source to sink and sink to source
            %   macroLength is rounded to the multiple of wave length
            %   source and sink specification can be
            %       one string specifying optic and port, like 'ITM-HR'
            %   example : addProp( 'FPprop', 'ITM-HR', 'ETM-HR', 3994.5);
            [idf, idb] = obj.props.addProp( varargin{:} );
        end
        
        %% Field quantities
        function setRF( obj, RF )
            % setRF( RF )
            % set RF sideband frequency given in Hz
            obj.fields.kSB = 2*pi*RF/ConstantTK.c;
        end
        
        function RF = getRF( obj )
            RF = obj.fields.kSB * ConstantTK.c / (2*pi);
        end
        
        function setInputType( obj, varargin )
            % setInputType( obj, varargin )
            % set the input laser mode
            [fieldMode, ind1, ind2, dxyz, nvec] = obj.optics.getInputType;
            eval( UtilTK.setVars( who, false, varargin{:} ) );
            
            obj.optics.setInputType( fieldMode, ind1, ind2, dxyz, nvec );
            obj.cavities.clearAllTemp;
            obj.startFFT;
        end
        
        % change the laser beam center (shift) and the pointing vector (tilt)
        % to make the input beam matches with the resonating field
        % It is assumed that the resonating field has been calculated
        % if adjust = 't', only tilt is adjusted, if adjust = 's', only shift is adjusted.
        % if adjust is empty, both are cleared
        %
        % if adjust = [x0, thetax0; y0, thetay0], it is used as initial direction of the laser in the cavity
        function cnout = adjustLaser( obj, adjust, cnoutIn )
            % laser field ID
            srcID = obj.optics.findLaser;
            % propagator ID and sink port (e.g., AR side of RM or ITM)
            [sinkID, propID] = obj.props.getPropTo( srcID, true );                                    
            if sinkID > 0
                % sink, AR side of RM/ITM
                % find field transmitted
                transID = obj.optics.findTransPort( sinkID );
                cnout = zeros(2,2);
                if transID > 0
                    if exist( 'cnoutIn', 'var' )
                        cnout = cnoutIn;
                    else
                        % transID is HR side out going field of RM/ITM
                        if isnumeric( adjust ) & (size(adjust) == [2,2])
                            CNXY = adjust;
                        else
                            CNXY = obj.HG1DFit( transID );
                        end
                        ABCD = obj.optics.findTransABCD( sinkID ) * obj.props.propList(propID).propABCD;
                        for ixy = 1 : 2
                            cnout(1:2,ixy) = ABCD \ CNXY(1:2,ixy);
                        end
                    end
                    if exist('adjust', 'var') && ischar( adjust )
                        if isempty( adjust )
                            cnout(1:2,1:2) = 0;
                        elseif lower(adjust) == 't'
                            % only tilt
                            cnout(1,1:2) = 0;
                        elseif lower(adjust) == 's'
                            % only shift
                            cnout(2,1:2) = 0;
                        end
                    end
                    % dxyz(3) is 0
                    obj.setInputType( 'dxyz', [cnout(1,1), cnout(1,2), 0], 'nvec', [cnout(2,1), cnout(2,2)] );
                end
            end
        end
        
        %% internal support functions
        
        % arguments for calc and lock
        %
        % varargin{1} : errorLimit [1e-5]
        % varargin{2} : Nprint [1]
        % varargin{3} : FPErrorLim
        % varargin{4} : reset
        
        function evol = calc( obj, varargin )
            % evol = calc( errorLimit, Nprint, FPErrorLim )
            if nargout > 0
                evol = obj.fieldAction( false, varargin{:} );
            else
                obj.fieldAction( false, varargin{:} );
            end
            obj.fields.calcIfNecessary;
        end
        
        function evol = lock( obj, varargin )
            % evol = lock( errorLimit, Nprint, FPErrorLim )
            if UtilTK.getVar( 'reset', true, varargin{:} ) == true
                obj.resetStatus(true);
                obj.props.resetDelL;
            end
            
            if nargout > 0
                evol = obj.fieldAction( true, varargin{:} );
            else
                obj.fieldAction( true, varargin{:} );
            end
            obj.fields.calcIfNecessary;
        end
        
        % field calculation code, with or without lock
        function evol = fieldAction( obj, doLock, varargin )
            evol = zeros(2,0);
            obj.props.propCounter = 0;
            
            % default values
            errorLimit = 0;    % amplitude variation is used to check the convergence
            Nprint = 1;        % print status report in every Nprint
            FPErrorLim = 0;    % FP cavity lock
            reset = 1;
            
            eval( UtilTK.setVars( who, true, varargin{:} ) );
            
            % mark that none has been calculated
            obj.cavities.resetFieldStatus;
            
            % this updates field status of laser
            obj.inject2IFO(reset);
            
            % guessList includes one dummy loop, and
            % do something only when real loop exists
            if obj.cavities.guessList(1).guessID > 0
                if errorLimit <= 0
                    errorLimit = 1e-5;
                end
                % loop case
                if doLock
                    obj.cavities.freezeLength( false );
                    % lock when needed
                    for ig = 1 : length( obj.cavities.guessList ) - 1
                        obj.cavities.guessList(ig).status = -1;
                    end
                else
                    obj.cavities.freezeLength( true );
                    % don't change
                    for ig = 1 : length( obj.cavities.guessList ) - 1
                        obj.cavities.guessList(ig).status = 0;
                    end
                end
                
                % if reset, reset all guessFields and pump fields
                if reset
                    for ig = 1 : length( obj.cavities.guessList ) - 1
                        guessField = obj.cavities.guessList(ig).guessID;
                        obj.fields.setDefField( guessField, 1 );
                        obj.fields.fieldList(guessField).EFFT = ...
                            obj.fields.fieldList(guessField).EFFT * obj.fields.fieldList(guessField).Eamp;
                        
                        % this should be redundant
                        obj.cavities.guessList(ig).initGuessField( obj.fields.fieldList(guessField).EFFT );
                        
                        obj.cavities.guessList(ig).fillPumpOuts;
                        
                        % a loop which has only one input, which is the end arm
                        if FPErrorLim > 0
                            if size(obj.cavities.guessList(ig).pumpIn,2) == 1
                                obj.cavities.guessList(ig).errorLimit = FPErrorLim;
                            end
                        end
                    end
                end
                
                %                errMax = 10;
                TSTART = tic;
                for iloop = 1 : 1000
                    err = 0;
                    PowerOld = obj.fields.fieldList( obj.cavities.guessList(1).guessID ).calcPower;
                    for ig = 1 : length( obj.cavities.guessList )-1
                        
                        if ig == 1 || ~obj.cavities.guessList(ig).onTheFly
                            err = max( err, obj.cavities.guessList(ig).action(errorLimit, Nprint) );
                        end
                    end
                    PowerNew = obj.fields.fieldList( obj.cavities.guessList(1).guessID ).calcPower;
                    dark = obj.lockDone;
                    if dark
                        dark = false;
                        if abs(PowerNew-PowerOld) < 0.001*PowerNew
                            if err < errorLimit
                                dark = true;
                            end
                            %{
                            if errMax <= errorLimit
                                % if errMax (error limit for each field finder) is the user setting
                                if err <= errMax
                                    % if the actual error estmation is less than requirement, done
                                    % otherwise, keep the loop using the same condition
                                    dark = true;
                                end
                            else
                                if errMax > 1
                                    % if errMax (err for each field finder) is large, make it 0.1
                                    errMax = 0.01;
                                else
                                    % otherwise, make it 1/100
                                    errMax = errMax / 100;
                                end
                            end
                            %}
                            %                         if err <= errorLimit
                            %                             % if the actual error estmation is less than requirement, done
                            %                             % otherwise, keep the loop using the same condition
                            %                             done = true;
                            %                         end
                        end
                    end
                    evol(1:2,size(evol,2)+1) = [ obj.props.propCounter; err ];
                    if dark || toc(TSTART) >= 10 || mod(iloop,10) == 0
                        % if more than 5 secods has elapsed or 10 loops
                        if doLock
                            doing = 'full lock';
                        else
                            doing = 'stationary';
                        end
                        fprintf(1, 'Main %s loop %d ... : err = %g, FFT count = %g, power changed from %g to %g\n', ...
                            doing, iloop, err, obj.props.propCounter, PowerOld, PowerNew );
                        TSTART = tic;
                    end
                    
                    if dark
                        break;
                    end
                end
            else
                % no cavity
                % use 1ppm for setting CD limit in Michelson case
                if errorLimit <= 0
                    errorLimit = 1e-6;
                end
                
                if doLock
                    for iloop = 1 : 1000
                        % make it dark
                        [dark, CD] = obj.darkPort( errorLimit );
                        evol(1:2,size(evol,2)+1) = [ obj.props.propCounter; CD ];
                        obj.cavities.resetFieldStatus;
                        obj.inject2IFO(true);
                        if dark ~= 0
                            break;
                        end
                    end
                end
            end
            
            obj.cavities.updateFieldStatus;
            
        end
        
        % judge is any criteria, except for loop phase, are OK or not, e.g., dark port is dark
        function done = lockDone( obj )
            if obj.cavities.freezeLength
                done = true;
            else
                done = obj.darkPort ~= 0;
            end
        end
        
        % dark > 0 when dark port is good dark
        % dark = 0 when dark is not good dark
        % dark < 0 when BS does not exist
        function [dark, CD, delL] = darkPort( obj, CDlimit, delLimit )
            
            dark = -1;  % default is no BS
            BSID = obj.optics.findNamedOptic( 'BS' );
            if BSID == 0
                % if there is no BS, that's it
                return;
            end
            toBSX = obj.optics.findPortID( BSID, -3 );
            toBSY = obj.optics.findPortID( BSID, -2 );
            if obj.props.getPropFrom( toBSX ) < 0 || obj.props.getPropFrom( toBSY ) < 0
                % if either port is not connected, that's it
                return;
            end
            
            % if there is no resonator, this might not be filled yet 2017-09-26
            obj.fields.calcIfNecessary(toBSX);
            obj.fields.calcIfNecessary(toBSY);
            
            if ~exist( 'CDlimit', 'var' )
                CDlimit = 10e-6;
            end
            if ~exist( 'delLimit', 'var' )
                delLimit = 1e-12;
            end
            
            REX = obj.optics.refl( toBSX );
            TEY = obj.optics.trans( toBSY );
            PREX = sum(sum( REX .* conj(REX) ));
            PTEY = sum(sum( TEY .* conj(TEY) ));
            REXTEY = sum(sum( REX .* conj(TEY) ));
            
            Pee = PREX + PTEY;
            Ree = 2*real(REXTEY);
            Iee = 2*imag(REXTEY);
            
            delL = 0;
            CD = ( Pee + Ree ) / ( Pee - Ree );
            
            if CD > CDlimit
                % dark port power = Pee + Ree cos(2 phi) - Iee sin(2 phi)
                % = Pee + Aee cos( 2 phi + atan(Iee/Ree) )
                % Aee = sqrt( Ree^2 + Iee^2 )
                % 2 phi + atan(Iee/Ree) = pi, phi = pi/2 - atan2(Iee, Ree)/2
                phi = mod( pi/2 - atan2(Iee, Ree)/2, pi );
                if phi > pi/2
                    phi = phi - pi;
                end
                delL = phi / (obj.fields.kCR + obj.fields.kSB) / 2;
                
                if abs(delL) > delLimit
                    Edark = REX*exp(1i*phi) + TEY*exp(-1i*phi);
                    Pdark = sum(sum(Edark .* conj(Edark)));
                    CDnew = Pdark / (2 * Pee );
                    
                    if CDnew < CD
                        CD = CDnew;
                        [~, propX1ID] = obj.props.getPropFrom( toBSX );
                        propX2ID = obj.props.findReverseProp( propX1ID );
                        
                        [~, propY1ID] = obj.props.getPropFrom( toBSY );
                        propY2ID = obj.props.findReverseProp( propY1ID );
                        
                        obj.props.propList(propX1ID).delL = obj.props.propList(propX1ID).delL - delL;
                        obj.props.propList(propX2ID).delL = obj.props.propList(propX2ID).delL - delL;
                        obj.props.propList(propY1ID).delL = obj.props.propList(propY1ID).delL + delL;
                        obj.props.propList(propY2ID).delL = obj.props.propList(propY2ID).delL + delL;
                    else
                        delL = 0;
                    end
                end
            end
            
            dark = abs(delL) < delLimit || CD < CDlimit;
        end
        
        % calculate field in a loop defined by pathLoop
        % if lockIt, propagator propID is adjuted to lock the cavity
        % calcuation information printed every printIt trial. 0 is no print
        function [err, numTry] = calcOne( obj, lockIt, printIt )
            if ~exist( 'lockIt', 'var' )
                lockIt = false;
            end
            if ~exist( 'printIt', 'var' )
                printIt = 0;
            end
            obj.inject2IFO(true);
            % calculate firlds in a cavity pathLoop,
            % optionally locking by adjuting propID propergator
            [err, numTry] = obj.cavities.compList(1).calc( lockIt, printIt );
        end
        
        % just lock
        function [err, numTry] = lockOne( obj )
            [err, numTry] = obj.calcOne( true );
        end
        
        function printAll( obj, fid )
            % function printAll( fid )
            % print information about fields, optics, etc.
            % If fid is a string, a file is created and the output is saved in the file.
            % If fid is not string, it is used as the file id to be used in fprintf
            
            fileOpened = false;
            if exist( 'fid', 'var' )
                % if fid is a text, it is assumed to be a file name to store
                if ischar( fid )
                    fid = fopen(fid, 'w');
                    fileOpened = true;
                end
            else
                % if nothing specified, it is 1 for screen
                fid = 1;
            end
            
            % print all information
            obj.fields.printAll( fid );
            obj.optics.printAll( fid );
            obj.props.printAll( fid );
            obj.cavities.printAll( fid );
            
            % if the file is opened here, close it
            if fileOpened
                fclose(fid);
            end
        end
        
        function printLoss( obj, fid )
            % function printLoss( fid )
            % print information about loss
            % If fid is a string, a file is created and the output is saved in the file.
            
            if exist( 'fid', 'var' ) && ischar( fid )
                fid = fopen(fid, 'w');
            else
                % if nothing specified, it is 1 for screen
                fid = 1;
            end
            
            obj.optics.printLoss( fid );
            
            if fid ~= 1  % fid is file opened
                fclose(fid);
            end
        end
        
        % from the field srcID, propagate and go through the first optic
        function transPort = inject2IFO( obj, setField )
            srcID = obj.optics.findLaser;
            if isnan( obj.fields.fieldList(srcID).invQ )
                transPort = 0;
                return;
            end
            
            if ~exist( 'setField', 'var' )
                setField = false;
            end
            
            % sinkID is input to the first optic hit by laser
            [sinkID, propID] = obj.props.getPropTo( srcID, true );
            if sinkID > 0
                obj.props.propField(propID);
                transPort = obj.optics.findTransPort( sinkID );
                if transPort > 0
                    if setField
                        obj.optics.trans( sinkID );
                    end
                end
                % May 10, 2016
                % bug fixed. This used to be elseif and no prompt reflection existed
                if obj.optics.findReflPort( sinkID ) > 0
                    if setField
                        obj.optics.refl( sinkID );
                    end
                end
                
                % laser and sink belongs to system wide guessList
                loopID = length( obj.cavities.guessList );
                obj.fields.markFieldStatus( srcID, loopID );
                obj.fields.markFieldStatus( sinkID, loopID );
            end
        end

        %% saving and restoring fields to restore a given state
        %    saveFields( tagID ) copies fields.fieldList(:).EFFT to savedField( tagID )
        %    retstoreFields( tagID [, saveID ] ) copies the savedField data back to fields.fieldList(:).EFFT
        %           if saveID is specified, saveFields( saveID ) is called before restoring, so same tagID can be used
        %
        function clearSavedFields( obj )
            obj.savedFields = repmat( struct('fieldID', 0, 'EFFT', obj.ffts.zeroMat), 0, 0 );
        end
        
        % function saveFields
        %    saves the current fields in local buffer
        %    savedFields = repmat( struct( 'fieldzeros( 0, 0, size(obj.ffts.zeroMat,1), size(obj.ffts.zeroMat,2) );
        %    fields = 'ALL' saves all fields
        function tagID = saveFields( obj, tagID, fields, saveName )
            if tagID == 0
                obj.clearSavedFields;
            end
            if tagID <= 0 || tagID > size( obj.savedFields, 1 )
                tagID = size( obj.savedFields, 1 ) + 1;
            end
            if exist( 'fields', 'var' )
                fieldIDs = obj.fields.conv2FieldIDList( fields );
            else
                if size( obj.savedFields, 1 ) > 0
                    fieldIDs = [obj.savedFields(end, :).fieldID];
                else
                    fieldIDs = 1:length(obj.fields.fieldList);
                end
            end
            for fid = 1 : length(fieldIDs)
                obj.savedFields(tagID, fid).fieldID = fieldIDs(fid);
                obj.savedFields(tagID, fid).EFFT = obj.fields.fieldList(fieldIDs(fid)).EFFT;
            end
            
            if exist('saveName', 'var')
                tempDataXXX = obj.savedFields(tagID,:);
                eval( sprintf('save %s.mat tempDataXXX;', saveName) );
                clear tempDataXXX;
            end
        end
        
        % function restoredGuess( restoredID, [saveID] )
        %    copies the saved guess fields to the current guess fields.
        %    If saveID is specified, the current guessField is saved in that set.
        %    With tagID < 0, fields are loaded from the file into main fields.fieldList.EFFT without using savedFields buffer
        function restoreFields( obj, tagID, saveID, loadName )
            if exist( 'loadName', 'var' )
                eval( sprintf( 'load %s.mat', loadName ) );
                if tagID < 0
                    for fid = 1 : size(tempDataXXX,2)
                        obj.fields.fieldList(tempDataXXX(fid).fieldID).EFFT = tempDataXXX(fid).EFFT;
                    end
                else
                    obj.savedFields(tagID, :) = tempDataXXX;
                end
            end
            if tagID > 0 && tagID <= size( obj.savedFields, 1 )
                tempID = 0;
                if exist( 'saveID', 'var' ) && ~isempty(saveID) && saveID > 0
                    if saveID == tagID
                        % current fields is stored to tagID and the saved field in tagID is restored
                        % so save the current field to tempID, restore tagID to fields then copy tempID to tagID
                        tempID = obj.saveFields(-1);
                    else
                        obj.saveFields( saveID );
                    end
                end
                for fid = 1 : size(obj.savedFields,2)
                    obj.fields.fieldList(obj.savedFields(tagID,fid).fieldID).EFFT = obj.savedFields(tagID, fid).EFFT;
                end
                if tempID > 0
                    % tempID has the current fields, so save those in tempID to tagID, and delete temp
                    for fid = 1 : size(obj.savedFields,2)
                        obj.savedFields(tagID, fid).EFFT = obj.savedFields(tempID, fid).EFFT;
                    end
                    obj.savedFields = obj.savedFields(1:end-1,:);
                end
            end
        end
        
        % save fields and delL
        function saveIFOState( obj, fileName )
            if ~exist( 'fileName', 'var' )
                fileName = sprintf('%s.mat', class(obj));
            end
            fieldDataInSavedFile___ = obj.fields;
            fieldDataInSavedFile___.ifo = nan;
            propDataInSavedFile___ = cell2mat({obj.props.propList(:).delL});
            miscInSavedFile___.Nfft = obj.ffts.Nfft;
            
            for id = 1 : length(fieldDataInSavedFile___.fieldList)
                fieldDataInSavedFile___.fieldList(id).ifo = nan;
            end
            save( fileName, 'fieldDataInSavedFile___', 'propDataInSavedFile___', 'miscInSavedFile___' );
            obj.fields.ifo = obj;
            for id = 1 : length(obj.fields.fieldList)
                obj.fields.fieldList(id).ifo = obj;
            end
        end
        
        % load fields and delL
        function loadIFOState( obj, fileName )
            if ~exist( 'fileName', 'var' )
                fileName = sprintf('%s.mat', class(obj));
            end
            
            % if file does not exist, dont bother
            if ~exist( fileName, 'file' )
                return;
            end
            
            load( fileName, 'miscInSavedFile___' );
            if ~exist('miscInSavedFile___', 'var') || miscInSavedFile___.Nfft == obj.ffts.Nfft
                
                load( fileName, 'fieldDataInSavedFile___', 'propDataInSavedFile___' );
                obj.fields = fieldDataInSavedFile___;
                
                obj.fields.ifo = obj;
                for id = 1 : length(obj.fields.fieldList)
                    obj.fields.fieldList(id).ifo = obj;
                end
                for id = 1 : length(obj.props.propList)
                    obj.props.propList(id).delL = propDataInSavedFile___(id);
                end
            else
                fprintf(1, '\n>>> Internal data inconsistent, saved data cannot be used\n');
                fprintf(1, '>>> NFFT(old) = %d, NFFT(new) = %d\n\n', miscInSavedFile___.Nfft, obj.ffts.Nfft);
            end
        end
    end
    
    methods ( Abstract )
        defineIFO( obj, varargin )
    end
    
    methods ( Static )
        function modeStr = mainModesStr( modes )
            modeStr = sprintf('%s\n%s\n', ...
                'index]   (m, n)   fraction / power   loss (frac*loss)   amplitude', ...
                '-------------------------------------------------------------------');
            for ind = 1 : size(modes,1)
                modeStr = sprintf('%s%2d] (%2d,%3d)   %8.4e / %8.4e   %8.4e (%8.4e)   %8.4e + %8.4ei\n', modeStr, ...
                    ind, modes(ind,1), modes(ind,2), modes(ind,4), abs(modes(ind,3)), modes(ind,5), modes(ind,7), real(modes(ind,3)), imag(modes(ind,3)) );
            end
            modeStr = sprintf('%s\n-----\n  fraction not covered by these modes = %12.4e\n', modeStr, 1-sum(modes(:,4)));
            modeStr = sprintf('%s  total of frac*loss = %12.4e\n', modeStr, sum(modes(:,7)));
        end
    end
end
