%
% classdef FFTTK
%   framework to prepare FFT-related basic quantities
%   based on FOG code
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% [Um, Vn] : N x N grid points in f-space
% [Xm, Yn] : N/2 x N/2 grid points in x-space
%
% FFT window size is set to be 1m,
% quantities with length dimention are multiplified by fac (assined for each object)
% quantities with frequency (1/length) are multiplied by 1/fac.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% March 24, 2017
%   zeroMat is defined in setNfft and is a normal property, not dependent
%
% October 30, 2014
%   major change for the update
%   removed bunch for cleanliness, moved some to UtilTK
%   zeroMat = zeros( N/2, N/2 ) is now accessed by using get function
%
% September 19, 2014
%   calcSumIn removed to be replaced by UtilTK.sumIn
%

classdef FFTTK < handle
    
    properties (Constant)
        % FFT window size unit is 1m
        windowUnit = 1
        % Real FFT window size is aliasMargin x physical window
        % to suppress the aliasing
        aliasMargin = 2
    end
    
%     properties (Dependent = true, SetAccess = private)
%         zeroMat
%     end

    properties
        Nfft      % number of grid points in full FFT space, Nfft/aliasMargin[2] for physical

        Dfft      % grid spacing, windowUnit[1] / Nfft
        % length dimentional quantities need to be multiplied by actual Wfft (full size)
        % e.g., xList = Xm * Wfft
        Xm        % [x,y] on N/2 x N/2 grid points
        Yn
        rmn_sq    % r^2 = Xm.^2 + Yn.^2
        rmn       % sqrt( rmn_sq )
        % frequency dimentional quantities need to be divided by actual Wfft
        Um        % [fx,fy] on N x N grid points
        Vn
        fmn_sq    % fx^2 + fy^2
        fmn       % sqrt( fx^2 + fy^2 )
        
        zeroMat   % zeros of the size of field
    end
    
    methods

        % no constructor needed
        
        function setNfft( obj, Nfft )

            obj.Nfft = 4*round(Nfft/4);
            obj.Dfft = FFTTK.windowUnit / obj.Nfft;

            % x-space grid points
            % -W/2+Dfft/2:Dfft:W/2-Dfft/2 = -W/2 + Dfft/2 + (0:N-1)*Dfft
            Grid_axis = -FFTTK.windowUnit/2 + obj.Dfft/2 + (0:obj.Nfft-1)*obj.Dfft;

            % f-space grid points
            % -fmax/2 + (0:N-1)*df = 1/W*(-N/2:1:N/2-1), fmax = 1/Dfft = N/W, df = 1/W
            Grid_axis_fft = -1/(2*obj.Dfft) + (0:obj.Nfft-1)/FFTTK.windowUnit;

            [obj.Xm,obj.Yn] = meshgrid(Grid_axis);
            [obj.Um,obj.Vn] = meshgrid(Grid_axis_fft);  % NxN

            % fx^2 + fy^2 with obj.Nfft^2 data points
            obj.fmn_sq = obj.Um.^2+obj.Vn.^2;
            obj.fmn = sqrt( obj.fmn_sq  );
            
            obj.Xm = UtilTK.useQuarter( obj.Xm );
            obj.Yn = UtilTK.useQuarter( obj.Yn );

            % r and r^2
            obj.rmn_sq=obj.Xm.^2+obj.Yn.^2;
            obj.rmn=sqrt(obj.rmn_sq);
            
            % zero field amplitude
            obj.zeroMat = zeros( size( obj.rmn ) );
        end
            
%         function zeroM = get.zeroMat( obj )
%             % zeroMat is a zero matrix with the size of physical window (N/2 x N/2)
%             zeroM = zeros( size( obj.rmn ) );
%         end
        
        function val = calcProdInt( obj, A, B, scale )
            % calcProdInt = integral of A * conj(B)  * scale^2
            if ~exist( 'scale', 'var' )
                scale = 1;
            end
            A( isnan(A) ) = 0;
            B( isnan(B) ) = 0;
            val = sum( sum( A .* conj(B) ) ) * ( obj.Dfft * scale )^2;
        end
        
    end
end
