%
% FieldObj : field object
%
% November 20, 2019
%   LGCoef and HGCoef fixed to handle varargin of (thetaX, thetaY, EFFT).
%   EFFT is standard format field
%   LGCoef and HGCoef shares function varargin4ModeCoef for common codes
%
% November 5, 2019
%   1) tileField added, which tilts the field by adding exp( -i k n ). The amplitude will need to be scaled by the change
%   of ethe coordinate size, but it will be 1 - cos(theta) ~ theta^2 and will be negligible.
%   2) LGCoef and HGCoef have varargin, which can be (thetaX, thetaY) for tilt, and last EFFT as speific amplitude (internal format)
%
% October 10, 2019
%   calcSumIn( obj, radius, x0, y0 ) suports offset
%
% August 8, 2018
%   A new variable portIDfrom added.
%   When a new field is added by refl or trans to a port which has already a field,
%   portIDfrom has the port ID or field ID.
%   This allows to associate multiple fields to one port.
%
% July 30, 2018
%   HGField2DatFID and LGField2DatFID has invQ as the last optional argument.
%
% May 18, 2018
%   updateField modified
%   several function arguments are changed for consistency, not functional change
%
% March 6, 2017
%   addEFFT is added to simpliy obj.EFFT = obj.EFFT + delEFFT
%
% Febraury 20, 2017
%   static and non-static member function to convert from standard to internal format added
%      Eint = conv2intSTATIC( Estd, phiCorr, rsq, Wfft )
%      Eint = conv2int( obj, Estd )
%
% September 19, 2014
%   gaussFit info are processed in modeFitData
%   calcSumIn uses UtilTK.sumIn, instead of FFTTK
%
% September 16, 2014
%   static member function added to convert from internal to standard
%       Estd = conv2stdSTATIC( Eint, phiCorr, rsq, Wfft )
%
% March 8th, 2014
%   changed the superclass of FieldObj from handle to matlab.mixin.Copyable
%   to allow copying existing fields
%
% March 3rd, 2014 HY
%   Analytic solution, Eamp, is moved from twiddle object to this FieldObj.
%
% September 14, 2013
%
% ff(1:numFlds) = struct( 'mirID', 0, 'port', 0, 'RoC', 1e18, 'CL', 0, ...
%    'w', 1, 'di', 0, 'Cw', 1, 'fac', 1, 'L', 1,  'CLa', 1, ...
%    'rmapC', 0, 'tmapC', 0, 'transToID', 0, 'temp', 0 );
%
%       mirID : mirror this field is attached to
%       port : port index, >=0 HR, < 0 AR, even in field, odd out field
%       RoC : mirror surface RoC
%       CL = 1 / field curvature, 1/RoC if wave front = mirror curvature
%       w  : beam size on the mirror
%
%
%
classdef FieldObj < matlab.mixin.Copyable
    
    % Using invQ and PrivateInvQ to implement setInvQ is explained in
    % "Avoiding Property Initialization Order Dependency"
    properties (Dependent)
        % modal model base
        % this is to set invQ dependent variables in invQ set function
        invQ % q = z + i z0; invQ = invRoC - i 2/k 1/w^2
    end
    
    properties(Access=private)
        PrivateInvQ
    end
    
    properties
        % IFO object
        ifo
        
        % name of the object
        name
        
        % calculated in set.invQ
        % values based on the nominal mode base
        z
        z0
        w
        invRoC
        
        % phiCorr is generalized invRoC
        % E(standard form) = E(internal form) * exp( -i k/2 r^2 phiCorr ) / Wfft
        % E(internal form) = E(standard form) * exp( +i k/2 r^2 phiCorr ) * Wfft
        phiCorr = 0;
        
        % FFT field in internal form. Use above formula or conv2stdSTATIC to convet to standard format
        % This points to an element of fieldManager.Amp defined in get.EFFT
        EFFT
        
        % optics this field is attached to
        optID
        
        % 'status' is a flag if this has been calculated - may carry some information in the future
        % This is cleared at the entrance of calc and lock calls, and fields in closed loops
        % are filled by non-zeros marking these have valid calculated field.
        % When getField is called, if this flag is 0, it is traced back to non-zero
        % fields to calculate.
        status = 0;
        
        % field amplitude calculated by twiddle
        % make this to 1 if twiddle initialization is to be removed.
        % Then, the printInfo will need to be changed.
        Eamp = nan;
        
        % When this field comes out from a port in a non-standard way,
        % like a new field added by FFTIFO.refl or trans, this has the standatd field ID or port ID.
        % This allows multiple fields be asociated with one port ID
        portIDfrom = 0;
    end
    
    methods
        
        function obj = FieldObj( name, ifo )
            obj.ifo = ifo;
            
            obj.name = name;
            obj.invQ = NaN;
        end
        
        function xList = getXlist( obj )
            xList = obj.ifo.optics.getXlist( obj.optID );
        end
        
        function Wfft = getWfft( obj )
            Wfft = obj.ifo.optics.getWfft( obj.optID );
        end
        
        function val = get.invQ( obj )
            val = obj.PrivateInvQ;
        end
        
        function set.invQ( obj, newInvQ )
            % set.invQ( newInvQ )
            % sets invQ and calculate various modal model quantities
            obj.PrivateInvQ = newInvQ;
            
            obj.w = MMTK.qinv2W( newInvQ, obj.ifo.fields.kCR );
            obj.invRoC = real( newInvQ );
            obj.z = real( 1/newInvQ );
            obj.z0 = imag( 1/newInvQ );
        end
        
        function addEFFT( obj, EFFTin )
            % obj.EFFT = obj.EFFT + EFFT
            obj.EFFT = obj.EFFT + EFFTin;
        end
        
        function updateField( obj )
            % function updateField( obj, fieldID, props, optics )
            % Those fields with status <= 0 have not be calculaed yet,
            % like the field going out from the AR side of ETM.
            % This one calculates the field using inputs to the optics
            if (obj.status > 0) || (obj.status == -1) || isnan( obj.invQ )
                % this field is ready (>0) or no known source (-1), so do nothing
                return
            end
            
            fieldID = find( obj.ifo.fields.fieldList == obj );
            if isempty(fieldID) || fieldID == 0
                return;
            end
            
            [srcID, propID] = obj.ifo.props.getPropFrom( fieldID );
            if srcID < 0
                % this is a start of a prop, so find interaction
                tempEFFT = obj.ifo.ffts.zeroMat;
                srcID = obj.ifo.optics.findReflPort( fieldID, true );
                if srcID > 0
                    % there is relection
                    % make the input to the relfection input port ready
                    obj.ifo.fields.fieldList(srcID).updateField;
                    % save refletion in tempEFFT
                    tempEFFT = obj.ifo.optics.refl( srcID );
                end
                srcID = obj.ifo.optics.findTransPort( fieldID, true );
                if srcID > 0
                    % there is transmission
                    % make the input to the tranmission input port ready
                    obj.ifo.fields.fieldList(srcID).updateField;
                    % save the total to tempEFFT
                    tempEFFT = tempEFFT + obj.ifo.optics.trans( srcID );
                end
                obj.ifo.fields.fieldList(fieldID).EFFT = tempEFFT;
            else
                % this is at the end of propagator, so let the start ready and fill
                obj.ifo.fields.fieldList(srcID).updateField;
                % let it prop from src to sink
                obj.ifo.props.propField( propID );
            end
            % mark it is done
            obj.status = 1024;
        end
        
        % standard format HG function with the mode defined at port fldID
        function [E, xList] = HGField2DatFID( obj, m, n, invQ )
            if ~exist( 'invQ', 'var' )
                invQ = obj.invQ;
            end
            [invRoCQ, wQ] = MMTK.qinv2RinvW( invQ, obj.ifo.fields.kCR );
            xList = obj.getXlist;
            E = HGTK.HGField2DRW( m, n, xList, xList, invRoCQ, wQ, obj.ifo.fields.kCR );
        end
        
        % standard format LG function with the mode defined at port fldID
        function [E, xList] = LGFieldatFID( obj, p, m, invQ )
            if ~exist( 'invQ', 'var' )
                invQ = obj.invQ;
            end
            [invRoCQ, wQ] = MMTK.qinv2RinvW( invQ, obj.ifo.fields.kCR );
            xList = obj.getXlist;
            E = LGTK.LGFieldRW( p, m, xList, xList, invRoCQ, wQ, obj.ifo.fields.kCR );
        end
        
        % varargin = (thetaX, thetaY, EFFT)
        function Cmn = HGCoef( obj, m, n, varargin )
            %  HG coefficient at fldID port, for a standard field Fld
            Cmn = zeros( length(m), length(n) );
            
            % Fld is tilted in this function
            [~, ~, Fld] = obj.varargin4ModeCoef( varargin{:} );
            
            for im = 1 : length(m)
                for in = 1 : length(n)
                    Cmn(im,in) = obj.ifo.ffts.calcProdInt( Fld, obj.HGField2DatFID( m(im), n(in) ), obj.getWfft );
                end
            end
        end
        
        % varargin = (thetaX, thetaY, EFFT)
        function Cpm = LGCoef( obj, p, m, varargin )
            %  LG coefficient at fldID port, for a standard field Fld
            Cpm = zeros( length(p), length(m) );
            
            % Fld is tilted in this function
            [~, ~, Fld] = obj.varargin4ModeCoef( varargin{:} );
            
            for ip = 1 : length(p)
                for im = 1 : length(m)
                    Cpm(ip,im) = obj.ifo.ffts.calcProdInt( Fld, obj.LGFieldatFID( p(ip), m(im) ), obj.getWfft );
                end
            end
        end
        
        function [thetaX, thetaY, Fld] = varargin4ModeCoef( obj, varargin )
            thetaX = 0;
            thetaY = 0;
            if isempty( varargin )
                Fld = obj.conv2std( obj.EFFT );
            else
                if length( varargin ) == 1 && length( varargin{1} ) > 1 % varargin = EFFT
                    Fld = varargin{1};
                else
                    if length( varargin) >= 3
                        Fld = varargin{3};
                        thetaY = varargin{2};
                        thetaX = varargin{1};
                    else
                        Fld = obj.conv2std( obj.EFFT );
                        if length( varargin ) >= 2 % thetaY is given
                            thetaY = varargin{2};
                            thetaX = varargin{1};
                        else
                            thetaY = 0;
                            if length( varargin ) >= 1 && ~isempty(varargin{1})
                                thetaX = varargin{1};
                            else
                                thetaX = 0;
                            end
                        end
                    end
                end
                
                if thetaX ~= 0 || thetaY ~= 0
                    Fld = obj.tiltField( thetaX, thetaY, Fld );
                end
                
            end
        end
        
        
        function outField = tiltField( obj, thetaX, thetaY, inField )
            if thetaX == 0 && thetaY == 0
                outField = inField;
            else
                xList = obj.getXlist;
                [xg, yg] = meshgrid( xList, xList );
                outField = exp(-1i*(thetaX*xg + thetaY*yg)*obj.ifo.fields.kCR) .* inField;
            end
        end
        
        function [info, cwR] = gaussFit( obj, message )
            % [info, cwR] = gaussFit( obj, message )
            % Fit the field by gaussian form
            % exp( -(x-x0)^2*(1/wx^2) + i/(2*Rx) )
            % cwR(1, 2) for x and y
            % cwR(:).c0 = center of gauss field
            % cwR(:).w  = the beam width
            % cwR(:).invRk = k / RoC
            if isempty( obj.EFFT )
                info = 'no field';
            else
                [cwR, info] = modeFitData( obj.getXlist, obj.getXlist, obj.conv2std(obj.EFFT), obj.ifo.fields.kCR );
                if obj.invRoC == 0
                    RoC = 'inf';
                else
                    RoC = sprintf('%-13.10g', 1/obj.invRoC);
                end
                qInv = obj.invRoC - 2i/(obj.ifo.fields.kCR * obj.w^2);
                info = sprintf('%s\n                          Mode base : w = %13.10g, R = %s, z = %13.10g', info, obj.w, RoC, real(1/qInv) );
            end
            if exist( 'message', 'var' )
                info = strcat(message, info);
            end
        end
        
        % function power = calcSumIn( obj, radius, x0, y0 )
        %   calcute power in a given radius
        function power = calcSumIn( obj, radius, x0, y0 )
            %        	power = obj.ifo.ffts.calcSumIn( obj.EFFT, radius/obj.ifo.optics.optList(obj.optID).Wfft );
            % magnitude of the field internal format is standard format * Wfft
            if isempty( obj.EFFT )
                power = 0;
            else
                if ~exist( 'x0', 'var' )
                    x0 = 0;
                end
                if ~exist( 'y0', 'var' )
                    y0 = 0;
                end
                power = UtilTK.sumIn( obj.getXlist, obj.EFFT / obj.getWfft, radius, radius, x0, y0 );
            end
        end
        
        function power = calcPower( obj, stdFld )
            % calculate power of the field
            % |EFFT| = standard amplitude * Wfft
            % integral of standard amp^2 = prodInt( stdamp, stdamp, Wfft )
            % = prodInt( stdamp*Wfft, stdamp*Wfft, 1 ) = prodInt( EFFT, EFFT, 1 )
            if exist('stdFld', 'var')
                power = obj.ifo.ffts.calcProdInt( stdFld, stdFld, obj.getWfft );
            else
                power = obj.ifo.ffts.calcProdInt( obj.EFFT, obj.EFFT, 1 );
            end
        end
        
        function Estd = conv2std( obj, Eint )
            % convert internal format to standard
            if ~exist( 'Eint', 'var' )
                Eint = obj.EFFT;
            end
            if isempty( Eint )
                fprintf(1, 'Field ''%s'' is empty\n', obj.name);
                Estd = [];
            else
                Estd = obj.conv2stdSTATIC( Eint, obj.phiCorr, obj.ifo.optics.getRsq(obj.optID), ...
                    obj.getWfft, obj.ifo.fields.kCR );
            end
        end
        
        function Eint = conv2int( obj, Estd )
            % convert standard format to internal format
            if nargout == 1
                Eint = obj.conv2intSTATIC( Estd, obj.phiCorr, obj.ifo.optics.getRsq(obj.optID), ...
                    obj.getWfft, obj.ifo.fields.kCR );
            else
                obj.EFFT = obj.conv2intSTATIC( Estd, obj.phiCorr, obj.ifo.optics.getRsq(obj.optID), ...
                    obj.getWfft, obj.ifo.fields.kCR );
            end
        end
        
        function info = printInfo( obj, all )
            if ~exist( 'all', 'var' )
                all = true;
            end
            %            obj.updateField;
            if isempty( obj.EFFT )
                if ~all
                    info = '';
                    return;
                end
                str = 'FFT field is empty';
            else
                str = sprintf( 'Power = %g', obj.calcPower );
            end
            if isnan( obj.Eamp )
                analStr = 'no analytic evaluation';
            else
                if obj.Eamp == 0
                    analStr = 'E(analytic) = 0';
                elseif imag(obj.Eamp) == 0
                    analStr = sprintf( 'E(analytic) = %g', real(obj.Eamp) );
                else
                    analStr = sprintf( 'E(analytic) = %g + i %g', real(obj.Eamp), imag(obj.Eamp) );
                end
            end
            info = sprintf('name: %s   ====   %s, %s, [status = %d]', obj.name, str, analStr, obj.status );
            
            info = sprintf('%s\nRoC = %g, w = %g, z = %g, z0 = %g, 1/phiCorr = %g', ...
                info, 1/obj.invRoC, obj.w, obj.z, obj.z0, 1/obj.phiCorr );
            info = sprintf('%s\nq = %g + i %g, 1/q = %g - i %g', ...
                info, real(1/obj.invQ), imag(1/obj.invQ), real(obj.invQ), abs(imag(obj.invQ)) );
            
        end
        
    end
    
    methods (Static)
        function Estd = conv2stdSTATIC( Eint, phiCorr, rsq, Wfft, kCR )
            % conversion from interal to standard format
            Estd = exp( -1i * kCR / 2 * phiCorr * rsq ) .* Eint / Wfft;
        end
        
        function Eint = conv2intSTATIC( Estd, phiCorr, rsq, Wfft, kCR )
            % conversion from standard to internal format
            Eint = exp( 1i * kCR / 2 * phiCorr * rsq ) .* Estd * Wfft;
        end
        
    end
    
end
