%
% GuessObj : information necessary for guess field calculation
%       This is build after cavity is analyzed and FFT are ready to be used
%       When ths object is created and the path is analyzed,
%       input fields may not be ready to be used
%

%% pathRec = struct('srcID', 0, 'action', 0, 'id', 0, 'statusID', 0 )
%    src field ID
%    action : (1) propagation or    <=
%             (2) reflection or     <=
%             (3) transmision       <=
%             (4) add pump field    <=
%             (5) fill guess field  <=
%             (6) save guess field  <=
%             (7) save pumpOut for others
%             (8) add actuation field
%             (9) void - do nothing
%     id :
%       prop           : propID
%       reflection     : none
%       transmission   : none
%       addActuation   : actuation field ID
%       add pump field : id is the GuessObj index from which the field is provided
%           field(srcID) = field(srcID) + dumpObj( id )
%     statusID :  N : add to exiting field N
%              :  0 : fill anew
%              : -N : store in pump field(N)
%
% April 12, 2020
%   new action actAddActuationField added.
%      This action add actuation field on the field which is going to start propagate
%   function addActuator( srcField, actField ) added
%       which modified the masterPath so that, just before prop(srcField), addActuator(actField) is inserted
% 
% February 8th, 2019
%   function goMasterPath( addGuessPump, resetStatus ) 
%       has new argument resetStatus to reset status of some fields which need to be calculated again.
%
% November 28, 2018
%   removeAddPumpField removed, no more used
%
% May 18, 2018
%   More information in debug output
%   geussCopy property for external use
%
% March 25, 2017
%   addPath returns the index of newly added path
%   areBothPaths added which checks if prop and prop_bck are both in the path
%
% March 14, 2017
%   In goMasterPath, isDebugOn is called before debug.print is called, because sometimes, the argument of
%   debug.print('someinfo') consumes time before print is called, which itseld checks if debug is on internally.
%
% March 6, 2017
%   fieldObj.EFFT = fieldObj.EFFT + delEFFT is replaced by fieldObj.addEFFT( delEFFT)
%
% March 2, 2017
%   Modified to be used with Richard's acceleration code
%   addPath( path ) accepts an cell array of field names, like { 'MC1-HRx-o', ... }
%   in addition to the legacy numerical format [7 20 ...].
% 
% October 23, 2014
%   Non static DebugTK is accessed using obj.ifo.debug handle, mostly for debug information printout
%
% July 31, 2014
%   pump field is default mode
%
% November 21th, 2013  Hiro Yamamoto
%

%%
classdef GuessObj < handle
    
    properties (Constant)
        %    action : (1) propagation or    <=
        %             (2) reflection or     <=
        %             (3) transmision       <=
        %             (4) add pump field    <=
        %             (5) fill guess field  <=
        %             (6) save guess field  <=
        %             (7) save pumpOut for others
        %             (8) add actuation field
        %             (9) no action
        actionNames = {'propagation   ', 'reflection    ', 'transmision   ', 'add pump fld  ', ...
                       'fill guess fld', 'save guess fld', 'save pumpOut  ', 'add actuation field', 'void', 'unknown action' };
        
        actPropagation = 1;
        actReflection = 2;
        actTransmission = 3;
        actAddPumpField = 4;
        actFillGuessField = 5;
        actSaveGuessField = 6;
        actSavePumpOut = 7;
        actAddActuationField = 8;
        actVoid = 9; % no action
    end
    
    properties
        name = 'unnamed'
        
        id  % index of this guess obj, correspnding to closedCavs(id)
        
        % example : laser (field ID 2) -> PRM HR (field ID 5)
        
        % pumpOut is to be used as sources for others
        % pumpOut{1,:} is the source to produce the pump field, like laser source ( 2 for the example )
        % pumpOut{2,:} is the portID to prepare for others, like field inside of PRM ( 5 for the example )
        % pumpOut{3,:} is the NxN field amp other pump field uses : field of 5 with a given input 2
        % in obj.fetchPumpField( id ), pumpID = find( pumpOut{2,:} == id ); field = pumpOut{3,pumpID}
        pumpOut
        
        % pumpIn is a list of pumpIn sources provided by others
        % pumpIn(1,:) : pump source guess ID (guess ID where this pump field is provided)
        % pumpIn(2,:) : external source ID  ( 2 for the example )
        % pumpIn(3,:) : my port to which this pump field is to be added ( 5 for the example )
        pumpIn
        
        % parameters for the accelaration code
        Nacc = 2;
        Nave = 0;
        Nprint = 5;
        lockPropID = -1;  % first prop connected to the guess field
        
        % handle of parent
        ifo
        
        % list of paths for this loop, part of the tree
        pathList
        
        % reassembled actions by comining all elements in pathList
        % each item in this list is executed one after another for one loop execution
        % buildMasterPath build masterPath
        %    obj.masterPath = repmat( struct('srcID', 0, 'action', 0, 'id', 0, 'statusID', 0 ), 1, 0 );
        masterPath
        
        % guessFieldID
        guessID
        % saved guess field
        % at the beginning of the loop, field(guessID) is saved in guessField
        % at the end of the loop, field(guessID) has the new field, guessField has the old field
        guessField
        
        % guessField(saveID, size(field)) : saved copy of guess field
        guessCopy

        % if this is ON, action for this guess field is called when its inputs are ready
        onTheFly = false;
        
        errorLimit = 1e-6;
        
        % round trip phase is tried to be set to this
        lockPhase = 0;
        
        % status
        % 0 : calc
        % -1 : calc, but for the first calc, lock it. After lock, this becomes 1
        %  1 : lock if not freezed
        %  2 : force lock
        status
     end
    
    methods
        % constructor
        function obj = GuessObj( name, id, ifo, guessID )
            obj.id = id;
            obj.ifo = ifo;
            
            if exist( 'guessID', 'var' )
                obj.guessID = guessID;
            else
                obj.guessID = nan; % to be filled
            end
            obj.onTheFly = false;

            % this is a special case, maintained separately, like laser ource input
            if id == 0
                obj.guessID = 0;
                obj.onTheFly = false;
            elseif id >= 1 && id <= length(obj.ifo.cavities.closedCavs)
                % first field of the close cavity
                obj.guessID = obj.ifo.cavities.closedCavs{id}(1);
                % if the cavity is a closed cavity all facing HR side, like FP or MC, do it right away
                obj.onTheFly = obj.ifo.cavities.isHRloop( obj.ifo.cavities.closedCavs{id} );
            end
            
            if obj.onTheFly
                obj.Nprint = 20;
            end
            
            if isempty( name )
                if obj.onTheFly
                    numMirrors = length(obj.ifo.cavities.closedCavs{id})-1;
                    switch numMirrors
                        case 2
                            guessName = 'FP';
                        case 3
                            guessName = 'MCTri';
                        otherwise
                            guessName = sprintf('MC%dmir', numMirrors );
                    end
                else
                    guessName = 'guess';
                end
                obj.name = sprintf( '%s%d', guessName, id );
            else
                obj.name = name;
            end
            
            obj.pumpOut = cell(3,0);
            obj.pumpIn = zeros(3,0);
            
            obj.pathList = cell(1,0);
            
            obj.initGuessField;
            
            obj.status = 0;
        end
        
        function info = printInfo( obj )
            info = sprintf('[%2d] ''%s'' : guessID = %3d\nNacc = %2d, Nave = %2d, Nprint = %2d\n', ...
                obj.id, obj.name, obj.guessID, obj.Nacc, obj.Nave, obj.Nprint);
            info = sprintf('%sonTheFlay = %d, errorLimit = %g, lockPhase = %g, status = %d', ...
                info, obj.onTheFly, obj.errorLimit, obj.lockPhase, obj.status);
        end
        
        % print masterPath content
        %     obj.masterPath = repmat( struct('srcID', 0, 'action', 0, 'id', 0, 'statusID', 0 ), 1, 0 );
        function printMasterPath( obj )
            for index = 1 : length( obj.masterPath )
                nameIndex = obj.masterPath(index).action;
                if nameIndex <= 0 || nameIndex >= length( obj.actionNames )
                    nameIndex = length( obj.actionNames );
                end
                if nameIndex == obj.actVoid || nameIndex == length( obj.actionNames )
                    sprintf('%2d) %s\n', index, obj.actionNames{nameIndex});
                else
                    pathID = obj.masterPath(index).id;
                    switch nameIndex
                        case obj.actPropagation % propagation
                            idStr = sprintf('propagtor ''%s''[%2d]', obj.ifo.props.getName(pathID), pathID );
                        case obj.actAddPumpField % add pump field
                            idStr = sprintf('pump field from guessList ''%s''[%2d]', ...
                                obj.ifo.cavities.guessList(pathID).name, pathID );
                        case obj.actAddActuationField % actuation field added
                            idStr = sprintf('actuation field ''%s''[%2d]', ...
                                obj.ifo.fields.fieldList(pathID).name, pathID );
                        otherwise
                            idStr = 'id not used';
                    end
                    statusID = obj.masterPath(index).statusID;
                    if isempty( statusID )
                        statusStr = 'statudID not used';
                    else
                        if statusID == 0
                            statusStr = sprintf('store output');
                        elseif statusID > 0
                            statusStr = sprintf('add output to ''%s''[%2d]', obj.ifo.fields.getName(abs(statusID)), abs(statusID));
                        else
                            statusID = -statusID;
                            statusStr = sprintf('store output in dump field of ''%s''[%2d]', obj.ifo.cavities.guessList(statusID).name, statusID);
                        end
                    end
                    fprintf('%2d) src ''%s''[%2d]; action %s; %s; %s\n', ...
                        index, obj.ifo.fields.getName(obj.masterPath(index).srcID), obj.masterPath(index).srcID, ...
                        obj.actionNames{nameIndex}, idStr, statusStr );
                end
            end
        end

        function clearTempFields( obj )
            for ipo = 1 : size(obj.pumpOut,2)
                obj.pumpOut{3,ipo} = [];
            end
            obj.initGuessField;
        end
        
        % calculate something
        function err = action( obj, errorLimit, Nprint )
            if ~exist( 'errorLimit', 'var' )
                errorLimit = obj.errorLimit;
            end
            if ~exist( 'Nprint', 'var' )
                Nprint = obj.Nprint;
            end
            
            % if freezeLength, then don't lock
            % calc, but lock once
            if obj.status == -1
                lockProp = obj.lockPropID;
                obj.status = 1;  % lock if not frozen
            else
                if obj.ifo.cavities.freezeLength
                    lockProp = 0;
                else
                    lockProp = obj.lockPropID;
                end
            end
            obj.ifo.debug.print(['%s >> guessList( %s ).providePump( %s ) called for %s using src field %s \n' ...
                '    calling %s.AccClosedCavityStationary\n'],  ...
                obj.name, ...
                obj.ifo.cavities.guessList( obj.pumpIn(1,1) ).name , obj.ifo.fields.findName(obj.guessID), ...
                obj.ifo.fields.findName(obj.pumpIn(3,1)), obj.ifo.fields.findName(obj.pumpIn(2,1)), ...
                obj.name );
            err = obj.AccClosedCavityStationary( obj.ifo.cavities.guessList( obj.pumpIn(1,1) ).providePump(obj.guessID) , ...
                errorLimit, lockProp, obj.Nacc, obj.Nave, Nprint );
        end
        
        % new single path is added to this database
        function pathIndex = addPath( obj, path )
            if ~isempty( path )
                if iscell( path )
                    pathOut = zeros(1, length( path ));
                    for index = 1 : length( path )
                        pathOut(index) = obj.ifo.fields.findNamedField( path{index} );
                    end
                    path = pathOut;
                end
                pathIndex = length( obj.pathList) + 1;
                obj.pathList{ pathIndex } = path;
            else
                pathIndex = 0;
            end
        end
        
        % test if reverse propagator is in the path
        function doublePath = areBothPaths( obj, propID, pathIndex )
            doublePath = false;
            if propID > 0
                inPort = obj.ifo.props.propList(propID).inPortID;
                if ~isempty( find( obj.pathList{ pathIndex }(1:end-1) == inPort, 1 ) )
                    propID = obj.ifo.props.findReverseProp( propID );
                    if propID > 0
                        inPort = obj.ifo.props.propList(propID).inPortID;
                        if ~isempty( find( obj.pathList{ pathIndex }(1:end-1) == inPort, 1 ) )
                            doublePath = true;
                        end
                    end
                end
            end
        end
        
        % set the guess field or empty for now
        % this needs to be called when field calculation starts, at least the first one
        function initGuessField( obj, inField )
            if exist( 'inField', 'var' )
                obj.guessField = inField;
            else
                obj.guessField = obj.ifo.cavities.ifo.ffts.zeroMat;
            end
        end
        
        %% building masterPath
        %
        % after FFT is ready, this is called to convert a sequence of path to
        % a sequence of events -
        % transmission, reflection and propagation
        % pumping into from other loops and
        % pump out preparation for other loops
        %
        function buildMasterPath( obj, saveOrRestoreGuessField )
            if ~exist( 'saveOrRestoreGuessField', 'var' )
                saveOrRestoreGuessField = obj.actSaveGuessField;
            end
            
            % analPath is an intermediate data
            % analPath{1,n) is source to analPath(2,n)
            analPath = zeros(3, length( obj.pathList{1} )-1);
            for in = 1 : length( obj.pathList{1} )-1
                analPath(1, in) = obj.pathList{1}(in);
                analPath(2, in) = obj.pathList{1}(in+1);
            end
            for ip = 2 : length( obj.pathList )
                thisPath = obj.pathList{ip};
                thisLeng = length(thisPath);
                for in = 1 : thisLeng-1
                    % find where mismatch occurs
                    analPathLeng = size( analPath, 2 );
                    if (analPath(1, in) ~= thisPath(in)) || ...
                            (analPath(2, in) ~= thisPath(in+1))
                        lengMove = length(thisPath) - in;
                        analPath(:, in+lengMove:analPathLeng+lengMove) = analPath(:, in:analPathLeng);
                        for ik = 0 : lengMove-1
                            analPath(1, in+ik) = thisPath(in+ik);
                            analPath(2, in+ik) = thisPath(in+ik+1);
                        end
                        break;
                    end
                    if analPathLeng ~= size( analPath, 2 )
                        break;
                    end
                end
            end
            
            % sort analPath
            % [-N,M] : external pump source, to be added to M
            % [-N,-M] : discard
            % [N,-M] : pump souce for other guess loop
            % [N,M] : prop
            in = 1;
            while in <= size( analPath, 2)
                keepIt = true;
                if analPath(1, in) < 0 && analPath(2, in) > 0
                    % external pump source, to be added
                    pumpInID = size(obj.pumpIn,2) + 1;
                    obj.pumpIn(2,pumpInID) = -analPath(1, in);
                    ipo = obj.ifo.cavities.findClosedCavID( obj.pumpIn(2,pumpInID) );
                    if ipo > 0
                        obj.pumpIn(1,pumpInID) = ipo;
                    else
                        obj.pumpIn(1,pumpInID) = length( obj.ifo.cavities.guessList );
                        obj.ifo.cavities.guessList(end).addPumpOut( -analPath(1, in), analPath(2, in) );
                    end
                    obj.pumpIn(3,pumpInID) = analPath(2, in);
                    keepIt = false;
                elseif analPath(1, in) < 0 && analPath(2, in) < 0
                    % for now not necessary
                    keepIt = false;
                elseif analPath(1, in) > 0 && analPath(2, in) < 0
                    % pump souce for other guess loop
                    obj.addPumpOut( analPath(1, in), -analPath(2, in) );
                    keepIt = true;
                end
                if keepIt
                    in = in + 1;
                else
                    analPath(:, in:end-1) = analPath(:, in+1:end);
                    analPath(:, end) = [];
                end
            end
            
            for ip = 1 : size( analPath, 2 )
                if analPath(2,ip) == obj.guessID || ~isempty( find( analPath(2,1:ip-1) == analPath(2,ip), 1 ) )
                    % guess field is filled at the beginning, so external source is to be added
                    % if already filled, add it
                    analPath(3,ip) = analPath(2,ip);
                else
                    % this is the first time, so fill it
                    analPath(3,ip) = 0;
                end
            end
            
            % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % masterPath is defined here
            %
            %   save current guess field to obj.guessField
            %   add pumpField for this guess field, if there is any
            %   go through the path
            %   guess field is calculated at the last action
            %   save pumpOut fields for others
            %
            % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            obj.masterPath = repmat( struct('srcID', 0, 'action', 0, 'id', 0, 'statusID', 0 ), 1, 0 );
            % at first, restore guess field
            %            obj.addOneAction( 5 );
            % save the current field at guessID to guessField
            obj.addOneAction( saveOrRestoreGuessField );
            obj.addPumpIn( obj.guessID );
            
            for ip = 1 : size( analPath, 2 )
                obj.addOneAction( analPath( 3, ip ), analPath( 1, ip ), analPath( 2, ip ) );
            end
            % save guess field
            %            obj.addOneAction( 6 );
            % calculate and store pump field for thers
            obj.addOneAction( obj.actSavePumpOut );
            
            for ip = length( obj.masterPath ) : -1 : 1
                for ip2 = ip-1 : -1 : 1
                    if obj.masterPath(ip2).srcID == obj.masterPath(ip).srcID && ...
                            obj.masterPath(ip2).action == obj.masterPath(ip).action
                        obj.masterPath(ip).action = obj.actVoid;
                        break;
                    end
                end
            end

         end
        
        
        % register pump field which are needed for others
        function numPumps = addPumpOut( obj, srcID, portID )
            numPumps = size( obj.pumpOut, 2 );
            for ip = 1 : numPumps
                if obj.pumpOut{1,ip} == srcID && obj.pumpOut{2,ip} == portID
                    return;
                end
            end
            numPumps = numPumps + 1;
            obj.pumpOut{1,numPumps} = srcID;
            obj.pumpOut{2,numPumps} = portID;
            obj.pumpOut{3,numPumps} = [];
        end                    
        
        %% prepare masterPath
%        actPropagation = 1;
%        actReflection = 2;
%        actTransmission = 3;
%        actAddPumpField = 4;
%        actFillGuessField = 5;
%        actSaveGuessField = 6;
%        actSavePumpOut = 7;
%        actAddActuationField = 8;

        % when nargin == 4, i.e., user provides action, srcID and sinkID, call addOnePair
        % id = 5, 6, 7 for preparing action 5,6,7
        function addOneAction( obj, action, id1, id2 )
            if nargin == 4
                obj.addOnePair( action, id1, id2 );
            else
                if action == obj.actFillGuessField || action == obj.actSaveGuessField
                    % save and restore guessField
                    ID = length( obj.masterPath ) + 1;
                    obj.masterPath(ID).action = action;
                    obj.masterPath(ID).srcID = obj.guessID;
                    if action == obj.actFillGuessField
                        obj.addPumpIn( obj.guessID );
                    end
                    
                elseif action == obj.actSavePumpOut
                    % save pumpOut for others
                    for ipo = 1 : size( obj.pumpOut, 2 )
                        destLoopID = obj.ifo.cavities.findClosedCavID( obj.pumpOut{2,ipo} );
                        if ~obj.ifo.cavities.guessList( destLoopID ).onTheFly
                            % for onTheFly loop, no need to save
                            % because it is saved explicitly before action is called
                            obj.preparePumpOut( ipo );
                        end
                    end
                end
            end
        end

        % use Eamp as the pump source amplitude and calculate pump field
        % this is, for example, PRC loop in aLIGO configuration before SRC loop is calculated
        function fillPumpOuts( obj )
            for ipo = 1 : size( obj.pumpOut, 2 )
                src = abs( obj.pumpOut{1,ipo} );  % src field, which propagates to sinkID
                sinkID = obj.ifo.props.getPropTo( src, true );
                % pump field is default mode
                obj.ifo.fields.setDefField( sinkID, 1 );
                obj.ifo.fields.fieldList(sinkID).EFFT = ...
                    obj.ifo.fields.fieldList(sinkID).EFFT * obj.ifo.fields.fieldList(sinkID).Eamp;
                obj.pumpOut{3,ipo} = obj.ifo.optics.trans( sinkID );
            end
        end
        
        % define action to prepare pumpout field
        function preparePumpOut( obj, ipo )
            % src props to sinkID to outID
            src = abs( obj.pumpOut{1,ipo} );
            outID = obj.pumpOut{2,ipo};
            sinkID = obj.ifo.props.getPropTo( src, true );
            % should be trasmitted
            obj.addReflOrTrans( sinkID, outID, -ipo );
        end
        
        function [propTo, propID] = addOneProp( obj, srcID )
            [propTo, propID] = obj.ifo.props.getPropTo( srcID );
            if propID <= 0
                return;
            end
            for ip = 1 : length( obj.masterPath )
                if obj.masterPath(ip).action == 1 && obj.masterPath(ip).id == propID
                    return;
                end
            end
            ID = length( obj.masterPath ) + 1;
            obj.masterPath(ID).srcID = srcID;
            obj.masterPath(ID).action = obj.actPropagation;
            obj.masterPath(ID).id = propID;
        end
        
        function [kind, ID] = addReflOrTrans( obj, srcID, sinkID, statusID )
            kind = 0;
            if sinkID < 0
                kind = -sinkID;  % ????
            else
                if obj.ifo.optics.findReflPort( srcID ) == sinkID
                    kind = 2;
                elseif obj.ifo.optics.findTransPort( srcID ) == sinkID
                    kind = 3;
                end
            end
            if kind == 0
                return;
            end
            ID = length( obj.masterPath ) + 1;
            obj.masterPath(ID).action = kind;
            obj.masterPath(ID).srcID = srcID;
            obj.masterPath(ID).statusID = statusID;
        end
        
        %  N to M : propagate and trans or refl
        % -N to M : case 4 get external pump source
        %  N to -M : call action if onTheFly flag is ON
        function addOnePair( obj, statusID, srcID, sinkID )
            if srcID > 0 && sinkID > 0
                % add prop
                [propTo, propID] = obj.addOneProp( srcID );
                if propID <= 0
                    return;
                end
                
                if sinkID == obj.guessID
                    statusID = 0;
                end
                if obj.addReflOrTrans( propTo, sinkID, statusID ) > 0
                    % if pumpIn is to be added
                    if sinkID ~= obj.guessID
                        obj.addPumpIn( sinkID );
                    end
                end
            elseif srcID < 0 && sinkID > 0
                % external pump field
                % for now, manually add when internal field is calculated
                %{
                ID = length( obj.masterPath ) + 1;
                obj.masterPath(ID).srcID = sinkID;
                obj.masterPath(ID).action = obj.actAddPumpField;
                obj.masterPath(ID).statusID = statusID;
                obj.masterPath(ID).id = obj.ifo.cavities.findClosedLoop( abs(srcID) );
                if obj.masterPath(ID).id == 0
                    % the last guessList is system wide server
                    obj.masterPath(ID).id = length( obj.ifo.cavities.guessList );
                end
                %}
            elseif srcID > 0 && sinkID < 0
                % go into other guess loop
                [~, propID] = obj.addOneProp( srcID );
                if propID <= 0
                    return;
                end
                iguess = obj.ifo.cavities.findClosedCavID( abs(sinkID) );
                if obj.ifo.cavities.guessList(iguess).onTheFly
                    ipo = find( [obj.pumpOut{2,:}] == abs(sinkID), 1 );
                    obj.preparePumpOut( ipo );
                    ID = length( obj.masterPath ) + 1;
                    obj.masterPath(ID).action = obj.actSavePumpOut;
                    obj.masterPath(ID).srcID = iguess;
                end
            end
        end
        
        % add action to pumpIn
        function addPumpIn( obj, srcID )
            ipi = find( obj.pumpIn(3, :) == srcID, 1 );
            if ~isempty( ipi )
                ID = length( obj.masterPath ) + 1;
                obj.masterPath(ID).srcID = srcID;
                obj.masterPath(ID).action = obj.actAddPumpField;
                obj.masterPath(ID).statusID = obj.pumpIn(2,ipi); % only non-zero
                obj.masterPath(ID).id = obj.pumpIn(1,ipi);  % pump source guess loop
            end
        end
        
        %% help to execute masterPath
        
        % on request, return pump field for other guess field calculation
        % outID is the field ID to which the pump field is to be added
        function fld = providePump( obj, outID )
            % find the pumpOut index which has the pump field
            ipo = find( [obj.pumpOut{2,:}] == outID );
            obj.ifo.debug.print('==> pumpOut requested to fill field %s, originated by field %s\n', ...
                obj.ifo.fields.findName(outID), obj.ifo.fields.findName(obj.pumpOut{1,ipo}) );
            if isempty( obj.pumpOut{3,ipo} )
                % if the dump field is not ready yet, calculate now
                src = abs( obj.pumpOut{1,ipo} );
                [sinkID, propID] = obj.ifo.props.getPropTo( src, true );
                obj.ifo.props.propField(propID);
                if obj.ifo.optics.findReflPort( sinkID ) == outID
                    obj.pumpOut{3,ipo} = obj.ifo.optics.refl( sinkID );
                    obj.ifo.debug.print('   pumpOut calculated by reflecting field %d\n', sinkID );
                elseif obj.ifo.optics.findTransPort( sinkID ) == outID
                    obj.pumpOut{3,ipo} = obj.ifo.optics.trans( sinkID );
                    obj.ifo.debug.print('   pumpOut calculated by transmissing field %d\n', sinkID );
                else
                    obj.pumpOut{3,ipo} = obj.ifo.cavities.fields.ffts.zeroMat;
                    obj.ifo.debug.print('   pumpOut filled by empty field\n' );
                end
            else
                obj.ifo.debug.print('   pumpOut field already prepared\n' );
            end
            fld = obj.pumpOut{3,ipo};
        end
        
        %% execute the masterPth instruction set
        
        % addGuessPump controls if the pump field for the guess field is to be added or not
        % For calculational convenience, pump field can be explicitly added to the guess field
        % before goMasterPath is called. When added already, make addGuessPump false to
        % avoid dumpling the pumping for the guess field
        %
        % When resetStatus is true, status of fields udated by propgagation, reflection and transmission are reset.
        % When the RTL is calculated, fields in the loop are evaluated to calculate the loss. 
        % These fields need to be calculated again when the nominal field values are needed.
        function goMasterPath( obj, addGuessPump, resetStatus )
            if ~exist( 'addGuessPump', 'var' )
                addGuessPump = false;
            end
            
            if ~exist( 'resetStatus', 'var' )
                resetStatus = false;
            end
            
            debugStatus = obj.ifo.debug.isDebugOn;
            if debugStatus
                emptyHeader = char(ones(1,length(obj.ifo.debug.title))*' ');
            end
            
            for istep = 1 : length( obj.masterPath )
                resetID = 0;
                srcID = obj.masterPath(istep).srcID;
                statusID = obj.masterPath(istep).statusID;
                ID = obj.masterPath(istep).id;
                nextAction = obj.masterPath(istep).action;
                
                switch nextAction
                    case obj.actVoid
                        if debugStatus
                            obj.ifo.debug.print( 'no action\n' );
                        end
                        
                    case obj.actPropagation   % propagation
                        obj.ifo.props.propField( ID );
                        if debugStatus                            
                            obj.ifo.debug.print( '%s\n%s%s\n', obj.ifo.props.propInfo(ID), ...
                                emptyHeader, obj.extraDebugInfo( nextAction, ID, emptyHeader ) );
                        end
                                         
                        if resetStatus
                            resetID = obj.ifo.props.propList(ID).outPortID;
                        end
                        
                    case obj.actAddActuationField  % add actuation field
                        if statusID > 0   
                            obj.ifo.fields.fieldList(statusID).EFFT = ...
                                obj.ifo.fields.fieldList( statusID ).EFFT + ...
                                obj.ifo.fields.fieldList( ID ).EFFT;
                            if debugStatus
                                obj.ifo.debug.print( 'Actuation field %s [%2d] added to field %s [%2d]\n', ...
                                obj.ifo.fields.fieldList(ID).name, ID, obj.ifo.fields.findName(statusID), statusID );
                            end
                        end

                    case obj.actReflection   % reflection
                        if statusID == 0
                            obj.ifo.optics.refl( srcID );
                            if debugStatus
                                obj.ifo.debug.print( '%s and output is stored\n%s%s\n', obj.ifo.optics.reflInfo( srcID ), ...
                                     emptyHeader, obj.extraDebugInfo( nextAction, srcID, emptyHeader ) );
                            end
                        elseif statusID > 0
                            obj.ifo.fields.fieldList( statusID ).addEFFT( obj.ifo.optics.refl( srcID ) );
                            if debugStatus
                                obj.ifo.debug.print( '%s\n     and output is added to field %s\n', ...
                                    obj.ifo.optics.reflInfo( srcID ), obj.ifo.fields.findName(statusID) );
                            end
                        else
                            obj.pumpOut( -statusID ) = ...
                                obj.ifo.optics.refl( srcID );
                            if debugStatus
                                obj.ifo.debug.print( '%s\n     and output is stored in pump field %s\n', ...
                                    obj.ifo.optics.reflInfo( srcID ), obj.ifo.fields.findName(obj.pumpOut{2, -statusID}) );
                            end
                        end
                        
                        if resetStatus
                            resetID = obj.ifo.optics.findReflPort(srcID);
                        end
                        
                    case obj.actTransmission   % transmission
                        if statusID == 0
                            obj.ifo.optics.trans( srcID );
                            if debugStatus
                                obj.ifo.debug.print( '%s and output is stored\n%s%s\n', obj.ifo.optics.transInfo( srcID ), ...
                                    emptyHeader, obj.extraDebugInfo( nextAction, srcID, emptyHeader ) );
                            end
                            
                        elseif statusID > 0   % BUG found on 2/13/2018, .EFFT was missing, never came here?
                            obj.ifo.fields.fieldList( statusID).EFFT = ...
                                obj.ifo.fields.fieldList( statusID ).EFFT + ...
                                obj.ifo.optics.trans( srcID );
                            if debugStatus
                                obj.ifo.debug.print( '%s\n     and output is added to field %s\n', ...
                                    obj.ifo.optics.transInfo( srcID ), obj.ifo.fields.findName(statusID) );
                            end
                        else
                            obj.pumpOut{ 3, -statusID } = ...
                                obj.ifo.optics.trans( srcID );
                            if debugStatus
                                obj.ifo.debug.print( '%s\n     and output is stored in pump field %s\n', ...
                                    obj.ifo.optics.transInfo( srcID ), obj.ifo.fields.findName(obj.pumpOut{2, -statusID}) );
                            end
                        end
                        
                        if resetStatus
                            resetID = obj.ifo.optics.findTransPort(srcID);
                        end

                    case obj.actAddPumpField  % add pump field
                        if addGuessPump || (srcID ~= obj.guessID)
                            % if dont add pump field for guess field, then dont add pump for guess
                            if statusID == 0
                                obj.ifo.fields.fieldList( srcID ).EFFT = ...
                                    obj.ifo.cavities.guessList(ID).providePump( srcID );
                                if debugStatus
                                    obj.ifo.debug.print( 'pump field stored to %s from guess ID %d\n', ...
                                        obj.ifo.fields.findName(srcID), ID );
                                end
                            else
                                obj.ifo.fields.fieldList( srcID ).addEFFT( obj.ifo.cavities.guessList(ID).providePump( srcID ) );
                                if debugStatus
                                    obj.ifo.debug.print( 'pump field added to %%s from guess %d\n', ...
                                        obj.ifo.fields.findName(srcID), ID );
                                end
                            end
                        end
                        
                    case obj.actFillGuessField  % restore guessfield
                        obj.ifo.fields.fieldList( obj.guessID ).EFFT = obj.guessField;
                        if debugStatus
                            obj.ifo.debug.print( 'guess field restored to field( %d )\n%s%s\n', obj.guessID, ...
                                emptyHeader, obj.extraDebugInfo( nextAction, obj.guessID, emptyHeader ) );
                        end
                        
                    case obj.actSaveGuessField  % save guess field
                        obj.guessField = obj.ifo.fields.fieldList( obj.guessID ).EFFT;
                        if debugStatus
                            obj.ifo.debug.print( 'guess field is saved from field( %s )\n', obj.ifo.fields.findName(obj.guessID) );
                        end
                    
                    case obj.actSavePumpOut  % call other guess field
                        obj.ifo.cavities.guessList( srcID ).action;
                        if debugStatus
                            obj.ifo.debug.print( 'action for guess field loop %d called\n' , srcID );
                        end
                        
                    otherwise
                        if debugStatus
                            obj.ifo.debug.print( 'unknown action\n' );
                        end
                end
                
                if resetStatus && resetID ~= 0
                    obj.ifo.resetStatus( resetID );
                end

            end
        end
        
        function msg = extraDebugInfo( obj, action, ID, emptyHeader )
            switch action
                case obj.actPropagation
                    % propagation phase
%                    Cmn = obj.ifo.fields.modeCoef( obj.ifo.props.propList(ID).HR2HRProp );
                    % this is between actual ports, not HR to HR
                    Cmn = obj.ifo.fields.modeCoef( [obj.ifo.props.propList(ID).inPortID, obj.ifo.props.propList(ID).outPortID] );
                    Cin = Cmn(1); Cout = Cmn(2);
                    msg = sprintf('    Cin = %e EXP(i %e pi), Cout =%e EXP(i %e pi), Cout/Cin = %e EXP(i %e pi)', ...
                        abs(Cin), atan2(imag(Cin), real(Cin))/pi, ...
                        abs(Cout), atan2(imag(Cout), real(Cout))/pi, ...
                        abs(Cout/Cin), atan2(imag(Cout/Cin), real(Cout/Cin))/pi );
                    % if prop ends on AR side, extend the propagation to HR
                    portIDs = obj.ifo.props.propList(ID).HR2HRProp;
                    if portIDs(2) ~= obj.ifo.props.propList(ID).outPortID
                        [opticID, portIndex] = obj.ifo.optics.findOptic( obj.ifo.props.propList(ID).outPortID );
                        if portIndex < 0
                            obj.ifo.optics.optList(opticID).trans( -portIndex );
                            Cout = obj.ifo.fields.modeCoef( portIDs(2) ); %, '', 0, 0, obj.ifo.optics.optList(opticID).trans( -portIndex ) );
                            msg = sprintf('%s\n%s    with prop in out ==>  Cout =%e EXP(i %e pi), Cout/Cin = %e EXP(i %e pi)', ...
                                msg, emptyHeader, abs(Cout), atan2(imag(Cout), real(Cout))/pi, ...
                                abs(Cout/Cin), atan2(imag(Cout/Cin), real(Cout/Cin)) / pi );
                        end
                    end

                case obj.actReflection 
                    outID = obj.ifo.optics.findReflPort(ID);
                    Cin = obj.ifo.LGCoef(ID,0,0);
                    Cout = obj.ifo.LGCoef(outID,0,0);
                    Aout = abs( Cout );
                    phiOut = atan2(imag(Cout), real(Cout)) / pi;
                    phiOutIn = atan2(imag(Cout/Cin), real(Cout/Cin)) / pi;
                    if phiOutIn > 0.5
                        phiOut = phiOut - 1;
                        phiOutIn = phiOutIn - 1;
                        Aout = -Aout;
                    elseif phiOutIn < -0.5
                        phiOut = phiOut + 1;
                        phiOutIn = phiOutIn + 1;
                        Aout = -Aout;
                    end                   
                    msg = sprintf('    Cin = %e EXP(i %e pi), Cout =%e EXP(i %e pi), Cout/Cin = %e EXP(i %e pi)', ...
                        abs(Cin), atan2(imag(Cin), real(Cin))/pi, ...
                        Aout, phiOut, Aout/abs(Cin), phiOutIn );
                    
                case obj.actTransmission
                    outID = obj.ifo.optics.findTransPort( ID );
                    Cin = obj.ifo.LGCoef(ID,0,0);
                    Cout = obj.ifo.LGCoef(outID,0,0);
                    msg = sprintf('    Cin = %e EXP(i %e pi), Cout =%e EXP(i %e pi), Cout/Cin = %e EXP(i %e pi)', ...
                        abs(Cin), atan2(imag(Cin), real(Cin))/pi, ...
                        abs(Cout), atan2(imag(Cout), real(Cout))/pi, ...
                        abs(Cout/Cin), atan2(imag(Cout/Cin), real(Cout/Cin))/pi );
                    
                case obj.actFillGuessField
                    Cout = obj.ifo.LGCoef(ID,0,0);
                    msg = sprintf('    Cin(guess %d) = (%e + %e pi i) = %e EXP(i %e pi)', ...
                        ID, real(Cout), imag(Cout), abs(Cout), atan2(imag(Cout), real(Cout))/pi );
                    
                otherwise
                    msg = 'nothing available';
            end
        end
        
        %% loop calculation
        
        % To use Epump = N for AccClosedCavityStationary and ClosedCavityStationary,
        % guess field needs to be initialized as the pumpIn field
        
        % use accelarated code by Richard
        % AccN : how many iterations for accelaration calculation
        % AveN : how many samples used for one iretation
        % if lockPropID > 0, then this prop is changed to move the cavity to locked point
        function [err, numTry] = ...
                AccClosedCavityStationary( obj, Epump, errorLimit, lockPropID, AccN, AveN, Nprint )
            if ~exist( 'lockPropID', 'var' )
                lockPropID = 0;
            end
            if isa( lockPropID, 'char' )
                nameStr = lockPropID;
                lockPropID = obj.ifo.props.findNamedProp( lockPropID );
                if isempty(lockPropID)
                    error('This propagator does not exist : ''%s''', nameStr);
                end
            end
            if lockPropID == -1
                lockPropID = obj.ifo.props.getPropID( obj.guessID );
            end
            
            if ~exist( 'Nprint', 'var' )
                Nprint = 0;
            end
            if ~exist( 'AccN', 'var' )
                AccN = 5;
            end
            if ~exist( 'AveN', 'var' )
                AveN = 2;
            end
            if AccN <= 1
                % no acceleration, just simple method
                [err, numTry] = obj.ClosedCavityStationary( Epump, errorLimit, lockPropID, Nprint );
                return;
            end
            
            if length( Epump ) == 1
                % is Epump is given by a number,
                % pump field is the value stored in the field(guessID) and
                % the true initial guess field is Empump * the pump field
                if Epump > 0
                    factor = Epump;
                else
                    factor = 1;
                end
                Epump = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                obj.ifo.fields.fieldList(obj.guessID).EFFT = ...
                    factor * obj.ifo.fields.fieldList(obj.guessID).EFFT;
            end
            
            NC = cell(AccN,1);
            NC(:) = {zeros(size(Epump))};
            EN = cell(AccN+1,1);
            EN(:) = {zeros(size(Epump))};
            
            if AveN==0
                AveN=ones(1,AccN);
            else
                AveN=AveN*ones(1,AccN);
                AveN(2:2:end)=AveN(2:2:end)+1;
            end
            
            numTry = 0;
            for iloop=1:1000
                if lockPropID > 0
                    obj.easyLock( lockPropID );
                end
                
                Elast = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                
                EN(:) = {zeros(size(Epump))}; NC(:) = {zeros(size(Epump))};
                for m=1:AccN
                    for l=1:AveN(m)
                        numTry = numTry + 1;
                        eGuess = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                        EN{m} = EN{m} + eGuess;
                        obj.goMasterPath( false );
                        NC{m} = NC{m} + (eGuess - obj.ifo.fields.fieldList(obj.guessID).EFFT);
                        obj.ifo.fields.fieldList(obj.guessID).addEFFT( Epump );
                        obj.ifo.debug.print('%s >> Epump added to %s\n', obj.name, obj.ifo.fields.findName(obj.guessID));
                        obj.ifo.debug.print('%s >> end of %d th averaging out of %d\n', obj.name, l, AveN(m));
                    end
                    EN{m}=EN{m}/AveN(m); NC{m}=NC{m}/AveN(m);
                    obj.ifo.debug.print('%s >> end of %d th acceleration of out %d\n', obj.name, m, AccN);
                end
                
                Et{1} = Epump;
                EG = R_getguessFin( NC',EN',Et);
                obj.ifo.fields.fieldList(obj.guessID).EFFT = EG{1};
                
                err = UtilTK.evalError( Elast, obj.ifo.fields.fieldList(obj.guessID).EFFT );
                
                if err < errorLimit
                    obj.goMasterPath( false );  % test convergence
                    obj.ifo.fields.fieldList(obj.guessID).EFFT = EG{1};
                end
                
                if Nprint > 0 && mod(iloop,Nprint)==0
                    fprintf(1, '    %s >> Acc Stationary finding : %d tries  power = %g, error = %g, FTT count = %g\n', ...
                        obj.name, numTry, obj.ifo.fields.fieldList(obj.guessID).calcPower, err, obj.ifo.props.propCounter  );
                end
                
                if err < errorLimit
                    break;
                end
            end
        end
        
        % stationary field in a closed cavity defined by cavityPath, with input field Epump, is calculated
        % this is no acceleration
        % When Epump is a positive number, Epump is the initial guess field value and the guess field is set to
        % be Epump x guess field. For a FP, set the initial value of guess field to be the pump in field and
        % set the Epump to be finess, and the convergence could be accelerated.
        %
        % if Nprint > 0, convergence information is printed every Nprint
        % if lockPropID > 0, then this prop is changed to move the cavity to locked point
        %
        % This is a copy of ClosedCavityStationary from cavityManager, customized for GuessObj.
        % The one in cavityManager will be removed in the future.
        %
        function [err, numTry] = ClosedCavityStationary( obj, Epump, errorLimit, lockPropID, Nprint )
            if ~exist( 'lockPropID', 'var' )
                % no locking if lockPropID is not given
                lockPropID = 0;
            end
            if isa( lockPropID, 'char' )
                % if lockPropID is given by name, change it to ID
                nameStr = lockPropID;
                lockPropID = obj.ifo.props.findNamedProp( lockPropID );
                if isempty(lockPropID)
                    error('This propagator does not exist : ''%s''', nameStr);
                end
            end
            if lockPropID == -1
                lockPropID = obj.ifo.props.getPropID( obj.guessID );
            end
            if ~exist( 'Nprint', 'var' )
                % default is no printing
                Nprint = 0;
            end
            
            if length( Epump ) == 1
                % is Epump is given by a number,
                % pump field is the value stored in the field(guessID) and
                % the true initial guess field is Empump * the pump field
                if Epump > 0
                    factor = Epump;
                else
                    factor = 1;
                end
                Epump = obj.ifo.fields.fieldList(obj.guessID).EFFT;
                obj.ifo.fields.fieldList(obj.guessID).EFFT = ...
                    factor * obj.ifo.fields.fieldList(obj.guessID).EFFT;
            end
            
            for numTry = 1:2000
                if (lockPropID > 0) && (mod(numTry, 10) == 1)
                    obj.easyLock( lockPropID );
                end
                
                obj.goMasterPath(false);
                obj.ifo.fields.fieldList(obj.guessID).addEFFT( Epump );
                err = UtilTK.evalError( obj.guessField, obj.ifo.fields.fieldList(obj.guessID).EFFT );
                 if err < errorLimit
                     obj.goMasterPath( false );  % test convergence
                     obj.ifo.fields.fieldList(obj.guessID).addEFFT( Epump );
                 end

                if (Nprint > 0) && (mod(numTry, Nprint) == 0)
                    fprintf(1, '    %s >> Stationary finding : %d  power = %g, error = %g, FFT count = %g\n', ...
                        obj.name, numTry, obj.ifo.fields.fieldList(obj.guessID).calcPower, err, obj.ifo.props.propCounter );
                end
                if err < errorLimit
                    break;
                end
            end
        end
        
        % change the microscopic path length of the propergator propID to make the round trip phase to null
        function delL = easyLock( obj, propID )
            % ??? lockPhase to be fixed
            delL = ( obj.calcLoopPhase - obj.lockPhase * 0 ) / (obj.ifo.fields.kCR + obj.ifo.fields.kSB);
            backID = obj.ifo.props.findReverseProp( propID );
            if backID > 0
                % this length change is applied for two paths, go and back, so make it half
                delL = delL / 2;
                obj.ifo.props.propList(backID).delL = obj.ifo.props.propList(backID).delL + delL;
            end
            obj.ifo.props.propList(propID).delL = obj.ifo.props.propList(propID).delL + delL;
        end
        
        % find round trip phase
        function loopPhase = calcLoopPhase( obj )
            current = obj.ifo.cavities.freezeLength( true );
            currStatus = obj.ifo.fields.fieldList(obj.guessID).status;
           
            % when the masterPath starts, gield(guessID) is stored in obj.guessField
            obj.goMasterPath( false, true );
            EEProd = sum(sum( obj.ifo.fields.fieldList(obj.guessID).EFFT .*conj(obj.guessField) ));
            loopPhase = atan2( imag(EEProd), real(EEProd) );
            % restore initial state
            obj.ifo.fields.fieldList(obj.guessID).EFFT = obj.guessField;
            obj.ifo.fields.fieldList(obj.guessID).status = currStatus;
            obj.ifo.fields.calcIfNecessary;
            obj.ifo.cavities.freezeLength( current );
        end
        
        function loss = roundtripLoss( obj )
            % function loss = roundtripripLoss( obj )
            % find round trip loss
            
            current = obj.ifo.cavities.freezeLength( true );
            currStatus = obj.ifo.fields.fieldList(obj.guessID).status;
            
            Pold = obj.ifo.fields.fieldList( obj.guessID ).calcPower;
            % when the masterPath starts, gield(guessID) is stored in obj.guessField
            obj.goMasterPath( false, true );
            
            Pnew = obj.ifo.fields.fieldList( obj.guessID ).calcPower;
            % at each bounce, reflectance is multiplied
            RProd = 1;
            for im = 1:length( obj.masterPath )
                % for each reflection, multiply reflectance on HR
                if obj.masterPath( im ).action == 2
                    [~, optID] = obj.ifo.optics.isHRside( obj.masterPath( im ).srcID );
                    %{
                    if isHR
                        rProd = rProd *  obj.ifo.optics.optList(optID).rHR;
                    else
                        rProd = rProd *  obj.ifo.optics.optList(optID).rAR;
                    end
                    %}
                    RProd = RProd * obj.ifo.optics.optList(optID).Rhr0;
                end
            end
            loss = 1 - Pnew/RProd/Pold;
            
            % restore initial state
            obj.ifo.fields.fieldList(obj.guessID).EFFT = obj.guessField;
            obj.ifo.fields.fieldList(obj.guessID).status = currStatus;
            obj.ifo.fields.calcIfNecessary;
            obj.ifo.cavities.freezeLength( current );
        end
        
    end
end
