% Hermite Gauss functions
%
% April 3rd, 2019
%     dxyz(3), a.ka.c0, is now the offset in z, renamed to dxyz
%     Field calculated using invQ, F0, is moved by (dxyz(1), dxyz(2), dxyz(3))
%     Field(x,y,z) = F0( x-dxyz(1), y-dxyz(2), z-dxyz(3) )
%     R and w are calculated using Q = Q input - dxyz(3)
%
% March 5th, 2019
%     modeList = HGModes( nm ) and LGModes added
%     These returns (index1, index2) of modes specified by nm
%     For nm >= 0, index+index2 <= nm 
%     For nm < 0,  index+index2 == nm
%     (nx, ny) = (0,nm), (1,nm), ... (nm,0)
%
%  November 28, 2018
%    HG1DN argument array can be row or column.
%
%  May 18, 2018
%    HermiteSum added, which calculates sums of Hermite polynomials with a given weight
%
% May 11, 2016
%   dispacement and tilt added to HGField
%   The cross term may not be correctly implemented. 
%
% Septemer 2nd, 2015
%   explicit argument of kCR
%
% October 14, 2013  Hiro Yamamoto  hiro@ligo.caltech.edu
%   first release
%

classdef HGTK < handle
    
    methods (Static)
  
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % modeList(id,1:2) = (index1, index2) with 
        %    nm >= 0 : index1 + index2 <= nm
        %    nm < 0  : index1 + index2 = abs(nm)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function modeList = HGModes( nm )
            if nm < 0
                mnmin = abs(nm);
            else
                mnmin = 0;
            end
            modeList = zeros(0,2);
            for imn = abs(nm) : -1 : mnmin
                for m = 0 : imn
                    modeList(size(modeList,1)+1,1:2) = [m, imn-m];
                end
            end            
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % 2D Hermite Gaussian with invQ as input
        % HGField2D( m, n, x, y, invQ ) or
        % HGField2DRW( m, n, x, y, invRoC, w )
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
        function E = HGField2D( m, n, x, y, invQ, k, dxyz, nvec )
            if ~exist( 'dxyz', 'var' )
                dxyz = [0,0,0];
            end
            if ~exist( 'nvec', 'var' )
                nvec = [0,0];
            end
          	[invRoC, w] = MMTK.qinv2RinvW( invQ, k );
            E = HGTK.HGField2DRW( m, n, x, y, invRoC, w, k, dxyz, nvec );
        end
        
        function E = HGField2DRW( m, n, x, y, invRoC, w, k, dxyz, nvec )
            if ~exist( 'dxyz', 'var' )
                dxyz = [0,0,0];
            end
            if ~exist( 'nvec', 'var' )
                nvec = [0,0];
            end
            if dxyz(3) ~= 0
                [invRoC, w] = MMTK.RinvWZ2RinvW( invRoC, w, -dxyz(3), k );
            end
            E = HGTK.HGField1D( n, y, invRoC, w, k, dxyz(2), nvec(2) ).' * ...
                HGTK.HGField1D( m, x, invRoC, w, k, dxyz(1), nvec(1) );
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % hermite gaussian field specified by
        % the inverse of RoC and the beam size
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function E = HGField1D( m, x, invRoC, w, k, c0, nvec )
            if ~exist( 'c0', 'var' )
                c0 = 0;
            end
            if ~exist( 'nvec', 'var' )
                nvec = 0;
            end
            x = x - c0;
            E = sqrt( sqrt(2)/w ) * HGTK.HG1DN( m, sqrt(2)*x/w ) .* exp(-1i*k*(invRoC/2*x.^2 + x*nvec) );
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % normalized (w=1) 1D Hermite Gaussian
        % int[-inf,inf] HG(m,x) HG(n,x) = 1 for m=n
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function data = HG1DN( m, x )
            if size(m,1) < size(m,2)
                m = m';
            end
            transX = size(x,1) > size(x,2);
            if transX
                x = x';
            end
            
            data = ...
                repmat(1 ./ sqrt( sqrt(pi) * pow2(m) .* gamma(m+1) ),1,size(x,2)) ...
                .* HGTK.Hermite( m, x ) ...
                .* repmat(exp( -x.^2/2 ), size(m,1), 1 );
            if transX
                data = data';
            end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % HermiteSum
        %   when coef(:,1) = m, coef(:,2) = n; coef(:,3) = weight are given
        %   data = sum( weight * Hermite(m, x(1,:)) * Hermite(n, y(:,1)) )
        %   is calculated. The amplitude is normalized so that a surface with this weigt
        %   induces HG(n,m) mode with amplitude of 2 k weight when a TEM00 is injected
        %   x and y are 2d data
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        function data = HermiteSum( x, y, coef, w )
            data = zeros( size(x) );
            for ig = 1:size( coef, 1 )
                data = data + coef(ig,3)  / sqrt( pow2(coef(ig,1)) * gamma(coef(ig,1)+1) * pow2(coef(ig,2)) * gamma(coef(ig,2)+1) ) ...
                    * repmat( HGTK.Hermite( coef(ig,1), sqrt(2)/w*x(1,:) ), size(y,1), 1 ) .* repmat( HGTK.Hermite( coef(ig,2), sqrt(2)/w*y(:,1).' ).', 1, size(x,2) );
            end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Hermite polynomial
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        function data = Hermite( m, x )
            if size(m,1) < size(m,2)
                m = m';
            end
            if size(x,1) > size(x,2)
                x = x';
            end
            xLeng = size(x,2);
            
            data = zeros( length(m), length(x) );
            for im = 1 : length(m)
                switch m(im)
                    case 0
                        data(im,1:xLeng) = 1;
                    case 1
                        data(im,:) = 2*x;
                    case 2
                        data(im,:) = -2 + 4*x.*x;
                    case 3
                        data(im,:) = ( 8*x.*x - 12 ) .* x;
                    case 4
                        x2 = x.*x;
                        data(im,:) = (16*x2 - 48).*x2 + 12;
                    case 5
                        x2 = x.*x;
                        data(im,:) = ((32*x2-160).*x2 + 120).*x;
                    otherwise
                        data(im,1:xLeng) = 1;
                        coeff = 1;
                        ms = 0;
                        s2 = m;
                        x22 = 4*x.*x;
                        for s = 1 : floor(m/2)
                            ms = ms + 1;
                            coeff = -coeff * s2*(s2-1) / ms;
                            s2 = s2 - 2;
                            data(im,:) = data(im,:) .* x22 + coeff;
                        end
                        if mod(m,2) ~= 0
                            data(im,:) = data(im,:) * 2 .* x;
                        end
                end
            end

        end
    end
end
