%
% mirror2x2 : mirror with fields of normal incidence
%
% August 5th, 2014
%   removeCurvature option is added to Mirror2x2
%
%   April 4, 2014
%       When finite optTheta, optPhi is set by addMirror, they were applied in the code
%       Only when tile removal was applied, they were used.
%       Fixed, always applied.
%
%   September 13, 2013
%
%             HR     AR
%  inPort 1  ==> || | ==> outPort 2
%                || |
%  outPort 1 <== || | <== inPort 2
%
%
% September 2, 2015
%    Using new OpticObj structure with ifo property, mostly
%    changing ConstantTK.kLam0 to obj.ifo.fields.kCR
%
%

classdef Mirror2x2Obj < MirrorObj
    
	properties (Constant)
        inputNames2x2 = { 'invRoCAR', 'removeTilt', 'removeCurvature' };
    end
    
    methods (Static)
        function tagList = getInputNames
            tagList = [ Mirror2x2Obj.inputNames2x2, MirrorObj.getInputNames ];
        end
    end
    
    methods
        function obj = Mirror2x2Obj( name, ifo, specs )
            
            obj = obj@MirrorObj( ifo, 2, 2, name, specs );
            
            UtilTK.updateProperties( obj, obj.inputNames2x2, specs );

            obj.suffix = {'HR', 'AR'};

            obj.reflMatch  = [1, 2];
            obj.transMatch = [2, 1];
                        
            obj.updateNdep;
 
        end
        
        % ABCD matrix of a simple 2x2 mirror, which is characterized by 
        % RoCHR, RoCAR, lens f val in substrate, thickness and refractive index
        function updateNdep( obj, ~, ~ )
            % refl(1) : HR in -> HR out, refl(2) : AR in -> AR out
            obj.reflABCD(1:2) = { MMTK.ReflN_ABCD( obj.invRoC ), ...
                MMTK.ReflAR_ABCD( obj.invFocal, obj.invRoC, obj.invRoCAR, obj.nRef, obj.Thickness ) };

            % lens by AR and HR side curvature
            lensAR = MMTK.LensR_ABCD( obj.invRoCAR, obj.nRef );
            lensHR = MMTK.LensR_ABCD( obj.invRoC , obj.nRef ) * MMTK.LensF_ABCD(obj.invFocal);
            % transmission through central substrate
            transM = MMTK.Prop_ABCD( obj.Thickness, obj.nRef );
            % lens and trans do not conmute, but no correct way ...
            % trans(1) : HR in -> AR out, trans(2) : AR in -> HR out
            obj.transABCD(1:2) = { lensAR * transM * lensHR, lensHR * transM * lensAR };
        end
        
        % build maps for standard mirror
        % it takes into account the incident angle and the maps are stored in reflFFT and transFFT
        % just for 2x2 mirrors
        % for finite angle incident angle case, they need to be properly refilled into other buffer
        % obj.RPYmap etc are used and kept for debugging purpose
        function buildMaps( obj )
 
            % load three maps
            [ obj.RPYmap, obj.RSXmap, obj.TPXmap, obj.TSYmap, specified ] = obj.prepare2x2Like;
            if ~specified
            	[ obj.RPYmap, obj.RSXmap, obj.TPXmap, obj.TSYmap ] = obj.buildDefault;
            end

            % mask is always 0, and side is only one for each
            obj.buildMirFFTmaps( [0,0,0,0], [1;2;1;2] );
            
            % mirror orientation by (optPhi, optTheta) applied here
            obj.doRemoveTiltCurvature( true );
            
        end
    end
end
