%
% OpticObj : basic optic obj, other optics, like mirror2, will be derived from this
%
%    1) create in and out ports array
%    2) assign a unique ID to this optic
%
% November 9, 2017
%    fillMissingQ updated for the case that the laser has no mode assigned.
%    This happened for a simple IMC using default setupBaseMode, because the cavity field does not go backward to laser.
%    For a mirror when an input field does not have q value assigned, Q values of transmitted or reflected field are
%    used to define Q of the input field, like MC1->MC3 is used to define laser->MC1.
%
% February 28, 2017
%    fillMissingQ is changed to do better tracing of q values
%
% September 2, 2015
%    New property ifo added, mostly to avoid using ContantTK.lam0/kLam0
%
% October 26, 2014
%   refl() is added, which just reflects by a perfect surface, i.e., out = fliplr(in)
%
% October 14, 2014
%    backScatter properties is added, which is false by default
%    This is used to manage guessObject onTheFly quantity. 
%    Without this, MC loops which couple by the back scattering goes into infinite loop.
%
% September 12, 2013  Hiro Yamamoto
%

% OpticObj is derived from matlab.mixin.Heterogeneous to have different optics
% in an array, like laser, mirror2x2, detector

classdef OpticObj < matlab.mixin.Heterogeneous & handle
    
    properties (SetObservable)
        % suffix of a name which specifies the sidem like AR or HRx
        % when it is channged from default to more meaningful one, the correspong field names are updated
        suffix
    end
    
    properties
        name = 'unnamed'

        % When apertureDef < 0, Aperture is calculated to be the beam size * abs(apertureDef)
        ApertureDef = -8;
        % extra outer space in which field is to be calculated.
        virtualAperture = 0;
        % actual aperture size
        Aperture = NaN;
        
        % when WfftUser ~= 0, 
        %   WfftUser > 0 : Wfft = WfftUser
        %   WfftUser < 0 : Wfft = wOpt * abs(WfftUser)
        % with the same Nfft. 
        % The adaptive scaling is not exactly the ratio of the beam size from mirror to mirror
        % This is to make the Wfft the same on test masses, or to make the Wfft be large on RMs without making other Wfft larger
        % These are all done in opticManager.calcDimension
        % This is physcical Wfft, and is multiplied by aliasMargin
        WfftUser = 0;
        
        % average beam size
        wOpt = NaN;
        % computational FFT window size, = physical Wfft x aliasMargin
        Wfft = NaN;
        
        % mirror center displacement
        % The HR side is used to define the direction of + axis and magnitude
        optX0 = 0;
        optY0 = 0;
        
        % this is only for mirrors, but for convenience defined here
        % invRoC is for HR, and positive for concave case
        invRoC = 0;
        % curvature on AR side is supported only for Mirror2x2, for now
        invRoCAR = 0;

        % each port is assigned a unique number, which is the field ID
        inPorts = zeros(0,0);
        outPorts = zeros(0,0);
        
        % this defines the direction of reflection and transmission
        % reflMatch(1) is the index of the output port where the reflected 
        % field goes out when in field comes to port 1
        reflMatch
        transMatch
        
        % modal model data
        % (2x2) cell array of reflection and transmission matrix
        % reflABCD(in) is the reflection from port index in to corresponding out port
        % these should be calculated in derived classes
        reflABCD = {};
        transABCD = {};
        
        % resolution
        % negative : fraction of the beam size, -16 means beam size / 16
        % poitive  : absolute value of the resolution
        resolution = -16
        
        % backscattering on HR side
        backScatter = false;
    end
    
    properties
        ifo
    end
    
    % function to replace a static variable
    methods
        % constructor
        function obj = OpticObj( ifo, name, nIn, nOut )
            
            obj.name = name;
            obj.ifo = ifo;
            
            % for each port, a new field is created and the fields index is used as the port ID
            for ii = 1 : nIn
                obj.inPorts(ii) = ...
                    ifo.fields.newField( obj.getPortName( 0, -ii ) ) ;
            end
            for ii = 1 : nOut
                obj.outPorts(ii) = ...
                    ifo.fields.newField( obj.getPortName( 0, ii ) ) ;
            end

            addlistener(obj,'suffix','PostSet',@obj.updatePortNames);
        end
           
        function set.Wfft( obj, Wfft )
            % for now, do nothing but just assign, some checking may be done in the future
            % to make Wfft variable, internal structure needs to be changed
            % Tocalculate Data in a area with Wfft, the data in other area needs to be large enough
            obj.Wfft = Wfft;
        end
        
        function Rsq = getRsq( obj )
            Rsq = obj.ifo.ffts.rmn_sq * obj.Wfft^2;
        end
        
        %% calculate reflection by a perfect reflector, in and out are both internal format
        % outField = fliplr( inField )
        function outField = refl( obj, inPortIndex, inField )
            if nargout == 0
                outID = obj.outPorts(obj.reflMatch(inPortIndex));
                obj.ifo.fields.fieldList(outID).EFFT = fliplr( inField );
%                obj.fields.fieldList(outID).status = -1; 
            else
                outField = fliplr( inField );
            end
        end
       
        function loss = opticLoss( obj, inPortID )
            if obj.isMirrorObj
                loss = obj.mirrorLoss( inPortID );
            else
                loss = nan;
            end
        end

    % HR side is always on the first half of the portIndex
        function onHR = isHRsideIndex( obj, portIndex )
            % it needs to be a mirror and the portIndex is first half
            onHR = obj.isMirrorObj & ( abs(portIndex) <= length( obj.inPorts ) / 2);
        end
        
        function bookKeeping( obj, optID )
            % let fields attached to this optic which optic they belong to
            for ip = 1 : length( obj.inPorts )
                obj.ifo.fields.fieldList( obj.inPorts(ip) ).optID = optID;
            end
            for ip = 1 : length( obj.outPorts )
                obj.ifo.fields.fieldList( obj.outPorts(ip) ).optID = optID;
            end
        end
        
        function setupMisc( ~ )
            % This is default do nothing setupMisc
            % optID assignment is done somewhere else
%             % let fields attached to this optic which optic they belong to
%             for ip = 1 : length( obj.inPorts )
%                 obj.fields.fieldList( obj.inPorts(ip) ).optID = optID;
%             end
%             for ip = 1 : length( obj.outPorts )
%                 obj.fields.fieldList( obj.outPorts(ip) ).optID = optID;
%             end
        end
        
        % if the aperture size is negative, set the aperture to be abs(apertture) x beam size
        % return the ratio of the aperture to beam size
        function winfac = calcAperture( obj, fields )
            % calculate the average of the beam sizes of all ports, which can be slighly different
            numW = 0;
            sumW = 0;
            for iport = 1 : length( obj.inPorts )
                w = fields.fieldList( obj.inPorts(iport) ).w;
                if w > 0
                    numW = numW + 1;
                    sumW = sumW + w;
                end
            end
            for iport = 1 : length( obj.outPorts )
                w = fields.fieldList( obj.outPorts(iport) ).w;
                if w > 0
                    numW = numW + 1;
                    sumW = sumW + w;
                end
            end
            % if numW = 0, something is wrong
            obj.wOpt = sumW / numW;
            if obj.ApertureDef < 0
                % if ApeetureDef < 0, use the beam size to calculate the effective aperture
                obj.Aperture = obj.wOpt * abs(obj.ApertureDef);
            elseif obj.ApertureDef == 0
                obj.Aperture = obj.wOpt * 8;
            else
                % if ApertureDef > 0, use it as is
                obj.Aperture = obj.ApertureDef;
            end
            % when there is a displacement, by optX0 and optY0, make the window size larger
            winfac = max( obj.Aperture+3*(abs(obj.optX0)+abs(obj.optY0)), obj.virtualAperture ) / obj.wOpt;
        end
        
        % if some Q of an input port is not filled, 
        % (1) if this port is connected to some source by prop, use that Q
        % (2) if transmissted port has some Q
        % (3) if reflected port has some Q
        % (4) if reverse directional port has some Q, i.e., use in for out or out for in
        % if any of the ports have Q assigned, return true
        
        function updated = fillMissingQ( obj, props )
            
            updated = false;
            
            % if all ports have good Q values, nothing to do.
            if isempty(isnan( cell2mat({obj.ifo.fields.fieldList( obj.inPorts ).invQ}))) && ...
                    isempty(isnan( cell2mat({obj.ifo.fields.fieldList( obj.outPorts ).invQ}) ))
                return;
            end
            
            % go through all inports
            for ii = 1 : length( obj.inPorts )
                sinkID = obj.inPorts(ii);
                inInvQ = obj.ifo.fields.fieldList( sinkID ).invQ;
                if isnan( inInvQ )
                    % this inport has no Q value
                    [srcID, propID] = props.getPropFrom( sinkID );
                    if srcID > 0
                        % this inport is connected to a prop
                        invQsrc = obj.ifo.fields.fieldList( srcID ).invQ;
                        if ~isnan( invQsrc )
                            % if the source of the prop connected to this inport has a valid Q, use it
                            obj.ifo.fields.fieldList( sinkID ).invQ = ...
                                MMTK.ABCD_qinv( props.getABCD(propID), invQsrc );
                            updated = true;
                        end
                    end
                end
                
                % if this is a mirror, then use reflection and transmission to fill
                if obj.isMirrorObj
                    
                    % Q value of this input port
                    inInvQ = obj.ifo.fields.fieldList( sinkID ).invQ;
                    
                    % this inport has q value so, use it to fill qs for out going fields via transmitted and reflected
                    if ~isnan( inInvQ )
                        outPortID = obj.outPorts(obj.transMatch(ii));
                        outInvQ = obj.ifo.fields.fieldList( outPortID ).invQ;
                        if isnan( outInvQ )
                            % transmissted output port is to be calculated by input Q
                            obj.ifo.fields.fieldList( outPortID ).invQ = ...
                                MMTK.ABCD_qinv( obj.transABCD{ii}, inInvQ );
                            updated = true;
                        end
                        
                        outPortID = obj.outPorts(obj.reflMatch(ii));
                        outInvQ = obj.ifo.fields.fieldList( outPortID ).invQ;
                        if isnan( outInvQ )
                            % input port has Q value, but not reflected output port
                            obj.ifo.fields.fieldList( outPortID ).invQ = ...
                                MMTK.ABCD_qinv( obj.reflABCD{ii}, inInvQ );
                            updated = true;
                        end
                        
                    % this input port has no Q assigned.
                    % Check if transmissted field or reflected field has q value assigned.
                    % if so, use that value to fill the q value of this field
                    else
                        outPortID = obj.outPorts(obj.transMatch(ii));
                        outInvQ = obj.ifo.fields.fieldList( outPortID ).invQ;
                        % if transmissted field has Q value assigned, use it to fill Q for this input field
                        if ~isnan( outInvQ )
                            obj.ifo.fields.fieldList( sinkID ).invQ = ...
                                MMTK.ABCD_qinvpRev( obj.transABCD{ii}, outInvQ );
                            updated = true;
                        
                        % if the transmitted field does not have qInv, try reflected field
                        else
                            outPortID = obj.outPorts(obj.reflMatch(ii));
                            outInvQ = obj.ifo.fields.fieldList( outPortID ).invQ;
                            % if transmissted field has Q value assigned, use it to fill Q for this input field
                            if ~isnan( outInvQ )
                                obj.ifo.fields.fieldList( sinkID ).invQ = ...
                                    MMTK.ABCD_qinvpRev( obj.reflABCD{ii}, outInvQ );
                                updated = true;
                            end
                        end
                        
                        inInvQ = obj.ifo.fields.fieldList( sinkID ).invQ;
                        % if new Q is found for this sink, fill the Q of the source field
                        if ~isnan( inInvQ )
                            [srcID, propID] = obj.ifo.props.getPropFrom( sinkID );
                            % if propagator is attached to this port
                            if srcID > 0
                                obj.ifo.fields.fieldList( srcID ).invQ = ...
                                    MMTK.ABCD_qinvpRev( obj.ifo.props.propList(propID).propABCD, inInvQ );
                            end
                        end
                    end
                end
            end
            
            if ~obj.isMirrorObj
                % if a port has valid Q for one direction, use it for the other
                % mostlikely for laser source and sink
                for ii = 1 : min( length(obj.inPorts), length(obj.outPorts) )
                    if isnan( obj.ifo.fields.fieldList( obj.inPorts(ii) ).invQ ) && ...
                            (~isnan( obj.ifo.fields.fieldList( obj.outPorts(ii) ).invQ ) )
                        % outport is known and inport is not known
                        obj.ifo.fields.fieldList( obj.inPorts(ii) ).invQ = ...
                            MMTK.reverseDir( obj.ifo.fields.fieldList( obj.outPorts(ii) ).invQ );
                        % copy this qval to the connected port
                        [srcID, propID] = obj.ifo.props.getPropFrom( obj.inPorts(ii) );
                        if srcID > 0
                            obj.ifo.fields.fieldList( srcID ).invQ = ...
                                MMTK.ABCD_qinvpRev( obj.ifo.props.propList(propID).propABCD, ...
                                    obj.ifo.fields.fieldList( obj.inPorts(ii) ).invQ );
                        end
                        updated = true;
                    elseif isnan( obj.ifo.fields.fieldList( obj.outPorts(ii) ).invQ ) && ...
                            (~isnan( obj.ifo.fields.fieldList( obj.inPorts(ii) ).invQ ))
                        % input is known and outport is not known
                        obj.ifo.fields.fieldList( obj.outPorts(ii) ).invQ = ...
                            MMTK.reverseDir( obj.ifo.fields.fieldList( obj.inPorts(ii) ).invQ );
                        % copy this qval to the connected port
                        [sinkID, propID] = obj.ifo.props.getPropTo( obj.outPorts(ii) );
                        if sinkID > 0
                            obj.ifo.fields.fieldList( sinkID ).invQ = ...
                                MMTK.ABCD_qinv( obj.ifo.props.propList(propID).propABCD, ...
                                    obj.ifo.fields.fieldList( obj.outPorts(ii) ).invQ );
                        end
                        updated = true;
                    end
                end
            end

        end
        
        % when port name is changed to better, copy that to field name
        function updatePortNames( obj, ~, ~ )
           for ii = 1 : length(obj.inPorts)
               obj.ifo.fields.setFieldName( obj.inPorts(ii), obj.getPortName( 0, -ii ) );
            end
            for ii = 1 : length( obj.outPorts )
                obj.ifo.fields.setFieldName( obj.outPorts(ii), obj.getPortName( 0, ii ) );
            end
        end
        
        % the reflection port and transmission port IDs
        % reflID = 0 when there is no reflection, -1 if the port does not belong to this optic
        % if reverse, opposite
        function reflID = getReflPort( obj, srcPortID, reverse )
            if ~exist( 'reverse', 'var' )
                reverse = false;
            end

            if reverse
                id = find( obj.outPorts == srcPortID );
            else
                id = find( obj.inPorts == srcPortID );
            end
 
            if isempty( id )
                reflID = -1;
            elseif isempty( obj.reflMatch )
                reflID = 0;
            else
                if reverse
                    for reflID = 1 : length(obj.inPorts)
                        if obj.reflMatch(reflID) == id
                            break;
                        end
                    end
                    reflID = obj.inPorts( reflID );
                else
                    reflID = obj.outPorts( obj.reflMatch(id) );
                end
            end
        end
        
        function transID = getTransPort( obj, srcPortID, reverse )
            if ~exist( 'reverse', 'var' )
                reverse = false;
            end
            
            if reverse
                id = find( obj.outPorts == srcPortID );
            else
                id = find( obj.inPorts == srcPortID );
            end
            if isempty( id )
                transID = -1;
            elseif isempty( obj.transMatch )
                transID = 0;
            else
                if reverse
                    for transID = 1:length( obj.inPorts )
                        if obj.transMatch(transID) == id
                            break;
                        end
                    end
                    transID = obj.inPorts( transID );
                else
                    transID = obj.outPorts( obj.transMatch(id) );
                end
            end
        end

        % if portID is one of the port of this objtc,
        % return > 0 for outPorts
        % return < 0 for inPorts
        % return = 0 if not in this optic
        function portIndex = getPortIndex( obj, portID )
            portIndex = find( obj.outPorts == portID, 1 );
            if isempty( portIndex )
                portIndex = -find( obj.inPorts == portID, 1 );
                if isempty( portIndex )
                    portIndex = 0;
                end
            end
            
        end
        
        function info = printInfo( obj )
            info = sprintf('name: ''%s'',  Wfft = %g,  Nfft = %d\ninPorts  = [ ', ...
                obj.name, obj.Wfft, obj.ifo.ffts.Nfft );
            for ii = 1:length( obj.inPorts )
                info = sprintf('%s%d ', info, obj.inPorts(ii));
            end
            info = sprintf('%s]  loss for each input port = [ ', info );
            for ii = 1:length( obj.inPorts )
                info = sprintf('%s%g ', info, obj.opticLoss( obj.inPorts(ii) ) );
            end
            info = sprintf('%s]\noutPorts = [ ', info);
            for ii = 1:length( obj.outPorts )
                info = sprintf('%s%d ', info, obj.outPorts(ii));
            end
            info = sprintf('%s]\n', info);
        end
        
        % if portID matches with any of the port ID, it specified the port
        % if portIndex > o, it is the index of output port
        % if portIndex < 0, it is the index of input port
        function portName = getPortName( obj, portID, portIndex, inOutName )
            if portID > 0
                match = find( obj.inPorts, portID );
                if ~isempty( match )
                    portIndex = -match;
                else
                    match = find( obj.outPorts, portID );
                    if ~isempty( match )
                        portIndex = match;
                    end
                end
            end
            if ~exist( 'inOutName', 'var' )
                inOutName = true;
            end
            if portIndex == 0
                portName = '';
            else
                portName = '-';
                if (abs(portIndex) > 0) && (length( obj.suffix ) >= abs(portIndex))
                	portName = strcat( portName, obj.suffix{abs(portIndex)} );
                end
                if portName == '-'
                    % no explicit name given
                    portName = strcat( portName, num2str(abs(portIndex)) );
                end
                inOutStr = '';
                if inOutName
                    if portIndex > 0
                        inOutStr = '-o';
                    elseif portIndex < 0
                        inOutStr = '-i';
                    end
                end
                portName = strcat( obj.name, portName, inOutStr );
            end
        end
        
        function [dx, dy] = getDXDY( obj )
            dx = obj.ifo.ffts.Dfft * obj.Wfft;
            dy = dx;
        end
        
        function [x,y,r] = getXYR( obj )
            x = obj.ifo.ffts.Xm * obj.Wfft;
            y = obj.ifo.ffts.Yn * obj.Wfft;
            if nargout == 3
                r = obj.ifo.ffts.rmn * obj.Wfft;
            end
        end

        % nothing to clean
        function cleanTemp( ~ )
        end
        
        function val = isMirrorObj( obj )
            val = isa( obj, 'MirrorObj' );
        end
        function val = isLaserObj( obj )
            val = isa( obj, 'LaserObj' );
        end
        function val = isDetectorObj( obj )
            val = isa( obj, 'DetectorObj' );
        end
        
    end
 
end