% 
% dataFileManager
%    Data is defined by a data file + formula
%    Map is defined in a plane without tilt.
%
%
% May 29, 2019
%    At the end of function [func, scatLOSS] = evalFormula( formula, map, vars, xl, yl )
%    Code has been modified for the case when there is no "func = ".
%
% November 29, 2018
%   loadMaps modified
%       if RoCFit = 0, remove up to tilt
%       if RoCFit is specified
%            if apertureFit < 0, remove piston and power
%            if apertureFit = 0, remove only piston
%            if apertureFit > 0, remove up to power
%          then the 1/newRoC = 1/curRoC + 1/RoCFit
%
% August 24, 2017
%   The power term of the formula is returned as a part of RoC.
%   This makes it possible to apply the tilt formula properly even when the formula part has power
%      del(x,y) = (x^2 / (2R*cos(tilt)) + y^2 / (2R/cos(tilt))) + A(x,y) cos(tilt)
%
% March 2, 2017
%   printInfo format change
%
% February 9, 2017
%   Variable phi = atan2(y,x) is added to evalFormula 
%   Now, r, rsq and phi are defined
%
% October 14, 2014
%   useMapXY added to loadMaps to control if the map file X,Y coordinate is to be used or FFT x-y is to be used
%
% October 10, 2014
%   evalFormula returns zero matrix if formula is empty and mat is empty, instead of 0 value
%
% August 12, 2014
%   When the formula has map( 'newX', 'newY' ), the mat is modified by using the coordinate change
%   When mat has one argument, it is interpreted as the rotation angle, i.e.,
%       map( theta ) = map( x*cos(theta)-y*sin(theta), x*sin(theta)+y*cos(theta) )
%
% July 8, 2016
%   fileInfo has now values of variables
%
% July 13, 2014
%   obj.variables for the formula are updated each time loadMaps is calloed
%
% July 8, 2014  Hiro Yamamoto
%   r = sqrt ( x.^2 + y.^2 ) added
%
% November 12, 2013 Hiro Yamamoto
%   initial version
%

classdef dataFileManager < handle
    properties
        name = 'unnamed'
        
        % format : struct('fileName','', 'formula', '', 'ApertureFit', 0, 'RoCFit', 0, 'rotMap', 0 )
        % rotMap : apply rot90( map, rotMap )
        fileList
        
        % it should be 2N cells, 'name1', val1, 'name2', var2, ...
        % These variables and variables in MirrorObj are can be complemental.
        % When only name1 is set in MirrorObj.updateMirMaps, 
        % name2=val2 in dataFileManager is used to complete the settings, 
        % i.e., the following sequence of codes are executed when the map is built
        %    name2 = val2 set in addMap, called once when the objet is created
        %    name1 = val1 set in updateMirMap, called each time the map is updated
        %    execute formula
        variables
    end
    
    methods
        %% constructor
        function obj = dataFileManager( name )
            obj.name = name;
            obj.fileList = repmat( struct('fileName','', 'formula', '', ...
                'ApertureFit', 0, 'RoCFit', 0, 'rotMap', 0 ), 1, 0 );
            obj.variables = cell(0);
        end

        function num = numFiles( obj )
            num = size( obj.fileList, 2 );
        end
        
        %% add new spec
        function addMap( obj, fileName, formula, ApertureFit, RoCFit, rotMap, varargin )
            % set default values for the main four variables
            eval( UtilTK.setDEFVALS('formula', '', 'ApertureFit', 0, 'RoCFit', 0, 'rotMap', 0) );
            % these are variables for the formula
            if ~UtilTK.validArgList( varargin{:} )
                error( 'Variable definition incorrect' );
            end
            id = obj.numFiles + 1;
            obj.fileList(id) = struct('fileName', fileName, 'formula', formula, ...
                'ApertureFit', ApertureFit, 'RoCFit', RoCFit, 'rotMap', rotMap );
            obj.variables{id} = varargin;
        end
        
        %% load multiple data files and add them together with formula applied, 
        %
        % If RoCFit is not 0, it is used to calculate invRoC even when ApertureFit is 0.
        % tilt and curvature removal works if aperture is > 0.
        % tilt and curvature removal applies only to map data, not formula.
        %
        % evalFormula uses (x,y) values of the map data, if there were any. 
        % If no map is defined, it uses nominal grid points of the optic.
        % So loadMaps can get different answers when the formula is defined 'map*0 + x' and just 'x' without map
        %
        % If useMapXY is true, use the x-y coordinates of the data file, and return those coordinates in xMap, yMap
        %
        function [map, invRoC, xMap, yMap] = loadMaps( obj, thisOptic, useMapXY )           
            % default is to use optics coordinate, not data map's
            if ~exist('useMapXY', 'var' )
                useMapXY = false;
            end
            
            % no curvaure specified
            invRoC = 0;
            
            if obj.numFiles > 0
                % prepare coordinate system using optic's data
                % optic coordinate
                [xOpt, yOpt] = thisOptic.getXYR;
                % empty map
                map = zeros( size(xOpt) );
                xOpt = xOpt(1,:);
                yOpt = yOpt(:,1);

                % go through all maps specified for this optic
                for id = 1 : obj.numFiles
                    if obj.fileList(id).RoCFit == 0
                        % if RoCFit = 0, don't remove power, but if Aperture is specified, remove up to tilt
                        removeZ = 1;
                    else
                        % curvature ia measured, so remove the curvature and add later
                        if obj.fileList(id).ApertureFit < 0
                            % remove only power
                            removeZ = 3;
                        elseif obj.fileList(id).ApertureFit == 0
                            removeZ = 0; % ApertureFit = 0 and removeZ = 0 mean to remove piston 
                        else
                            removeZ = 2;
                        end
                        % lenses can be added ...
                        invRoC = invRoC + 1/obj.fileList(id).RoCFit;
                    end
                    
                    % update variables (defined in addMap) using the thisOptic.variables values (reset in updateMirMap)
                    % option 1 : only replace. addMap should have registered all necessary variables
                    %    changed from 2 to 1 to suppress warning...
                    obj.variables{id} = ...
                        UtilTK.updateArgList( obj.variables{id}, thisOptic.variables, 1 );
                    
                    % when there is no data map,
                    if isempty( obj.fileList(id).fileName )
                        % just use the formula, and RoC removal is not applied
                        mapOne = obj.evalFormula( obj.fileList(id).formula, '', ...
                            obj.variables{id}, xOpt, yOpt );
                        xMap0 = xOpt;
                        yMap0 = yOpt;
                    else
                        % when data file is avaibale, load the data
                        % if RoC is provided and apertureFit > 0, power/tilt term is removed
                        [ mapOne, xMap0, yMap0 ] = ...
                            loadOneDataFile( obj.fileList(id).fileName, ...
                            abs(obj.fileList(id).ApertureFit), removeZ, thisOptic.Aperture );
                        mapOne( isnan( mapOne ) ) = 0;
                        % then formula is applied
                        mapOne = obj.evalFormula( obj.fileList(id).formula, mapOne, ...
                            obj.variables{id}, xMap0, yMap0 );
                    end
                    % apply mirror map rotation before tilting
                    mapOne = rot90( mapOne, obj.fileList(id).rotMap );
                    switch mod( obj.fileList(id).rotMap, 4 )
                        case 0   % no rotation
                            xMap = xMap0;
                            yMap = yMap0;
                        case 1   %  90 degree rotation
                            xMap = yMap0;
                            yMap = -xMap0(end:-1:1);
                        case 2   % 180 degree
                            xMap = -xMap0(end:-1:1);
                            yMap = -yMap0(end:-1:1);
                        case 3   % 270 degree
                            xMap = -yMap0(end:-1:1);
                            yMap = xMap0;
                    end
                        
                    % to make is consistent, don't add multiple data files
                    % also no coordinate change and no lowPass filter needed
                    if useMapXY
                        map = mapOne;
                        break;
                    end

                    % convert the grid size to the one defined by the FFT simulation and add
                    % when the optic is tilted and to be rorated, one original bin size shrinks by cos(incAngle)
                    % measured in the incoming field coordinate, or the mirror size becomes cos(theta) smaller                    
                    mapOne = UtilTK.lowPassMap( mapOne, xMap, yMap, xOpt, yOpt );
                    mapOne( isnan( mapOne ) ) = 0;
                    map = map + mapOne;
                end
                
                if ~useMapXY
                    xMap = xOpt;
                    yMap = yOpt;
                end
            else
                % to save memory
                map = []; % zeros(size(thisOptic.getXYR));            
            end
        end

        %% print information about data files
        function info = fileInfo( obj )
            info = '';
            for id = 1 : obj.numFiles
                if isempty( obj.fileList(id).fileName )
                    info = sprintf('%s  [%d] no file specified :', info, id );
                else
                    info = sprintf('%s  [%d] file name = ''%s''\n', info, id, obj.fileList(id).fileName );
                end
                if ~isempty( obj.fileList(id).formula )
                    info = sprintf('%s      formula = ''%s''\n', info, obj.fileList(id).formula );
                end
                if obj.fileList(id).ApertureFit ~= 0
                    info = sprintf('%s      aperture = %g :', info, obj.fileList(id).ApertureFit );
                end
                if obj.fileList(id).RoCFit ~= 0
                    info = sprintf('%s  RoC = %g :', info, obj.fileList(id).RoCFit );
                end
                if obj.fileList(id).rotMap ~= 0
                    info = sprintf('%s  map rotated by = %d degree :', info, obj.fileList(id).rotMap*90 );
                end
                info = sprintf('%s\n', info );
                varLeng = floor( length( obj.variables{id} ) / 2 );
                if varLeng > 0
                    info = sprintf('%s  [*] values of %d variables\n', info, varLeng );
                    for iv = 1 : varLeng
                        info = sprintf('%s      (%d) %s = %g\n', ...
                            info, iv, obj.variables{id}{iv*2-1}, obj.variables{id}{iv*2});
                    end
                end
            end
            if ~isempty( info )
                info = sprintf( 'File lists in ''%s''\n%s', obj.name, info );
            end
        end
        
    end
    
    methods (Static)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        %	function [func,scatLOSS] = evalFormula( formula, map, vars, xl, yl )
        %       formula can use x, y, r, rsq, phi and map
        %       if map is provided, [x,y] = meshgrid( xl, yl )
        %       otherwiae, [x,y] = getXYR, i.e., grids for FFT on this optic
        %       vars and defVars are cells arrays which are 'name1', val1, 'name2', var2, etc
        %       defVars are spefieid in dataFileManager.addMap and 
        %       vars are defined in MirrorObj.buildMirrorMaps
        %
        % 0) defVars updated each time updateMirMaps used
        % 
        % 1) If the formula has an expression "func = ", 
        % then it is assumed func value is defined in the formula.
        %     xnew = x * cos(theta) - y * sin(theta); func = xnew.^2;
        %
        % 2) otherwise, the value of this formula is used as func value.
        %     ( x * cos(theta) - y * sin(theta) ).^2
        %
        % 3) if there is an expression map( something ), it is assumed to be a coordinate change
        %    map( newX, newY ) : newX/newY can be values or formula, like x*cos(theta)-y*sin(theta)
        %    map( theta ) : the map is rotated by theta
        %
        % 4) a string MAP is treated as map, not interpred as the coordinate change.
        %
        %
        % 5) if formula is emoty and map is empty, zero matrix is returned, instead of 0 value
        %
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        function [func, scatLOSS] = evalFormula( formula, map, vars, xl, yl )
            scatLOSS = 0;
            if isempty( strtrim(formula) )
                if isempty( map )
                    func = zeros( length(yl), length(xl) );
                else
                    func = map;
                end
            else
                % formula exists
                % evaluate the formula using (x, y, map)
                % (x,y) can be used in the formula, so keep them
                [x,y] = meshgrid( xl, yl );
                rsq = x.^2 + y.^2;
                r = sqrt( rsq );
                phi = atan2(y,x);

                % map can be used in formula
                if isempty( map )
                	map = zeros( size(x) );
                end
                
                for iv = 1 : floor(size(vars,2)/2)
                    oneVal = sprintf('%s = vars{1,iv*2};', vars{1,iv*2-1});
                    eval( oneVal );
                end

                % if formula has mat with two arguments, they are used to redefine the coordinate
                % like 'cos*x - sin*y', 'sin*x + cos*y' to rotate mirrors
                % formula is the change of coordinates, so this formula rotates obmirror by -THETA
                [formula, newX, newY] = dataFileManager.modifymap( formula );
                if ~isempty( newX ) && ~isempty( newY )
                    % newX and newY are used in formula
                    newX = eval( newX );
                    newY = eval( newY );
                end
                
                % replace MAP by map so that MAP( r.^2 < 0.5^2 ) = 0; is not modified by modifymap
                indmap = regexp(formula, 'MAP');
                if ~isempty( indmap )
                    for ii = 1 : length(indmap)
                        formula( indmap(ii):indmap(ii)+2 ) = 'map';
                    end
                end
                
                % x, y and r can be used in eval(formula)
                % if no express like func = formula, prepend "func =" to the last expression.
                formula = strtrim( formula );
                if isempty( regexp( formula, 'func\ *=', 'once' ) )
                    % if there is ';' at the end, remove it
                    if formula(end) == ';'
                        formula = formula(1:end-1);
                    end
                    semiCol = find( formula == ';', 1, 'last' );
                    % if there is a ';', here to the end is the expression for formula
                    if isempty(semiCol)
                        formula = strcat( 'func = ', formula );
                    else
                        formula = strcat( formula(1:semiCol), 'func = ', formula(semiCol+1:end) );
                    end
                end
                eval( strcat(formula, ';') );                                                
             end
        end
        
        function [mapped, newX, newY] = modifymap( formula )
            % mapped = pickmap( func )
            % When func is 'something map('xnew','ynew') something'
            % returns mapped = 'something UtilTK.mapping(x,y,map,xnew,ynew) something'
            % When there is only one argument, it is assumed to be the angle of rotation
            % If func is something map('phi') something, it returns
            % with xnew = x cos(phi) - y sin(phi), ynew = x sin(phi) + y cos(phi)
                     
            newX = '';
            newY = '';

            % mat( something )
            pos0 = regexp(formula, '\<\<map\(.*\)', 'ONCE');
            
            if ~isempty( pos0 )
                % map( something )
                [argList, endPos] = UtilTK.pickArg( formula( pos0+3 : end ) );

                % only one value in map(), so it is angle
                if length( argList ) == 1
                    
                    theta = strtrim( argList{1} );
                    if ~isempty( theta )
                        newX = strcat( 'x*cos(', theta, ')-y*sin(', theta, ')');
                        newY = strcat( 'x*sin(', theta, ')+y*cos(', theta, ')');
                    end
                elseif length( argList ) == 2
                    % there are two arguments, mat( newX, newY )
                    newX = strtrim( argList{1} );
                    newY = strtrim( argList{2} );
                end
            end
            
            if isempty( newX ) || isempty( newY )
                mapped = formula;
            else
                % mapped = 'something UtilTK.mapping(x,y,map,xnew,ynew) something'
                mapped = strcat(formula(1:(pos0-1)), 'UtilTK.mapping(x, y, map, newX, newY)', formula(pos0+endPos+3:end));
            end
        end
        
%         % check if varargin for variable data passing makes sense
%         function ok = checkVarFormat( varargin )
%             ok = false;
%             if mod(size(varargin,2),2) ~= 0
%                 return;
%             end
%             for iv = 1 : size(varargin,2)/2
%                 if ~ischar( varargin{iv*2-1} )
%                     return;
%                 end
%             end
%             ok = true;
%         end
    end
end