%
% [amap, xl, yl] = loadOneDataFile( ( fileName, ApertureFit [0], removeZ [2], mirrorAperture[0] )
% Load data from a file and return data(NY,NX), xl(1,NX) and yl(NY,1)
% This function cashes loaded data for faster access for repeated access.
%
% Datafile format supported are (as of September 2015)
%    Metropro data, Wyko OPD, LMA col, ASCII GINT, hdf5 and SIS_Nxxx_Wyyy
%
% if mirrorAperture > 0 is provided, data out of this aperture is filled by NaN
%
% If ApertureFit is non zero, zernite terms are removed.
% If ApertureFit < 0, full mirror size is used as the aperture.
% 
%       removeZ = 0 : piston
%               = 1 : up to tilt
%               = 2 : up to power <= default
%               = 3 : piston and power, no tilt
%
% loadOneDataFile('CLEARCASH') or loadOneDataFile without argument removes all chashe data
% loadOneDataFile('LIST') shows the list of cash data
%
% apertureFit > 0 : removeZ = 0 (piston), 1 (piston, tilt), 2 (piston, tilt, power), 3 (piston, power)
% apertureFit = 0 : only removeZ = 0 to remove piston
% apertureFit < 0 (any negative) : aperturFit = mirrorSise, removeZ = 0, 1, 2, 3
%
%
% To support a new format data, add the data loader of that format in this function
% 
% February 14, 2019
%   new data file format, loadMATdata, added.
%   Save a data and coordinate in a matlab file, and this loads in that data.
%   X and Y coordinate name starts with X and Y, and the value name can be anything so long as it does not start with X nor Y.
%
% November 29, 2018
%   apertureFit and removeZ usage modified.
%   default (apertureFit = 0, removeZ = 2) just loads the data, without correction
%   hdf5 for TCS data supported
%
% July 6, 2017
%   ApertureFit < 0 case uses the mirror size as the ApertureFit value
%
% September 3rd, 2015
%   ASCII GINT format added
%
% October 19th, 2014
%   LMA col file loader is changed to loadLMAcol, which correctly handled off-centered maps
%
% October 6th, 2014
%   Default values of ApertureFit and removeZ are explicitly set at the beginning.
%
% August 5th, 2014
%   removeTiltPower is replaced by zernikeCoef
%
% January 10, 2014  HY
%       1) Wyko OPD format added
%       2) NaNs are removed by average values of surrounding valid data
%

function [amap, xl, yl] = loadOneDataFile( fileName, ApertureFit, removeZ, mirrorAperture )

if nargin == 0 || strcmp( fileName, 'CLEARCASH' )
    clearCash;
    return;
end

if strcmp( fileName, 'LIST' )
    listCash;
    return;
end

eval( UtilTK.setDEFVALS('ApertureFit', 0, 'removeZ', 2, 'mirrorAperture', 0 ) );

[amap, xl, yl] = findCashed( fileName, mirrorAperture, ApertureFit, removeZ );

if ~isempty( amap )
    return;
end

%% modify this section to add a new data format
if ~isempty( regexpi( fileName, '.col$' ) )
    % LMA col file
    [amap, xl, yl] = loadLMAcol( fileName );
    %    [amap, xl, yl] = FT_read_virgo_map_FOGP( fileName );
    
elseif ~isempty( regexpi( fileName, '.mat$' ) )
    % matlab data file
    [amap, xl, yl] = loadMATdata( fileName );
    
elseif ~isempty( regexpi( fileName, '.hdf5$' ) )
    [amap, xl, yl] = loadHDF5( fileName );
    
elseif ~isempty( regexpi( fileName, '.gnt$' ) )
    % ASCII GINT file
    [amap, xl, yl] = loadASCIIGNT( fileName );
    
elseif ~isempty( regexpi( fileName, '.opd$' ) )
    % Wyko OPD file
    [amap, wavelength, aspect, pixel] = readopd( fileName );
    amap( abs(amap) > 1e10 ) = NaN;
    amap = wavelength * amap';
    [ysize, xsize] = size( amap );
    xl = (-xsize/2+0.5:xsize/2-0.5)* pixel;
    yl = (-ysize/2+0.5:ysize/2-0.5)* pixel * aspect;
elseif  ~isempty(regexp( fileName, '_N\d', 'once')) &&  ~isempty(regexp( fileName, '_W\d', 'once'))
    % SIS *_W600_N512 ascii data
    [amap, xl, yl] = loadSISData( fileName );
    % elseif ~isempty( regexpi( fileName, '.txt$' ) )
    %     % simple text data, either 3 column [x,y,z] or two column [r,z]
    %     % for two column case, if r ranges from negative to positive,
    %     % it is assumed to be data along x axis and 2d data is generated by interpolating long phi
    %     [amap, xl, yl] = laodTEXTData( fileName );
else
    % for the MetroPro data, if ApertureFit is given, then remove up to curvature
    % this map is to be used with RoCFit, which is the reference used to measure the surface
    [ amap, xl, yl ] = loadMetroProData( fileName );
end

%%
if isempty( amap )
    fprintf( 1, 'File ''%s'' could not be loaded\n', fileName );
else
    if mirrorAperture > 0
        % row vector + column vector is not supported in older version, so use meshgrid
        [xg,yg] = meshgrid( xl, yl );
        amap( xg.^2+yg.^2 > mirrorAperture^2/4 ) = NaN;
    end
    % replace NaNs in the mirror area by averages of surrounding pixels.
    amap = fillNaN( amap );
    
    if ApertureFit == 0
        if removeZ == 0
            ind = find( ~isnan(amap) );
            amap(ind) = amap(ind) - mean( amap(ind) );
        end
    else
        A = ApertureFit;
        if A < 0
            if isnan(amap(1,1))
                ind = find( ~isnan(amap) );
            else
                ind = find( amap ~= amap(1,1) );
            end
            [Xm, Yn] = meshgrid(xl, yl);
            A = min( max(Xm(ind)) - min(Xm(ind)), max(Yn(ind)) - min(Yn(ind)) ) * 0.99;
        end
        % nmList = 2 to remove up to tilt, nmList = 3 to remove up to power
        % just to avoid unknown case
        if removeZ == 0  % 0 is to remove piston
            amap = zernikeCoef( amap, xl, yl, A, 0 );
        elseif removeZ == 1 || removeZ == 2 % removeZ+1 = 2 : up to titl x and tilt y, = 3 : up to power
            amap = zernikeCoef( amap, xl, yl, A, removeZ + 1 );
        elseif removeZ == 3 %piston and power
            amap = zernikeCoef( amap, xl, yl, A, [0,0; 2,0] );
        end
    end
    
    addToCash( fileName, mirrorAperture, ApertureFit, removeZ, amap, xl, yl );
end
end

function clearCash
global FOGPRIME13DATA;
FOGPRIME13DATA = [];
end

function listCash
global FOGPRIME13DATA;

nCash = size( FOGPRIME13DATA, 2 );
if nCash == 0
    fprintf(1, 'no data file cash\n');
else
    fprintf(1, '%d cashe(s) in FOGPRIME13DATA global data\n', nCash);
    for id = 1:nCash
        fprintf(1,'%d) ''%s'' [%d,%d]\n', id, FOGPRIME13DATA{1,id}, ...
            size(FOGPRIME13DATA{5,id},1), size(FOGPRIME13DATA{5,id},2) );
    end
end
end

function [amap, xl, yl] = findCashed( fileName, mirrorAperture, ApertureFit, removeZ )

global FOGPRIME13DATA;

if isempty( FOGPRIME13DATA )
    FOGPRIME13DATA = cell(7,0);
end

amap = [];
xl = [];
yl = [];

for id = 1 : size( FOGPRIME13DATA, 2 )
    if strcmp( FOGPRIME13DATA{1,id}, fileName ) && FOGPRIME13DATA{2,id} == mirrorAperture && ...
            FOGPRIME13DATA{3,id} == ApertureFit && FOGPRIME13DATA{4,id} == removeZ
        amap = FOGPRIME13DATA{5,id};
        xl = FOGPRIME13DATA{6,id};
        yl = FOGPRIME13DATA{7,id};
    end
end

end

function addToCash( fileName, mirrorAperture, ApertureFit, removeZ, amap, xl, yl )

global FOGPRIME13DATA;

id = size( FOGPRIME13DATA, 2 ) + 1;
FOGPRIME13DATA{1,id} = fileName;
FOGPRIME13DATA{2,id} = mirrorAperture;
FOGPRIME13DATA{3,id} = ApertureFit;
FOGPRIME13DATA{4,id} = removeZ;
FOGPRIME13DATA{5,id} = amap;
FOGPRIME13DATA{6,id} = xl;
FOGPRIME13DATA{7,id} = yl;

end