%
% opticManager
%
%   maintain optics
%
%   optList is a list of propagators
%
%   obj = opticManager( fields )
%
%   id = addMirror( obj, name, invRoC, Aperture, Thickness, T, Lhr, Lar )
%   id = addBS    ( obj, name, invRoC, Aperture, Thickness, T, Lhr, Lar, incAngle )
%
%   reflID = findReflPort( obj, inPortID )
%   transID = findTransPort( obj, inPortID )
%
%   portID = findPortID( obj, optID, portIndex )
%       portIndex > 0 for output port, < 0 for input port
%
%   reflection
%       outField = refl( obj, inPortID, inField )
%
%   transmission
%       outField = trans( obj, inPortID, inField )
%
%
% April 3rd, 2019
%   c0 for laser beam displacement is renamed to dxyz.
%
% January 20, 2019
%   function dat = lossSummary( obj, mirrorID )
%      loss around the mirrorID or all mirrors     
%
% August 8, 2018
%   findOptic has a second argument, (portID, tryUsePortIDfrom)
%   If this exists and true, usePortIDfrom of field portID is also checked to find the optic and portID
%   This is for fields which are associted with ports with regular fields.
%   When refl or trans are used to create a new field to a port which have regular field associtaed.
%   fieldObj.usePortIDfrom is the field ID of this regular field, and this alloww
%
% May 18, 2018
%   findABCD works for reflection, as well as transmission
%
% September 8, 2017
%   function [qInv, rootPort, Nfft] = getSourceFieldMode( obj ) added
%   This returns positive rootPort if a source port is found
%   Returns positive Nfft if source field from other IFO is supplied (i.e., Nfft is given externally)
%   qInv is not nan if provided
%
%   In function [portID, optID] = findLaser( obj ), both portID and optID are set to 0 if failed
%
% April 26, 2017
%   in getDeapth, thicknessInSubs( obj, n, correctOffcentering ) call sets correctOffcentering = false
%   because the distance between BS and ITM takes this effect into account.
%
% April 25, 2017
%   getHRRoC returns thetaX and thetaY
%
% March 28, 2017
%   getDeapth has second argument added which selects if physical * nRef or physica / nRef are calculated
%
% March 13, 2017
%   getGouyMul did not return value
%
% March 4, 2017
%   calcDimension is modified so that Nfft is properly calculated when WfftUser is specified.
%
% May 8, 2016
%  addLaser has new variables, 'nvec' (pointing vector) and 'c0' (beam center)
%
% September 2, 2015
%   addLaser accepts lamCR variable to set the CR wave length, which is
%   stored in fieldManager. kCR is filled as well by hand.
%
% May 15, 2015
%   In getDeapth, nRef was used wrong, and propL, longitudinal propagation
%   length calculated in PropObj.setupMisc, was wrong. It make the SB
%   resonance incorrect. Now, getDeapth calls MirrorObj.thicknessInSubs.
%
% October 26, 2014
%   handle of FFTIFO itself is saved in all manager objects,
%   instead of handle of multiple manager handles
%
% October 13, 2014
%   function id = addMirrorGeneric( obj, MirrorObjName, name, varargin )
%   added to replace addMirror2x2 and addMirror4x4, place class name string in MirrorObjName
%
% October 10, 2014
%    For function findPortID( obj, optID, portIndex ), optID for LaserObj and DetectorObj handled correctly
%
% August 24, 2014
%   addLaser stop using inputParser, instead uses UtilTK.setVars
%
% July 31, 2014
%   addLaser accepts field type (LG or HG) and their indexes
%
% March 29, 2014    HY
%   WfftUser is implemented. Wfft is forced to be this value, without affecting other optics.
%   Internal Wfft is 2xWfftUser
%

classdef opticManager < handle
    
    properties
        ifo
        
        optList     % array of optics
        
        winFac      % physical window size > winFac * beam size
    end
    
    methods
        % constructors
        function obj = opticManager( ifo )
            obj.ifo = ifo;
            obj.optList = OpticObj.empty(0);
        end
        
        function Rsq = getRsq( obj, optID )
            % Rsq = getRsq( optID )
            % = x^2 + y^2 in (N/2 x N/2) matrix
            Rsq = obj.optList(optID).getRsq;
        end
        
        function xList = getXlist( obj, optID )
            xList = obj.ifo.ffts.Xm(1,:) * obj.getWfft( optID );
        end
        
        function Wfft = getWfft( obj, optID )
            Wfft = obj.optList( obj.findNamedOptic(optID) ).Wfft;
        end
        
%        function dat = lossSummary( obj, mirrorID )
%            dat(id,5,2) : dat(1:id) is for id'th mirror
%            dat(  ,portID,1) : optID of input ports portID = 1~2 or 1:4, mirrorID for portID = 5
%            dat(  ,portID,2) : loss for the port for portID = 1~4 or the total loss for portID = 5
%            dat(1,1:4,1) : portID of the first optic, like PRM-HR-i or BS-HRx-i
%            dat(1,1:4,2) : loss for the field which goes to the input port
%            dat(1,5,1) : mirrorID of the first optic, PRM or ITM or like
%            dat(1,5,2) : total loss of power which goes into this mirror
        function dat = lossSummary( obj, mirrorID )
            dat = 0;
            if exist('mirrorID', 'var')
                mirrorID = obj.findNamedOptic(mirrorID);
                if obj.optList(mirrorID).isMirrorObj
                    dat = obj.optList( mirrorID ).lossSummary;
                end
            else
                dat = zeros(0,5,2);
                for mid = 1 : length( obj.optList )
                    if obj.optList(mid).isMirrorObj
                        dat( size(dat,1) + 1, 1:5, 1:2 ) = obj.optList( obj.findNamedOptic(mid) ).lossSummary;
                    end
                end
            end
        end
        
        function dat = lossSummaryN( obj, mirrorID )
            if exist('mirrorID', 'var')
                mirrorID = obj.findNamedOptic(mirrorID);
                if obj.optList(mirrorID).isMirrorObj
                    dat = obj.optList( mirrorID ).mirrorLossN;
                end
            else
                for mid = 1 : length( obj.optList )
                    if obj.optList(mid).isMirrorObj
                        datN = obj.optList( obj.findNamedOptic(mid) ).mirrorLossN;
                        if ~exist( 'dat', 'var' )
                            dat = datN(1);
                            for il = 2 : length( datN )
                                dat( 1, size(dat,2) + 1) = datN(il);
                            end
                        else
                            for il = 1 : length( datN )
                                dat( 1, size(dat,2) + 1) = datN(il);
                            end
                        end
                   end
                end
            end
        end
        
        function [qInv, rootPort, Nfft] = getSourceFieldMode( obj )
            qInv = nan;
            Nfft = 0;
            [rootPort, rootOptic] = obj.findLaser;
            if rootOptic > 0
                srcFieldID = obj.optList(rootOptic).srcFieldID;
                qInv = obj.optList(rootOptic).laserInvQ;
                if srcFieldID > 0
                    Nfft = obj.optList(rootOptic).srcNfft;
                elseif isnan(qInv)
                    rootPort = 0;
                end
            end
        end
        
        % find the optic to which portID is attached
        % if portID is an output port, find
        %       the input portID which transmisstes to the output port and
        %       the output port this input port reflects to
        % if portID is an input port, find
        %       the output port to which the input port transmits to and
        %       the the input port which reflects to this output port
        %
        function [inPort, outPort] = findInOutPort( obj, portID )
            inPort = 0;
            outPort = 0;
            [optID, portIndex] = obj.findOptic( portID );
            if optID > 0
                if portIndex > 0
                    % portID was an output port
                    % find input pott which transmits to this port
                    % reverse it, so find the source whcih transmits to this port
                    inPort = obj.findTransPort( portID, true );
                    outPort = obj.findReflPort( inPort );
                elseif portIndex < 0
                    % portID was an input port
                    % find the output port to which portID transmits to
                    % and find input which reflects to the output port
                    % following lines were wrong !!!
                    outPort = obj.findTransPort( portID );
                    inPort = obj.findReflPort( outPort, true );
                end
            end
        end
        
        function loss = opticLoss( obj, inPorts )
            loss = nan(1, length( inPorts ) );
            for id = 1 : length( inPorts )
                optID = obj.findOptic( inPorts(id) );
                loss(id) = obj.optList( optID ).opticLoss( inPorts(id) );
            end
        end
        
        % for 2x2, 1 is on HR, 2 is on AR
        % for 4x4, 1,2 is on HR, 3,4 is on AR
        function yesOnAR = isOnARSidePortID( obj, portID )
            [optID, portIndex] = obj.findOptic( portID );
            yesOnAR = abs(portIndex) > length( obj.optList(optID).outPorts ) / 2;
        end
        
        % reflection
        function outField = refl( obj, inPortID, inField )
            [optID, portIndex] = obj.findOptic( inPortID );
            if exist('inField', 'var')
                if nargout == 0
                    obj.optList(optID).refl( abs(portIndex), inField );
                else
                    outField = obj.optList(optID).refl( abs(portIndex), inField );
                end
            else
                if nargout == 0
                    obj.optList(optID).refl( abs(portIndex), ...
                        obj.ifo.fields.fieldList(inPortID).EFFT);
                else
                    outField = obj.optList(optID).refl( abs(portIndex), ...
                        obj.ifo.fields.fieldList(inPortID).EFFT);
                end
            end
        end
        
        function [isHR, optID] = isHRside( obj, portID )
            % either non-mirror port (laser or detector) or HR side of a mirror
            [optID, portIndex] = obj.findOptic( portID );
            % if AR side, calculate the transverse deapth
            isHR = (~obj.optList(optID).isMirrorObj) || obj.optList(optID).isHRsideIndex( portIndex );
        end
        
        % if portID is on the AR side, return thicness / cosAngInSub * nRef
        % n =  1 : depth = physical distance * nRef, this is longtitudinal optical length
        % n = -1 : depth = physical distance / nRef, this is transversal optical length
        function deapth = getDeapth( obj, portID, n )
            if ~exist( 'n', 'var' )
                n = 1;
            end
            [isHR, optID] = obj.isHRside( portID );
            if isHR
                deapth = 0;
            else
                % this takes care of the change of the propagation distance
                % due to the off-centering of the beam on the AR side
                deapth = obj.optList(optID).thicknessInSubs(n);
            end
        end
        
        function Nfft = bookKeeping( obj )
            for optID = 1 : length( obj.optList )
                obj.optList(optID).bookKeeping( optID );
            end
            Nfft = obj.calcDimension;
        end
        
        % transmission
        function outField = trans( obj, inPortID, inField )
            [optID, portIndex] = obj.findOptic( inPortID );
            if exist('inField', 'var')
                if nargout == 0
                    obj.optList(optID).trans( abs(portIndex), inField );
                else
                    outField = obj.optList(optID).trans( abs(portIndex), inField );
                end
            else
                if nargout == 0
                    obj.optList(optID).trans( abs(portIndex), ...
                        obj.ifo.fields.fieldList(inPortID).EFFT);
                else
                    outField = obj.optList(optID).trans( abs(portIndex), ...
                        obj.ifo.fields.fieldList(inPortID).EFFT);
                end
            end
        end
        
        function [map, xList] = getHRmap( obj, optID, removeRoC )
            % [map, xList] = getHRmap( optID, removeRoC )
            % return HR map with or without nominal curvature removed
            if ~exist( 'removeRoC', 'var' )
                removeRoC = true;
            end
            % convert name to numeric
            optID = obj.findNamedOptic( optID );
            map = obj.optList( optID ).getHRmap( removeRoC );
            xList = obj.getXlist( optID );
        end
        
        function [RoC, thetaX, thetaY, piston] = getHRRoC( obj, optID, radius )
            % convert name to numeric
            optID = obj.findNamedOptic( optID );
            [RoC, thetaX, thetaY, piston] = obj.optList( optID ).getHRRoC( radius );
        end
        
        % wOpt, calculate average beam size on each optic
        % calculate the aperture size for those whose physical size is very large by using the wOpt
        % calculate the Wfft as wOpt * winfac * wind * FFTTK.aliasMargin (=2)
        function Nfft = calcDimension( obj )
            % using field's q value, calculate the aperture size for those
            % whose physical size does not matter, like RM mirrors
            % winfac = 8 FFT window is at least 8 times of the beam size,
            % which is the limit set for RM etc.
            obj.winFac = abs(obj.optList(1).ApertureDef);
            for iopt = 1 : length( obj.optList )
                obj.winFac = max( obj.winFac, obj.optList(iopt).calcAperture( obj.ifo.fields ) );
            end
            % this is the common scaling factor used to define Wfft for each optic
            Nfft = 0;
            for iopt = 1 : length( obj.optList )
                % calculate Wfft with standard method
                Wfft = obj.winFac * obj.optList(iopt).wOpt;
                % find resolution
                if obj.optList(iopt).resolution > 0
                    res = obj.optList(iopt).resolution;
                else
                    res = obj.optList(iopt).wOpt / abs(obj.optList(iopt).resolution);
                end
                
                ratio = 1;
                % if WfftUser is provided, use that Wfft
                if obj.optList(iopt).WfftUser ~= 0
                    if obj.optList(iopt).WfftUser > 0
                        newWfft = obj.optList(iopt).WfftUser;
                    else
                        newWfft = abs(obj.optList(iopt).WfftUser) * obj.optList(iopt).wOpt;
                    end
                    ratio = newWfft / Wfft;
                    % when the beam is flat, w = inf, WfftUser will be much smaller than beam size
                    % so no check if ratio < 1
                    if ratio > 2 % || ratio < 0.8
                        fprintf(1,'> For mirror ''%s'', Wfft set by user is off by factor %g from the optimal Wfft, %g\n', ...
                            obj.optList(iopt).name, ratio, Wfft );
                    end
                    
                    % new Wfft is larger than old setting
                    if ratio >= 1
                        % if the apserture is defined by relative scaling, use larger Wfft
                        if obj.optList(iopt).ApertureDef < 0
                            obj.optList(iopt).Aperture = newWfft;
                        end
                    else
                        if obj.optList(iopt).Aperture > newWfft
                            fprintf(1,'> For mirror ''%s'', Wfft set by user, %g, is smaller than the aperture zize (%g)\n', ...
                                obj.optList(iopt).name, newWfft, obj.optList(iopt).Aperture );
                        end
                    end
                    Wfft = newWfft;
                end
                
                Wfft = Wfft * FFTTK.aliasMargin;
                % calculate Nfft using standard Wfft and resolution
                Nfft = max( Nfft, Wfft/res );
                
                obj.optList(iopt).Wfft = Wfft;
                
                % if inPort is connected to a propagator with 0 distance, set the Wfft of that optic to be the same as this
                % this is the case of laser source to input mirror. Laser source Wfft is calculated using field info
                if ratio ~= 1
                    for ip = 1 : length( obj.optList(iopt).inPorts )
                        [srcID, propID] = obj.ifo.props.getPropFrom( obj.optList(iopt).inPorts(ip) );
                        if srcID > 0 && obj.ifo.props.getPropLeng( propID ) == 0
                            obj.optList(obj.findOptic(srcID)).Wfft = Wfft;
                        end
                    end
                end
                
            end
            Nfft = 2^ceil( log(Nfft)/log(2) );
        end
        
        function setupMisc( obj )
            % now all apertures are setup
            % let do any necessary work for each optic
            % 1) inform fields attached to the optic which they are attached to
            % 2) various transmission and reflection maps ...
            % 3) misc
            for iopt = 1 : length( obj.optList )
                obj.optList(iopt).setupMisc;
            end
        end
        
        % find laser source
        function [portID, optID] = findLaser( obj )
            for iopt = 1 : length( obj.optList )
                if isa( obj.optList(iopt), 'LaserObj' )
                    optID = iopt;
                    portID = obj.optList(iopt).outPorts(1);
                    return;
                end
            end
            optID = 0;
            portID = 0;
        end
        
        % find reflection or transmission from inID to outID
        function outABCD = findABCD( obj, inID, outID )
            if obj.findTransPort( inID ) == outID
                outABCD = obj.findTransABCD( inID );
            elseif obj.findReflPort( inID ) == outID
                outABCD = obj.findReflABCD( inID );
            else
                outABCD = eye(2,2);
            end
        end
        
        % fill Qs for ports not filled, like input port of the end mirror
        % most importantly, input to detector could be speficied, not output
        % when out is filled, while not in, propagate back to in
        % then connect the newly found one to the other end
        function fillMissingQ( obj, props )
            updated = true;
            while updated
                updated = false;
                for iopt = 1 : length( obj.optList )
                    if obj.optList(iopt).fillMissingQ( props )
                        updated = true;
                    end
                end
            end
        end
        
        % get ABCD matrix
        function [outABCD, outID] = findReflABCD( obj, portID )
            outABCD = eye(2,2);
            outID = obj.findReflPort( portID );
            [optID, portIndex] = obj.findOptic( outID );
            if optID > 0
                outABCD = obj.optList(optID).reflABCD{portIndex};
            end
            
        end
        
        function [outABCD, outID] = findTransABCD( obj, portID )
            outABCD = eye(2,2);
            outID = obj.findTransPort( portID );
            [optID, portIndex] = obj.findOptic( outID );
            if optID > 0
                outABCD = obj.optList(optID).transABCD{portIndex};
            end
        end
        
        % total number of out going fields
        function numOutgoing = findNumOutgoingFields( obj )
            numOutgoing = sum( length([obj.optList(:).outPorts]) );
        end
        
        % find portID in the reverse direction, i.e., inPortID will return outPortID, and vice sersa
        function revID = findReverseID( obj, portID )
            [optID, portIndex] = obj.findOptic( portID );
            if optID > 0
                revID = obj.findPortID( optID, -portIndex );
            else
                revID = 0;
            end
        end
        
        function info = reflInfo( obj, inID )
            opticID = obj.findOptic( inID );
            outID = obj.findReflPort(inID);
            info = sprintf( 'Reflection by ''%s'' : field ''%s'' (%d) to ''%s'' (%d)', ...
                obj.optList(opticID).name, obj.ifo.fields.fieldList(inID).name, inID, ...
                obj.ifo.fields.fieldList(outID).name, outID );
        end
        
        function info = transInfo( obj, inID )
            opticID = obj.findOptic( inID );
            outID = obj.findTransPort( inID );
            info = sprintf( 'Transmitted through ''%s'' : field ''%s'' (%d) to ''%s'' (%d)', ...
                obj.optList(opticID).name, obj.ifo.fields.fieldList(inID).name, inID, ...
                obj.ifo.fields.fieldList(outID).name, outID );
        end
        
        
        % print all optics
        function printAll( obj, fid )
            if ~exist( 'fid', 'var' )
                fid = 1;
            end
            fileOpened = ischar( fid );
            if fileOpened
                fid = fopen(fid, 'w');
            end
            
            fprintf( fid, '\n=== Information about optics ===\n\n');
            for ii = 1:length(obj.optList)
                fprintf( fid, '%d) %s\n', ii, obj.optList(ii).printInfo );
            end
            
            if fileOpened
                fclose(fid);
            end
        end
        
        function printLoss( obj, fid )
            dat = obj.lossSummary;
            lossTotal = sum( dat(:,5,2) );

            fprintf( fid, '\n=== Information about losses : total loss = %g ===\n\n', lossTotal );

            for imid = 1 : size(dat,1)
                fprintf( fid, 'loss of mirror ''%6s'' (%2d) : total loss = %g\n', ...
                    obj.optList(dat(imid,end,1)).name, dat(imid,end,1), dat(imid,end,2) );
                for ip = 1 : 4
                    if ~isnan(dat(imid,ip,2))
                        fprintf( fid, '    portID ''%s'' (%2d) : loss = %g\n', ...
                            obj.ifo.fields.fieldList(dat(imid,ip,1)).name, dat(imid,ip,1), dat(imid,ip,2) );
                    end
                end
            end            
        end
        
        % find output port after srcPort is reflected to
        % if reverse, find input port which reflects to the src Port
        function [reflID, optID] = findReflPort( obj, srcPortID, reverse )
            if ~exist( 'reverse', 'var' )
                reverse = false;
            end
            for optID = 1:length(obj.optList)
                reflID = obj.optList(optID).getReflPort(srcPortID, reverse);
                if reflID >= 0
                    return;
                end
            end
            reflID = -1;
        end
        
        function [transID, optID] = findTransPort( obj, srcPortID, reverse )
            % find output port after srcPort is transmited to
            % if reverse, find input port which transmits to the src Port
            if ~exist( 'reverse', 'var' )
                reverse = false;
            end
            for optID = 1:length(obj.optList)
                transID = obj.optList(optID).getTransPort(srcPortID, reverse);
                if transID > 0
                    return;
                end
            end
            transID = -1;
        end
        
        % add optics
        function id = addLaser( obj, name, varargin )
            
            id = length( obj.optList ) + 1;
            obj.optList( id ) = LaserObj( obj.ifo, name, varargin{:} );
            obj.ifo.fields.lamCR = obj.optList(id).lamCR;
            obj.ifo.fields.kCR = 2*pi/obj.ifo.fields.lamCR;
            
        end
        
        function [fieldMode, ind1, ind2, dxyz, nvec] = getInputType( obj )
            [~, laserOptID] = obj.findLaser;
            [fieldMode, ind1, ind2, dxyz, nvec] = obj.optList(laserOptID).getInputType;
        end
        
        function setInputType( obj, fieldMode, ind1, ind2, dxyz, nvec )
            [~, laserOptID] = obj.findLaser;
            obj.optList(laserOptID).setInputType( fieldMode, ind1, ind2, dxyz, nvec );
        end
        
        function gmul = getGouyMul( obj )
            [~, laserOptID] = obj.findLaser;
            gmul = obj.optList(laserOptID).getGouyMul;
        end
        
        function id = addDetector( obj, name )
            id = length( obj.optList ) + 1;
            obj.optList( id ) = DetectorObj( obj.ifo, name );
        end
        
        function id = addMirrorGeneric( obj, MirrorObjName, name, varargin )
            % function id = addMirrorGeneric( obj, MirrorObjName, name, varargin
            % add a mirror of class 'MirrorObjName', specified by vararg. Return 0 is failed
            % replace addMirror2x2 and addMirror4x4
            %           identified = UtilTK.identifyTags( MirrorObjH.getInputNames, varargin{:} );
            eval( strcat('identified = UtilTK.identifyTags( ', MirrorObjName, '.getInputNames, varargin{:} );') );
            if ~isempty(identified)
                id = length( obj.optList ) + 1;
                %                obj.optList( id ) = MirrorObjH( name, obj.ifo.fields, identified );
                eval( strcat('obj.optList( id ) = ', MirrorObjName, '( name, obj.ifo, identified );') );
            else
                id = 0;
            end
        end
        
        function id = findNamedOptic ( obj, name )
            % find an id of an optic by name
            if ischar(name)
                id = find( strcmp( name, { obj.optList(:).name } ), 1 );
                if isempty( id )
                    id = 0;
                end
            else
                id = name;
            end
        end
        
        % get HR side RoC of an optic specified by the port ID
        function invRoC = findInvRoCHR( obj, portID )
            opticID = obj.findOptic( portID );
            if opticID > 0
                invRoC = obj.optList(opticID).invRoC;
            else
                invRoC = 0;
            end
        end
        
        % find the optic ID and portIndex for a given portID
        function [optID, portIndex] = findOptic( obj, portID, usePortIDfrom )
            if portID > 0
                if exist('usePortIDfrom', 'var') && usePortIDfrom
                    % this is the case field object is built, so field.optID and usePortIDfrom are usable
                    if length( obj.ifo.fields.fieldList ) >= portID
                        optID = obj.ifo.fields.fieldList(portID).optID;
                        if optID > 0
                            % portIDfrom is the main fieldID or portID to which this fieldID is attached to
                            portIDfrom = obj.ifo.fields.fieldList(portID).portIDfrom;
                            if portIDfrom > 0
                                portIndex = obj.optList(optID).getPortIndex( portIDfrom );
                                if portIndex ~= 0
                                    return;
                                end
                            end
                        end
                    end
                end
                for optID = 1:length(obj.optList)
                    portIndex = obj.optList(optID).getPortIndex( portID );
                    if portIndex ~= 0
                        return;
                    end
                end
            end
            optID = 0;
            portIndex = 0;
        end
        
        % get port name by optic ID and portIndex
        % portIndex > 0 for output port, < 0 for input port
        function name = getPortName( obj, optID, portIndex, inOutName )
            if ~exist( 'inOutName', 'var' )
                inOutName = true;
            end
            if ischar( optID )
                optID = obj.findNamedOptic( optID );
            end
            name = obj.optList(optID).getPortName( 0, portIndex, inOutName );
        end
        
        % return if the portID is the one of output port
        function isOutput = isOutputPort( obj, portID )
            [~, portIndex] = obj.findOptic( portID );
            isOutput = portIndex > 0;
        end
        
        function clearTemp( obj, optIDs )
            if ~exist( 'objIDs', 'var' )
                optIDs = 1:length(obj.optList);
            end
            for id = optIDs
                obj.optList(id).cleanTemp;
            end
        end
        
        function portID = findPortID( obj, optID, portIndex )
            % function portID = findPortID( obj, optID, portIndex )
            % portIndex > 0 for output port, < 0 for input port
            % new format by one argument, like BS-HRx-o, instead of old format, 'BS', 1
            % LaserObj and DetectorObj can be specified without HR/AR nor i/o
            
            if nargin == 2 && ischar( optID )
                % this is new format, there should be three parts, optID, side and in or out
                separator = find( optID == '-' );
                if length( separator ) < 2  % not like 'ITM-HR-o'
                    if isempty( separator )
                        % 'src'
                        optIndex = obj.findNamedOptic( optID );
                    else
                        % 'src-o'
                        optIndex = obj.findNamedOptic( optID(1:separator(1)-1) );
                    end
                    
                    % this spec can be used only for LaserObj or DetectorObj
                    if obj.optList(optIndex).isLaserObj
                        portIndex = 1;
                        if length( separator ) == 1 && strcmpi(optID(separator(1)+1:end),'i')
                            portIndex = -1;
                        end
                    elseif obj.optList(optIndex).isDetectorObj
                        portIndex = -1;
                    else
                        error( 'Optics specification ''%s'' incorrect. It should be ''optID-side-i/o', optID );
                    end
                    optID = optIndex;
                else
                    portIndex = obj.portName2portIndex(optID(separator(1)+1:separator(2)-1));
                    if strcmpi(optID(separator(2)+1:end),'i')
                        portIndex = -portIndex;
                    end
                    optID = optID(1:separator(1)-1);
                end
            end
            
            if ischar( optID )
                optID = obj.findNamedOptic( optID );
            end
            
            if optID > 0
                portID = -1;
                if portIndex > 0
                    if portIndex <= length(obj.optList( optID ).outPorts)
                        portID = obj.optList( optID ).outPorts( portIndex );
                    end
                else
                    if (-portIndex) <= length(obj.optList( optID ).inPorts)
                        portID = obj.optList( optID ).inPorts( -portIndex );
                    end
                end
            else
                portID = -1;
            end
        end
    end
    
    methods ( Static )
        function portIndex = portName2portIndex( portName )
            if ischar( portName )
                switch lower(portName)
                    case {'hr', 'out', 'in', '1'}
                        portIndex = 1;
                    case 'ar'
                        portIndex = 2;
                    case 'hrx'
                        portIndex = 1;
                    case 'hry'
                        portIndex = 2;
                    case 'arx'
                        portIndex = 3;
                    case 'ary'
                        portIndex = 4;
                    otherwise
                        error('from port specification error');
                end
            else
                portIndex = portName;
            end
        end
    end
    
end
