% function [Eout, xout, yout] = singleFFT( L, Ein, xin, yin, [xout0, yout0, lam] )
%     Huygen's integral with Fresnesl approximation to one integral, which is quivalent to FFT from x to lam L / W x n
%     This ia based on the idea by Tega Edo to propagate field
%
%     Ein, xin and yin is propagated by L. xin and yin can be an array of x and y axis coordinaes or two dimentional.
%     If xout0 and yout0 are not specified, coordinates are chosen based on the FFT convention, no interpolation etc.
%     If they are given, fields are interpolated to those points using interp2.
%     If xout0 and yout0 are 0 and 0, xout and yout based on the FFT convention, lam L / W x n is used 
%     If L < 0, spherical phase exp(-i k (x^2 + y^2 ) / 2L ) is excluded
%     This is independent of SIS code

%     example to compare with SIS calculation
%{
       llo = LLODRFPMAcc('armPower',0, 'res', 0.002, 'BSID',0, 'ITMXID',0,'ETMXID',0,'ITMYID',0,'ETMYID',0);
       llo.lock('errorLimit',1e-4);
       [E2o,x2] = llo.getField('PRM2-HRy-o');   % field on PRM2 going toward PRM3
       [E3i,x3] = llo.getField('PRM3-HRx-i');   % field on PRM3 coming from PRM2
       [E3iS, x3S,y3S] = singleFFT(16.174, E2o,x2,x2, 0, 0, 1.064e-6);  % propagate E2o toward PRM3 using singleFFT
       [EIXo,xIX] = llo.getField('ITMX-HR-o');  % field on ITMX going toward ETMX
       [EEXi,xEX] = llo.getField('ETMX-HR-i');  % field on ETMX coming from ITMX
       [EEXiS, xEXS,yEXS] = singleFFT(3994.5, EIXo,xIX,xIX, 0, 0, 1.064e-6); % propagate EIXo toward ETMX using singleFFT
%}
% May 24, 2020
%   Many clean ups
%
% March 31, 2020
%   First implementation
%
function [Eout, xout, yout] = singleFFT( L, Ein, xin, yin, xout0, yout0, lam )

% in case xin and yin are one dimentioal array, make them two dimentional
if size(xin,1) == 1 && size(yin,1) == 1
    [xin,yin] = meshgrid( xin, yin );
end
N = size(xin,1);
dx = xin(1,2)-xin(1,1);
dy = yin(2,1)-yin(1,1);
if ~exist( 'lam', 'var' )
    lam = 1.064e-6;
end
lamL = lam * abs(L);
if L == 0
    xout = xin;
    yout = yin;
    Eout = Ein;
else
    xout = lamL / (N*dx) * (-N/2:N/2-1);
    yout = lamL / (N*dy) * (-N/2:N/2-1);
    [xout,yout] = meshgrid( xout, yout );
    
    Ein( isnan(Ein) ) = 0;
    Eout = 1i/lamL*fftshift(fft2(ifftshift( Ein .* exp(-1i*pi/lamL*(xin.^2+yin.^2)) ))) * dx * dy;
    if L > 0
        Eout = exp(-1i*pi/lamL*(xout.^2+yout.^2)).* Eout;
    end
end

if exist( 'xout0', 'var' ) && exist( 'yout0', 'var' ) && length( xout0 ) > 1 && length( yout0 ) > 1 
    % in case xout0 and yout0 are one dimentioal array, make them two dimentional
    if size(xout0,1) == 1 && size(yout0,1) == 1
        [xout0,yout0] = meshgrid( xout0, yout0 );
    end
    %{
    indx = find( xout0(1,:) <= xout(end) & xout0(1,:) >= xout(1) );
    indy = find( yout0(:,1) <= yout(end) & yout0(:,1) >= yout(1) );
    xout0 = xout0( indy, indx );
    yout0 = yout0( indy, indx );
    Eout = interp2( xout, yout, Eout, xout0, yout0 );
    xout = xout0;
    yout = yout0;
    %}
    Eout = interp2( xout, yout, Eout, xout0, yout0, 'cubic', 0 );
    xout = xout0;
    yout = yout0;
end